@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one
  * or more contributor license agreements.  See the NOTICE file
  * distributed with this work for additional information
@@ -48,6 +48,18 @@ public interface RaftClientConfigKeys {
     static void setRequestTimeout(RaftProperties properties, TimeDuration timeoutDuration) {
       setTimeDuration(properties::setTimeDuration, REQUEST_TIMEOUT_KEY, timeoutDuration);
     }
+
+    String WATCH_REQUEST_TIMEOUT_KEY = PREFIX + ".watch.request.timeout";
+    TimeDuration WATCH_REQUEST_TIMEOUT_DEFAULT =
+        TimeDuration.valueOf(10000, TimeUnit.MILLISECONDS);
+    static TimeDuration watchRequestTimeout(RaftProperties properties) {
+      return getTimeDuration(properties.getTimeDuration(WATCH_REQUEST_TIMEOUT_DEFAULT.getUnit()),
+          WATCH_REQUEST_TIMEOUT_KEY, WATCH_REQUEST_TIMEOUT_DEFAULT, getDefaultLog());
+    }
+    static void setWatchRequestTimeout(RaftProperties properties,
+        TimeDuration timeoutDuration) {
+      setTimeDuration(properties::setTimeDuration, WATCH_REQUEST_TIMEOUT_KEY, timeoutDuration);
+    }
   }
 
   interface Async {
@@ -63,14 +75,17 @@ public interface RaftClientConfigKeys {
       setInt(properties::setInt, MAX_OUTSTANDING_REQUESTS_KEY, outstandingRequests);
     }
 
-    String SCHEDULER_THREADS_KEY = PREFIX + ".scheduler-threads";
-    int SCHEDULER_THREADS_DEFAULT = 3;
-    static int schedulerThreads(RaftProperties properties) {
-      return getInt(properties::getInt, SCHEDULER_THREADS_KEY,
-          SCHEDULER_THREADS_DEFAULT, getDefaultLog(), requireMin(1));
-    }
-    static void setSchedulerThreads(RaftProperties properties, int schedulerThreads) {
-      setInt(properties::setInt, SCHEDULER_THREADS_KEY, schedulerThreads);
+    interface Experimental {
+      String PREFIX = Async.PREFIX + "." + Experimental.class.getSimpleName().toLowerCase();
+
+      String SEND_DUMMY_REQUEST_KEY = PREFIX + ".send-dummy-request";
+      boolean SEND_DUMMY_REQUEST_DEFAULT = true;
+      static boolean sendDummyRequest(RaftProperties properties) {
+        return getBoolean(properties::getBoolean, SEND_DUMMY_REQUEST_KEY, SEND_DUMMY_REQUEST_DEFAULT, getDefaultLog());
+      }
+      static void setSendDummyRequest(RaftProperties properties, boolean sendDummyRequest) {
+        setBoolean(properties::setBoolean, SEND_DUMMY_REQUEST_KEY, sendDummyRequest);
+      }
     }
   }
 
