@@ -17,16 +17,18 @@
  */
 package org.apache.ratis.client.impl;
 
+import java.util.Optional;
 import org.apache.ratis.protocol.*;
 import org.apache.ratis.thirdparty.com.google.protobuf.ByteString;
 import org.apache.ratis.proto.RaftProtos.*;
 import org.apache.ratis.util.ProtoUtils;
 import org.apache.ratis.util.ReflectionUtils;
 
-import java.util.Arrays;
 import java.util.List;
 import java.util.stream.Collectors;
 
+import static org.apache.ratis.proto.RaftProtos.RaftClientReplyProto.ExceptionDetailsCase.ALREADYCLOSEDEXCEPTION;
+import static org.apache.ratis.proto.RaftProtos.RaftClientReplyProto.ExceptionDetailsCase.LEADERNOTREADYEXCEPTION;
 import static org.apache.ratis.proto.RaftProtos.RaftClientReplyProto.ExceptionDetailsCase.NOTLEADEREXCEPTION;
 import static org.apache.ratis.proto.RaftProtos.RaftClientReplyProto.ExceptionDetailsCase.NOTREPLICATEDEXCEPTION;
 import static org.apache.ratis.proto.RaftProtos.RaftClientReplyProto.ExceptionDetailsCase.STATEMACHINEEXCEPTION;
@@ -155,7 +157,7 @@ public interface ClientProtoUtils {
       if (reply.getMessage() != null) {
         b.setMessage(toClientMessageEntryProtoBuilder(reply.getMessage()));
       }
-      ProtoUtils.addCommitInfos(reply.getCommitInfos(), i -> b.addCommitInfos(i));
+      ProtoUtils.addCommitInfos(reply.getCommitInfos(), b::addCommitInfos);
 
       final NotLeaderException nle = reply.getNotLeaderException();
       final StateMachineException sme;
@@ -166,8 +168,7 @@ public interface ClientProtoUtils {
         if (suggestedLeader != null) {
           nleBuilder.setSuggestedLeader(suggestedLeader.getRaftPeerProto());
         }
-        nleBuilder.addAllPeersInConf(
-            ProtoUtils.toRaftPeerProtos(Arrays.asList(nle.getPeers())));
+        nleBuilder.addAllPeersInConf(ProtoUtils.toRaftPeerProtos(nle.getPeers()));
         b.setNotLeaderException(nleBuilder.build());
       } else if ((sme = reply.getStateMachineException()) != null) {
         StateMachineExceptionProto.Builder smeBuilder =
@@ -187,6 +188,35 @@ public interface ClientProtoUtils {
             .setLogIndex(nre.getLogIndex());
         b.setNotReplicatedException(nreBuilder);
       }
+
+      final LeaderNotReadyException lnre = reply.getLeaderNotReadyException();
+      if (lnre != null) {
+        LeaderNotReadyExceptionProto.Builder lnreBuilder = LeaderNotReadyExceptionProto.newBuilder()
+            .setServerId(ProtoUtils.toRaftGroupMemberIdProtoBuilder(lnre.getServerId()));
+        b.setLeaderNotReadyException(lnreBuilder);
+      }
+
+      final AlreadyClosedException ace = reply.getAlreadyClosedException();
+      if (ace != null) {
+        final Throwable t = ace.getCause() != null ? ace.getCause() : ace;
+        AlreadyClosedExceptionProto.Builder aceBuilder = AlreadyClosedExceptionProto.newBuilder()
+            .setExceptionClassName(t.getClass().getName())
+            .setErrorMsg(ace.getMessage())
+            .setStacktrace(ProtoUtils.writeObject2ByteString(ace.getStackTrace()));
+        b.setAlreadyClosedException(aceBuilder);
+      }
+
+      final RaftClientReplyProto serialized = b.build();
+      final RaftException e = reply.getException();
+      if (e != null) {
+        final RaftClientReply deserialized = toRaftClientReply(serialized);
+        if (!Optional.ofNullable(deserialized.getException())
+            .map(Object::getClass).filter(e.getClass()::equals).isPresent()) {
+          throw new AssertionError("Corruption while serializing reply= " + reply
+              + " but serialized=" + serialized + " and deserialized=" + deserialized, e);
+        }
+      }
+      return serialized;
     }
     return b.build();
   }
@@ -217,39 +247,42 @@ public interface ClientProtoUtils {
         b.setGroup(ProtoUtils.toRaftGroupProtoBuilder(reply.getGroup()));
         b.setIsRaftStorageHealthy(reply.isRaftStorageHealthy());
         b.setRole(reply.getRoleInfoProto());
-        ProtoUtils.addCommitInfos(reply.getCommitInfos(), i -> b.addCommitInfos(i));
+        ProtoUtils.addCommitInfos(reply.getCommitInfos(), b::addCommitInfos);
       }
     }
     return b.build();
   }
 
-  static RaftClientReply toRaftClientReply(
-      RaftClientReplyProto replyProto) {
+  static RaftClientReply toRaftClientReply(RaftClientReplyProto replyProto) {
     final RaftRpcReplyProto rp = replyProto.getRpcReply();
+    final RaftGroupMemberId serverMemberId = ProtoUtils.toRaftGroupMemberId(rp.getReplyId(), rp.getRaftGroupId());
+
     final RaftException e;
     if (replyProto.getExceptionDetailsCase().equals(NOTLEADEREXCEPTION)) {
       NotLeaderExceptionProto nleProto = replyProto.getNotLeaderException();
       final RaftPeer suggestedLeader = nleProto.hasSuggestedLeader() ?
           ProtoUtils.toRaftPeer(nleProto.getSuggestedLeader()) : null;
-      final RaftPeer[] peers = ProtoUtils.toRaftPeerArray(
-          nleProto.getPeersInConfList());
-      e = new NotLeaderException(RaftPeerId.valueOf(rp.getReplyId()),
-          suggestedLeader, peers);
+      final List<RaftPeer> peers = ProtoUtils.toRaftPeers(nleProto.getPeersInConfList());
+      e = new NotLeaderException(serverMemberId, suggestedLeader, peers);
     } else if (replyProto.getExceptionDetailsCase() == NOTREPLICATEDEXCEPTION) {
       final NotReplicatedExceptionProto nre = replyProto.getNotReplicatedException();
       e = new NotReplicatedException(nre.getCallId(), nre.getReplication(), nre.getLogIndex());
     } else if (replyProto.getExceptionDetailsCase().equals(STATEMACHINEEXCEPTION)) {
       StateMachineExceptionProto smeProto = replyProto.getStateMachineException();
-      e = wrapStateMachineException(RaftPeerId.valueOf(rp.getReplyId()),
-          smeProto.getExceptionClassName(), smeProto.getErrorMsg(),
-          smeProto.getStacktrace());
+      e = wrapStateMachineException(serverMemberId,
+          smeProto.getExceptionClassName(), smeProto.getErrorMsg(), smeProto.getStacktrace());
+    } else if (replyProto.getExceptionDetailsCase().equals(LEADERNOTREADYEXCEPTION)) {
+      LeaderNotReadyExceptionProto lnreProto = replyProto.getLeaderNotReadyException();
+      e = new LeaderNotReadyException(ProtoUtils.toRaftGroupMemberId(lnreProto.getServerId()));
+    } else if (replyProto.getExceptionDetailsCase().equals(ALREADYCLOSEDEXCEPTION)) {
+      AlreadyClosedExceptionProto aceProto = replyProto.getAlreadyClosedException();
+      e = wrapAlreadyClosedException(aceProto.getExceptionClassName(),
+          aceProto.getErrorMsg(), aceProto.getStacktrace());
     } else {
       e = null;
     }
     ClientId clientId = ClientId.valueOf(rp.getRequestorId());
-    final RaftGroupId groupId = ProtoUtils.toRaftGroupId(rp.getRaftGroupId());
-    return new RaftClientReply(clientId, RaftPeerId.valueOf(rp.getReplyId()),
-        groupId, rp.getCallId(), rp.getSuccess(),
+    return new RaftClientReply(clientId, serverMemberId, rp.getCallId(), rp.getSuccess(),
         toMessage(replyProto.getMessage()), e,
         replyProto.getLogIndex(), replyProto.getCommitInfosList());
   }
@@ -280,8 +313,7 @@ public interface ClientProtoUtils {
   }
 
   static StateMachineException wrapStateMachineException(
-      RaftPeerId serverId, String className, String errorMsg,
-      ByteString stackTraceBytes) {
+      RaftGroupMemberId memberId, String className, String errorMsg, ByteString stackTraceBytes) {
     StateMachineException sme;
     if (className == null) {
       sme = new StateMachineException(errorMsg);
@@ -290,7 +322,7 @@ public interface ClientProtoUtils {
         Class<?> clazz = Class.forName(className);
         final Exception e = ReflectionUtils.instantiateException(
             clazz.asSubclass(Exception.class), errorMsg, null);
-        sme = new StateMachineException(serverId, e);
+        sme = new StateMachineException(memberId, e);
       } catch (Exception e) {
         sme = new StateMachineException(className + ": " + errorMsg);
       }
@@ -301,6 +333,27 @@ public interface ClientProtoUtils {
     return sme;
   }
 
+  static AlreadyClosedException wrapAlreadyClosedException(
+      String className, String errorMsg, ByteString stackTraceBytes) {
+    AlreadyClosedException ace;
+    if (className == null) {
+      ace = new AlreadyClosedException(errorMsg);
+    } else {
+      try {
+        Class<?> clazz = Class.forName(className);
+        final Exception e = ReflectionUtils.instantiateException(
+            clazz.asSubclass(Exception.class), errorMsg, null);
+        ace = new AlreadyClosedException(errorMsg, e);
+      } catch (Exception e) {
+        ace = new AlreadyClosedException(className + ": " + errorMsg);
+      }
+    }
+    StackTraceElement[] stacktrace =
+        (StackTraceElement[]) ProtoUtils.toObject(stackTraceBytes);
+    ace.setStackTrace(stacktrace);
+    return ace;
+  }
+
   static Message toMessage(final ClientMessageEntryProto p) {
     return Message.valueOf(p.getContent());
   }
@@ -316,7 +369,7 @@ public interface ClientProtoUtils {
   static SetConfigurationRequest toSetConfigurationRequest(
       SetConfigurationRequestProto p) {
     final RaftRpcRequestProto m = p.getRpcRequest();
-    final RaftPeer[] peers = ProtoUtils.toRaftPeerArray(p.getPeersList());
+    final List<RaftPeer> peers = ProtoUtils.toRaftPeers(p.getPeersList());
     return new SetConfigurationRequest(
         ClientId.valueOf(m.getRequestorId()),
         RaftPeerId.valueOf(m.getReplyId()),
@@ -328,8 +381,7 @@ public interface ClientProtoUtils {
       SetConfigurationRequest request) {
     return SetConfigurationRequestProto.newBuilder()
         .setRpcRequest(toRaftRpcRequestProtoBuilder(request))
-        .addAllPeers(ProtoUtils.toRaftPeerProtos(
-            Arrays.asList(request.getPeersInNewConf())))
+        .addAllPeers(ProtoUtils.toRaftPeerProtos(request.getPeersInNewConf()))
         .build();
   }
 
