@@ -17,31 +17,35 @@
  */
 package org.apache.ratis.client.impl;
 
+import org.apache.ratis.client.ClientRetryEvent;
 import org.apache.ratis.client.RaftClient;
-import org.apache.ratis.client.RaftClientConfigKeys;
 import org.apache.ratis.client.RaftClientRpc;
 import org.apache.ratis.conf.RaftProperties;
+import org.apache.ratis.proto.RaftProtos.RaftClientRequestProto.TypeCase;
 import org.apache.ratis.proto.RaftProtos.ReplicationLevel;
 import org.apache.ratis.proto.RaftProtos.SlidingWindowEntry;
 import org.apache.ratis.protocol.*;
 import org.apache.ratis.retry.RetryPolicy;
-import org.apache.ratis.util.*;
-import org.apache.ratis.util.function.FunctionUtils;
+import org.apache.ratis.util.CollectionUtils;
+import org.apache.ratis.util.JavaUtils;
+import org.apache.ratis.util.Preconditions;
+import org.apache.ratis.util.TimeoutScheduler;
 
 import java.io.IOException;
 import java.io.InterruptedIOException;
-import java.util.*;
-import java.util.concurrent.*;
+import java.util.Arrays;
+import java.util.Collection;
+import java.util.Objects;
+import java.util.Optional;
+import java.util.concurrent.CompletableFuture;
+import java.util.concurrent.ConcurrentLinkedQueue;
 import java.util.concurrent.atomic.AtomicInteger;
 import java.util.concurrent.atomic.AtomicLong;
+import java.util.function.Consumer;
 import java.util.function.Function;
-import java.util.function.LongFunction;
 import java.util.function.Supplier;
 import java.util.stream.Stream;
 
-import static org.apache.ratis.proto.RaftProtos.RaftClientRequestProto.TypeCase.STALEREAD;
-import static org.apache.ratis.proto.RaftProtos.RaftClientRequestProto.TypeCase.WATCH;
-
 /** A client who sends requests to a raft service. */
 final class RaftClientImpl implements RaftClient {
   private static final AtomicLong CALL_ID_COUNTER = new AtomicLong();
@@ -70,53 +74,6 @@ final class RaftClientImpl implements RaftClient {
     }
   }
 
-  static class PendingAsyncRequest extends PendingClientRequest
-      implements SlidingWindow.ClientSideRequest<RaftClientReply> {
-    private final Function<SlidingWindowEntry, RaftClientRequest> requestConstructor;
-    private final long seqNum;
-    private volatile boolean isFirst = false;
-
-    PendingAsyncRequest(long seqNum, Function<SlidingWindowEntry, RaftClientRequest> requestConstructor) {
-      this.seqNum = seqNum;
-      this.requestConstructor = requestConstructor;
-    }
-
-    @Override
-    RaftClientRequest newRequestImpl() {
-      return requestConstructor.apply(ProtoUtils.toSlidingWindowEntry(seqNum, isFirst));
-    }
-
-    @Override
-    public void setFirstRequest() {
-      isFirst = true;
-    }
-
-    @Override
-    public long getSeqNum() {
-      return seqNum;
-    }
-
-    @Override
-    public boolean hasReply() {
-      return getReplyFuture().isDone();
-    }
-
-    @Override
-    public void setReply(RaftClientReply reply) {
-      getReplyFuture().complete(reply);
-    }
-
-    @Override
-    public void fail(Throwable e) {
-      getReplyFuture().completeExceptionally(e);
-    }
-
-    @Override
-    public String toString() {
-      return "[seq=" + getSeqNum() + "]";
-    }
-  }
-
   private final ClientId clientId;
   private final RaftClientRpc clientRpc;
   private final Collection<RaftPeer> peers;
@@ -125,11 +82,9 @@ final class RaftClientImpl implements RaftClient {
 
   private volatile RaftPeerId leaderId;
 
-  /** Map: id -> {@link SlidingWindow}, in order to support async calls to the RAFT service or individual servers. */
-  private final ConcurrentMap<String, SlidingWindow.Client<PendingAsyncRequest, RaftClientReply>>
-      slidingWindows = new ConcurrentHashMap<>();
   private final TimeoutScheduler scheduler;
-  private final Semaphore asyncRequestSemaphore;
+
+  private final Supplier<OrderedAsync> orderedAsync;
 
   RaftClientImpl(ClientId clientId, RaftGroup group, RaftPeerId leaderId,
       RaftClientRpc clientRpc, RaftProperties properties, RetryPolicy retryPolicy) {
@@ -142,9 +97,10 @@ final class RaftClientImpl implements RaftClient {
     Preconditions.assertTrue(retryPolicy != null, "retry policy can't be null");
     this.retryPolicy = retryPolicy;
 
-    asyncRequestSemaphore = new Semaphore(RaftClientConfigKeys.Async.maxOutstandingRequests(properties));
-    scheduler = TimeoutScheduler.newInstance(RaftClientConfigKeys.Async.schedulerThreads(properties));
+    scheduler = TimeoutScheduler.getInstance();
     clientRpc.addServers(peers);
+
+    this.orderedAsync = JavaUtils.memoize(() -> OrderedAsync.newInstance(this, properties));
   }
 
   @Override
@@ -160,13 +116,8 @@ final class RaftClientImpl implements RaftClient {
     return scheduler;
   }
 
-  private SlidingWindow.Client<PendingAsyncRequest, RaftClientReply> getSlidingWindow(RaftClientRequest request) {
-    return getSlidingWindow(request.is(STALEREAD)? request.getServerId(): null);
-  }
-
-  private SlidingWindow.Client<PendingAsyncRequest, RaftClientReply> getSlidingWindow(RaftPeerId target) {
-    final String id = target != null? target.toString(): "RAFT";
-    return slidingWindows.computeIfAbsent(id, key -> new SlidingWindow.Client<>(getId() + "->" + key));
+  OrderedAsync getOrderedAsync() {
+    return orderedAsync.get();
   }
 
   @Override
@@ -191,23 +142,7 @@ final class RaftClientImpl implements RaftClient {
 
   private CompletableFuture<RaftClientReply> sendAsync(
       RaftClientRequest.Type type, Message message, RaftPeerId server) {
-    if (!type.is(WATCH)) {
-      Objects.requireNonNull(message, "message == null");
-    }
-    try {
-      asyncRequestSemaphore.acquire();
-    } catch (InterruptedException e) {
-      return JavaUtils.completeExceptionally(IOUtils.toInterruptedIOException(
-          "Interrupted when sending " + type + ", message=" + message, e));
-    }
-
-    final long callId = nextCallId();
-    final LongFunction<PendingAsyncRequest> constructor = seqNum -> new PendingAsyncRequest(seqNum,
-        slidingWindowEntry -> newRaftClientRequest(server, callId, message, type, slidingWindowEntry));
-    return getSlidingWindow(server).submitNewRequest(constructor, this::sendRequestWithRetryAsync
-    ).getReplyFuture(
-    ).thenApply(reply -> handleStateMachineException(reply, CompletionException::new)
-    ).whenComplete((r, e) -> asyncRequestSemaphore.release());
+    return getOrderedAsync().send(type, message, server);
   }
 
   RaftClientRequest newRaftClientRequest(
@@ -240,7 +175,7 @@ final class RaftClientImpl implements RaftClient {
 
   private RaftClientReply send(RaftClientRequest.Type type, Message message, RaftPeerId server)
       throws IOException {
-    if (!type.is(WATCH)) {
+    if (!type.is(TypeCase.WATCH)) {
       Objects.requireNonNull(message, "message == null");
     }
 
@@ -248,6 +183,7 @@ final class RaftClientImpl implements RaftClient {
     return sendRequestWithRetry(() -> newRaftClientRequest(server, callId, message, type, null));
   }
 
+  // TODO: change peersInNewConf to List<RaftPeer>
   @Override
   public RaftClientReply setConfiguration(RaftPeer[] peersInNewConf)
       throws IOException {
@@ -257,7 +193,7 @@ final class RaftClientImpl implements RaftClient {
     // also refresh the rpc proxies for these peers
     addServers(Arrays.stream(peersInNewConf));
     return sendRequestWithRetry(() -> new SetConfigurationRequest(
-        clientId, leaderId, groupId, callId, peersInNewConf));
+        clientId, leaderId, groupId, callId, Arrays.asList(peersInNewConf)));
   }
 
   @Override
@@ -303,146 +239,95 @@ final class RaftClientImpl implements RaftClient {
         peersInNewConf.filter(p -> !peers.contains(p))::iterator);
   }
 
-  private void sendRequestWithRetryAsync(PendingAsyncRequest pending) {
-    final CompletableFuture<RaftClientReply> f = pending.getReplyFuture();
-    if (f.isDone()) {
-      return;
-    }
-
-    final RaftClientRequest request = pending.newRequest();
-    sendRequestAsync(request, pending.getAttemptCount()).thenAccept(reply -> {
-      if (f.isDone()) {
-        return;
-      }
-      if (reply == null) {
-        LOG.debug("schedule* attempt #{} with policy {} for {}", pending.getAttemptCount(), retryPolicy, request);
-        scheduler.onTimeout(retryPolicy.getSleepTime(),
-            () -> getSlidingWindow(request).retry(pending, this::sendRequestWithRetryAsync),
-            LOG, () -> "Failed* to retry " + request);
-      } else {
-        f.complete(reply);
-      }
-    }).exceptionally(FunctionUtils.consumerAsNullFunction(f::completeExceptionally));
-  }
-
-  private RaftClientReply sendRequestWithRetry(
-      Supplier<RaftClientRequest> supplier)
-      throws InterruptedIOException, StateMachineException, GroupMismatchException {
-    for(int attemptCount = 0;; attemptCount++) {
+  private RaftClientReply sendRequestWithRetry(Supplier<RaftClientRequest> supplier) throws IOException {
+    for(int attemptCount = 1;; attemptCount++) {
       final RaftClientRequest request = supplier.get();
-      final RaftClientReply reply = sendRequest(request);
-      if (reply != null) {
-        return reply;
-      }
-      if (!retryPolicy.shouldRetry(attemptCount)) {
-        return null;
-      }
+      IOException ioe = null;
       try {
-        retryPolicy.getSleepTime().sleep();
+        final RaftClientReply reply = sendRequest(request);
+        if (reply != null) {
+          return reply;
+        }
+      } catch (GroupMismatchException | StateMachineException e) {
+        throw e;
+      } catch (IOException e) {
+        ioe = e;
+      }
+
+      final ClientRetryEvent event = new ClientRetryEvent(attemptCount, request, ioe);
+      final RetryPolicy.Action action = retryPolicy.handleAttemptFailure(event);
+      if (!action.shouldRetry()) {
+        throw (IOException)noMoreRetries(event);
+      }
+
+      try {
+        action.getSleepTime().sleep();
       } catch (InterruptedException e) {
         throw new InterruptedIOException("retry policy=" + retryPolicy);
       }
     }
   }
 
-  private CompletableFuture<RaftClientReply> sendRequestAsync(
-      RaftClientRequest request, int attemptCount) {
-    LOG.debug("{}: send* {}", clientId, request);
-    return clientRpc.sendRequestAsync(request).thenApply(reply -> {
-      LOG.debug("{}: receive* {}", clientId, reply);
-      reply = handleNotLeaderException(request, reply, true);
-      if (reply != null) {
-        getSlidingWindow(request).receiveReply(
-            request.getSlidingWindowEntry().getSeqNum(), reply, this::sendRequestWithRetryAsync);
-      } else if (!retryPolicy.shouldRetry(attemptCount)) {
-        handleAsyncRetryFailure(request, attemptCount);
-      }
-      return reply;
-    }).exceptionally(e -> {
-      if (LOG.isTraceEnabled()) {
-        LOG.trace(clientId + ": Failed* " + request, e);
-      } else {
-        LOG.debug("{}: Failed* {} with {}", clientId, request, e);
-      }
-      e = JavaUtils.unwrapCompletionException(e);
-      if (e instanceof IOException && !(e instanceof GroupMismatchException)) {
-        if (!retryPolicy.shouldRetry(attemptCount)) {
-          handleAsyncRetryFailure(request, attemptCount);
-        } else {
-          handleIOException(request, (IOException) e, null, true);
-        }
-        return null;
-      }
-      failAllAsyncRequests(request, e);
-      return null;
-    });
+  Throwable noMoreRetries(ClientRetryEvent event) {
+    final int attemptCount = event.getAttemptCount();
+    final Throwable throwable = event.getCause();
+    if (attemptCount == 1 && throwable != null) {
+      return throwable;
+    }
+    return new RaftRetryFailureException(event.getRequest(), attemptCount, retryPolicy, throwable);
   }
 
-  static RaftRetryFailureException newRaftRetryFailureException(
-      RaftClientRequest request, int attemptCount, RetryPolicy retryPolicy) {
-    return new RaftRetryFailureException(
-        "Failed " + request + " for " + (attemptCount-1) + " attempts with " + retryPolicy);
-  }
-
-  private void handleAsyncRetryFailure(RaftClientRequest request, int attemptCount) {
-    final RaftRetryFailureException rfe = newRaftRetryFailureException(request, attemptCount, retryPolicy);
-    failAllAsyncRequests(request, rfe);
-  }
-
-  private void failAllAsyncRequests(RaftClientRequest request, Throwable t) {
-    getSlidingWindow(request).fail(request.getSlidingWindowEntry().getSeqNum(), t);
-  }
-
-  private RaftClientReply sendRequest(RaftClientRequest request)
-      throws StateMachineException, GroupMismatchException {
+  private RaftClientReply sendRequest(RaftClientRequest request) throws IOException {
     LOG.debug("{}: send {}", clientId, request);
-    RaftClientReply reply = null;
+    RaftClientReply reply;
     try {
       reply = clientRpc.sendRequest(request);
     } catch (GroupMismatchException gme) {
       throw gme;
     } catch (IOException ioe) {
-      handleIOException(request, ioe, null, false);
+      handleIOException(request, ioe);
+      throw ioe;
     }
     LOG.debug("{}: receive {}", clientId, reply);
-    reply = handleNotLeaderException(request, reply, false);
-    reply = handleStateMachineException(reply, Function.identity());
+    reply = handleLeaderException(request, reply, null);
+    reply = handleRaftException(reply, Function.identity());
     return reply;
   }
 
-  static <E extends Throwable> RaftClientReply handleStateMachineException(
-      RaftClientReply reply, Function<StateMachineException, E> converter) throws E {
+  static <E extends Throwable> RaftClientReply handleRaftException(
+      RaftClientReply reply, Function<RaftException, E> converter) throws E {
     if (reply != null) {
-      final StateMachineException sme = reply.getStateMachineException();
-      if (sme != null) {
-        throw converter.apply(sme);
+      final RaftException e = reply.getException();
+      if (e != null) {
+        throw converter.apply(e);
       }
     }
     return reply;
   }
 
   /**
-   * @return null if the reply is null or it has {@link NotLeaderException};
-   *         otherwise return the same reply.
+   * @return null if the reply is null or it has
+   * {@link NotLeaderException} or {@link LeaderNotReadyException}
+   * otherwise return the same reply.
    */
-  RaftClientReply handleNotLeaderException(RaftClientRequest request, RaftClientReply reply,
-      boolean resetSlidingWindow) {
-    if (reply == null) {
+  RaftClientReply handleLeaderException(RaftClientRequest request, RaftClientReply reply,
+                                        Consumer<RaftClientRequest> handler) {
+    if (reply == null || reply.getException() instanceof LeaderNotReadyException) {
       return null;
     }
     final NotLeaderException nle = reply.getNotLeaderException();
     if (nle == null) {
       return reply;
     }
-    return handleNotLeaderException(request, nle, resetSlidingWindow);
+    return handleNotLeaderException(request, nle, handler);
   }
 
   RaftClientReply handleNotLeaderException(RaftClientRequest request, NotLeaderException nle,
-      boolean resetSlidingWindow) {
-    refreshPeers(Arrays.asList(nle.getPeers()));
+      Consumer<RaftClientRequest> handler) {
+    refreshPeers(nle.getPeers());
     final RaftPeerId newLeader = nle.getSuggestedLeader() == null ? null
         : nle.getSuggestedLeader().getId();
-    handleIOException(request, nle, newLeader, resetSlidingWindow);
+    handleIOException(request, nle, newLeader, handler);
     return null;
   }
 
@@ -455,45 +340,43 @@ final class RaftClientImpl implements RaftClient {
     }
   }
 
+  void handleIOException(RaftClientRequest request, IOException ioe) {
+    handleIOException(request, ioe, null, null);
+  }
+
   void handleIOException(RaftClientRequest request, IOException ioe,
-      RaftPeerId newLeader, boolean resetSlidingWindow) {
+      RaftPeerId newLeader, Consumer<RaftClientRequest> handler) {
     LOG.debug("{}: suggested new leader: {}. Failed {} with {}",
         clientId, newLeader, request, ioe);
     if (LOG.isTraceEnabled()) {
       LOG.trace("Stack trace", new Throwable("TRACE"));
     }
 
-    if (resetSlidingWindow) {
-      getSlidingWindow(request).resetFirstSeqNum();
-    }
+    Optional.ofNullable(handler).ifPresent(h -> h.accept(request));
+
     if (ioe instanceof LeaderNotReadyException) {
       return;
     }
 
     final RaftPeerId oldLeader = request.getServerId();
-    final RaftPeerId curLeader = request.getServerId();
+    final RaftPeerId curLeader = leaderId;
     final boolean stillLeader = oldLeader.equals(curLeader);
     if (newLeader == null && stillLeader) {
       newLeader = CollectionUtils.random(oldLeader,
           CollectionUtils.as(peers, RaftPeer::getId));
     }
-    LOG.debug("{}: oldLeader={},  curLeader={}, newLeader{}", clientId, oldLeader, curLeader, newLeader);
+    LOG.debug("{}: oldLeader={},  curLeader={}, newLeader={}", clientId, oldLeader, curLeader, newLeader);
 
     final boolean changeLeader = newLeader != null && stillLeader;
-    if (changeLeader) {
-      LOG.debug("{}: change Leader from {} to {}", clientId, oldLeader, newLeader);
-      this.leaderId = newLeader;
+    final boolean reconnect = changeLeader || clientRpc.shouldReconnect(ioe);
+    if (reconnect) {
+      if (changeLeader && oldLeader.equals(leaderId)) {
+        LOG.debug("{} {}: client change Leader from {} to {} ex={}", groupId,
+            clientId, oldLeader, newLeader, ioe.getClass().getName());
+        this.leaderId = newLeader;
+      }
+      clientRpc.handleException(oldLeader, ioe, true);
     }
-    clientRpc.handleException(oldLeader, ioe, changeLeader);
-  }
-
-  void assertAsyncRequestSemaphore(int expectedAvailablePermits, int expectedQueueLength) {
-    Preconditions.assertTrue(asyncRequestSemaphore.availablePermits() == expectedAvailablePermits);
-    Preconditions.assertTrue(asyncRequestSemaphore.getQueueLength() == expectedQueueLength);
-  }
-
-  void assertScheduler(int numThreads) {
-    Preconditions.assertTrue(scheduler.getNumThreads() == numThreads);
   }
 
   long getCallId() {
@@ -507,6 +390,7 @@ final class RaftClientImpl implements RaftClient {
 
   @Override
   public void close() throws IOException {
+    scheduler.close();
     clientRpc.close();
   }
 }
