@@ -17,12 +17,15 @@
  */
 package org.apache.ratis.client.impl;
 
+import org.apache.ratis.client.ClientRetryEvent;
 import org.apache.ratis.client.impl.RaftClientImpl.PendingClientRequest;
 import org.apache.ratis.protocol.ClientId;
 import org.apache.ratis.protocol.GroupMismatchException;
 import org.apache.ratis.protocol.NotLeaderException;
 import org.apache.ratis.protocol.RaftClientReply;
 import org.apache.ratis.protocol.RaftClientRequest;
+import org.apache.ratis.protocol.RaftException;
+import org.apache.ratis.retry.RetryPolicies;
 import org.apache.ratis.retry.RetryPolicy;
 import org.apache.ratis.util.JavaUtils;
 import org.slf4j.Logger;
@@ -56,7 +59,7 @@ public interface UnorderedAsync {
         () -> client.newRaftClientRequest(null, callId, null, type, null));
     sendRequestWithRetry(pending, client);
     return pending.getReplyFuture()
-        .thenApply(reply -> RaftClientImpl.handleStateMachineException(reply, CompletionException::new));
+        .thenApply(reply -> RaftClientImpl.handleRaftException(reply, CompletionException::new));
   }
 
   static void sendRequestWithRetry(PendingClientRequest pending, RaftClientImpl client) {
@@ -73,14 +76,18 @@ public interface UnorderedAsync {
     client.getClientRpc().sendRequestAsyncUnordered(request).whenCompleteAsync((reply, e) -> {
       try {
         LOG.debug("{}: attempt #{} receive~ {}", clientId, attemptCount, reply);
-        reply = client.handleNotLeaderException(request, reply, false);
+        final RaftException replyException = reply != null? reply.getException(): null;
+        reply = client.handleLeaderException(request, reply, null);
         if (reply != null) {
           f.complete(reply);
           return;
         }
-        final RetryPolicy retryPolicy = client.getRetryPolicy();
-        if (!retryPolicy.shouldRetry(attemptCount)) {
-          f.completeExceptionally(RaftClientImpl.newRaftRetryFailureException(request, attemptCount, retryPolicy));
+        RetryPolicy retryPolicy = client.getRetryPolicy();
+        final ClientRetryEvent event = new ClientRetryEvent(attemptCount, request,
+            replyException != null? replyException: e);
+        final RetryPolicy.Action action = retryPolicy.handleAttemptFailure(event);
+        if (!action.shouldRetry()) {
+          f.completeExceptionally(client.noMoreRetries(event));
           return;
         }
 
@@ -94,9 +101,14 @@ public interface UnorderedAsync {
 
           if (e instanceof IOException) {
             if (e instanceof NotLeaderException) {
-              client.handleNotLeaderException(request, (NotLeaderException) e, false);
-            } else if (!(e instanceof GroupMismatchException)) {
-              client.handleIOException(request, (IOException) e, null, false);
+              client.handleNotLeaderException(request, (NotLeaderException) e, null);
+              retryPolicy = ((NotLeaderException) e).getSuggestedLeader() != null ?
+                  RetryPolicies.retryForeverNoSleep() : retryPolicy;
+            } else if (e instanceof GroupMismatchException) {
+              f.completeExceptionally(e);
+              return;
+            } else {
+              client.handleIOException(request, (IOException) e);
             }
           } else {
             if (!client.getClientRpc().handleException(request.getServerId(), e, false)) {
@@ -107,10 +119,10 @@ public interface UnorderedAsync {
         }
 
         LOG.debug("schedule retry for attempt #{}, policy={}, request={}", attemptCount, retryPolicy, request);
-        client.getScheduler().onTimeout(retryPolicy.getSleepTime(), () -> sendRequestWithRetry(pending, client),
-            LOG, () -> clientId + ": Failed~ to retry " + request);
+        client.getScheduler().onTimeout(action.getSleepTime(),
+            () -> sendRequestWithRetry(pending, client), LOG, () -> clientId + ": Failed~ to retry " + request);
       } catch (Throwable t) {
-        LOG.error(clientId + ": XXX Failed " + request, t);
+        LOG.error(clientId + ": Failed " + request, t);
         f.completeExceptionally(t);
       }
     });
