@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one
  * or more contributor license agreements.  See the NOTICE file
  * distributed with this work for additional information
@@ -18,7 +18,6 @@
 package org.apache.ratis.protocol;
 
 import org.apache.ratis.proto.RaftProtos.CommitInfoProto;
-import org.apache.ratis.proto.RaftProtos.RaftClientRequestProto;
 import org.apache.ratis.util.JavaUtils;
 import org.apache.ratis.util.Preconditions;
 import org.apache.ratis.util.ProtoUtils;
@@ -44,13 +43,23 @@ public class RaftClientReply extends RaftClientMessage {
 
   /**
    * This field is the log index of the transaction
-   * if (1) the request is {@link RaftClientRequestProto.TypeCase#WRITE} and (2) the reply is success.
+   * if (1) the request is {@link org.apache.ratis.proto.RaftProtos.RaftClientRequestProto.TypeCase#WRITE} and (2) the
+   * reply is success.
    * Otherwise, this field is not used.
    */
   private final long logIndex;
   /** The commit information when the reply is created. */
   private final Collection<CommitInfoProto> commitInfos;
 
+  @SuppressWarnings("parameternumber")
+  public RaftClientReply(ClientId clientId, RaftGroupMemberId serverId,
+      long callId, boolean success, Message message, RaftException exception,
+      long logIndex, Collection<CommitInfoProto> commitInfos) {
+    this(clientId, serverId.getPeerId(), serverId.getGroupId(),
+        callId, success, message, exception, logIndex, commitInfos);
+  }
+
+  @SuppressWarnings("parameternumber")
   public RaftClientReply(
       ClientId clientId, RaftPeerId serverId, RaftGroupId groupId,
       long callId, boolean success, Message message, RaftException exception,
@@ -67,8 +76,10 @@ public class RaftClientReply extends RaftClientMessage {
       Preconditions.assertTrue(!success,
           () -> "Inconsistent parameters: success && exception != null: " + this);
       Preconditions.assertTrue(ReflectionUtils.isInstance(exception,
-          NotLeaderException.class, NotReplicatedException.class, StateMachineException.class,
-          RaftRetryFailureException.class), () -> "Unexpected exception class: " + this);
+          AlreadyClosedException.class,
+          NotLeaderException.class, NotReplicatedException.class,
+          LeaderNotReadyException.class, StateMachineException.class),
+          () -> "Unexpected exception class: " + this);
     }
   }
 
@@ -130,6 +141,11 @@ public class RaftClientReply extends RaftClientMessage {
     return message;
   }
 
+  /** If this reply has {@link AlreadyClosedException}, return it; otherwise return null. */
+  public AlreadyClosedException getAlreadyClosedException() {
+    return JavaUtils.cast(exception, AlreadyClosedException.class);
+  }
+
   /** If this reply has {@link NotLeaderException}, return it; otherwise return null. */
   public NotLeaderException getNotLeaderException() {
     return JavaUtils.cast(exception, NotLeaderException.class);
@@ -145,8 +161,12 @@ public class RaftClientReply extends RaftClientMessage {
     return JavaUtils.cast(exception, StateMachineException.class);
   }
 
-  /** If this reply has {@link RaftRetryFailureException}, return it; otherwise return null. */
-  public RaftRetryFailureException getRetryFailureException() {
-    return JavaUtils.cast(exception, RaftRetryFailureException.class);
+  public LeaderNotReadyException getLeaderNotReadyException() {
+    return JavaUtils.cast(exception, LeaderNotReadyException.class);
+  }
+
+  /** @return the exception, if there is any; otherwise, return null. */
+  public RaftException getException() {
+    return exception;
   }
 }
