@@ -30,6 +30,8 @@ import static org.apache.ratis.proto.RaftProtos.RaftClientRequestProto.TypeCase.
  */
 public class RaftClientRequest extends RaftClientMessage {
   private static final Type WRITE_DEFAULT = new Type(WriteRequestTypeProto.getDefaultInstance());
+  private static final Type WATCH_DEFAULT = new Type(
+      WatchRequestTypeProto.newBuilder().setIndex(0L).setReplication(ReplicationLevel.MAJORITY).build());
 
   private static final Type DEFAULT_READ = new Type(ReadRequestTypeProto.getDefaultInstance());
   private static final Type DEFAULT_STALE_READ = new Type(StaleReadRequestTypeProto.getDefaultInstance());
@@ -47,12 +49,15 @@ public class RaftClientRequest extends RaftClientMessage {
         : new Type(StaleReadRequestTypeProto.newBuilder().setMinIndex(minIndex).build());
   }
 
+  public static Type watchRequestType() {
+    return WATCH_DEFAULT;
+  }
   public static Type watchRequestType(long index, ReplicationLevel replication) {
     return new Type(WatchRequestTypeProto.newBuilder().setIndex(index).setReplication(replication).build());
   }
 
   /** The type of a request (oneof write, read, staleRead, watch; see the message RaftClientRequestProto). */
-  public static class Type {
+  public static final class Type {
     public static Type valueOf(WriteRequestTypeProto write) {
       return WRITE_DEFAULT;
     }
@@ -99,8 +104,8 @@ public class RaftClientRequest extends RaftClientMessage {
       this(WATCH, watch);
     }
 
-    public boolean is(RaftClientRequestProto.TypeCase typeCase) {
-      return getTypeCase().equals(typeCase);
+    public boolean is(RaftClientRequestProto.TypeCase tCase) {
+      return getTypeCase().equals(tCase);
     }
 
     public RaftClientRequestProto.TypeCase getTypeCase() {
@@ -127,7 +132,7 @@ public class RaftClientRequest extends RaftClientMessage {
       return (WatchRequestTypeProto)proto;
     }
 
-    static String toString(ReplicationLevel replication) {
+    public static String toString(ReplicationLevel replication) {
       return replication == ReplicationLevel.MAJORITY? "": "-" + replication;
     }
 
@@ -158,7 +163,7 @@ public class RaftClientRequest extends RaftClientMessage {
 
   private final SlidingWindowEntry slidingWindowEntry;
 
-  RaftClientRequest(ClientId clientId, RaftPeerId serverId, RaftGroupId groupId, long callId, Type type) {
+  public RaftClientRequest(ClientId clientId, RaftPeerId serverId, RaftGroupId groupId, long callId, Type type) {
     this(clientId, serverId, groupId, callId, null, type, null);
   }
 
