@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one
  * or more contributor license agreements.  See the NOTICE file
  * distributed with this work for additional information
@@ -17,12 +17,14 @@
  */
 package org.apache.ratis.protocol;
 
+import org.apache.ratis.retry.RetryPolicy;
+
 /**
- * Retry failure as per the retryPolicy defined.
+ * Retry failure as per the {@link RetryPolicy} defined.
  */
 public class RaftRetryFailureException extends RaftException {
-
-  public RaftRetryFailureException(String msg) {
-    super(msg);
+  public RaftRetryFailureException(
+      RaftClientRequest request, int attemptCount, RetryPolicy retryPolicy, Throwable cause) {
+    super("Failed " + request + " for " + attemptCount + " attempts with " + retryPolicy, cause);
   }
 }
\ No newline at end of file
