@@ -17,34 +17,32 @@
  */
 package org.apache.ratis.retry;
 
+import org.apache.ratis.util.JavaUtils;
+import org.apache.ratis.util.Preconditions;
 import org.apache.ratis.util.TimeDuration;
 
+import java.util.function.Supplier;
+
 /**
  * A collection of {@link RetryPolicy} implementations
  */
 public interface RetryPolicies {
-  /**
-   * Keep retrying forever with zero sleep.
-   */
+  /** For any requests, keep retrying forever with no sleep between attempts. */
   static RetryPolicy retryForeverNoSleep() {
     return Constants.RETRY_FOREVER_NO_SLEEP;
   }
 
+  /** No retry. */
   static RetryPolicy noRetry() {
     return Constants.NO_RETRY;
   }
 
-  /**
-   * Keep retrying forever with fixed sleep.
-   */
-  static RetryPolicy retryForeverWithSleep(TimeDuration sleepTime) {
+  /** For any requests, keep retrying forever with a fixed sleep time between attempts. */
+  static RetryForeverWithSleep retryForeverWithSleep(TimeDuration sleepTime) {
     return new RetryForeverWithSleep(sleepTime);
   }
 
-  /**
-   * Keep trying a limited number of times, waiting a fixed time between attempts,
-   * and then fail by re-throwing the exception.
-   */
+  /** For any requests, keep retrying a limited number of attempts with a fixed sleep time between attempts. */
   static RetryLimited retryUpToMaximumCountWithFixedSleep(int maxAttempts, TimeDuration sleepTime) {
     return new RetryLimited(maxAttempts, sleepTime);
   }
@@ -58,8 +56,8 @@ public interface RetryPolicies {
     private RetryForeverNoSleep() {}
 
     @Override
-    public boolean shouldRetry(int attemptCount) {
-      return true;
+    public Action handleAttemptFailure(Event event) {
+      return RetryPolicy.RETRY_WITHOUT_SLEEP_ACTION;
     }
 
     @Override
@@ -72,8 +70,8 @@ public interface RetryPolicies {
     private NoRetry() {}
 
     @Override
-    public boolean shouldRetry(int attemptCount) {
-      return false;
+    public Action handleAttemptFailure(Event event) {
+      return RetryPolicy.NO_RETRY_ACTION;
     }
 
     @Override
@@ -82,67 +80,41 @@ public interface RetryPolicies {
     }
   }
 
+  /** For any requests, keep retrying forever with a fixed sleep time between attempts. */
   class RetryForeverWithSleep implements RetryPolicy {
     private final TimeDuration sleepTime;
 
-    private String myString;
-
-    RetryForeverWithSleep(TimeDuration sleepTime) {
-      if (sleepTime.isNegative()) {
-        throw new IllegalArgumentException(
-            "sleepTime = " + sleepTime.getDuration() + " < 0");
-      }
+    private RetryForeverWithSleep(TimeDuration sleepTime) {
+      Preconditions.assertTrue(!sleepTime.isNegative(), () -> "sleepTime = " + sleepTime + " < 0");
       this.sleepTime = sleepTime;
     }
 
     @Override
-    public TimeDuration getSleepTime() {
-      return sleepTime;
-    }
-
-    @Override
-    public boolean shouldRetry(int attemptCount) {
-      return true;
+    public Action handleAttemptFailure(Event event) {
+      return () -> sleepTime;
     }
 
     @Override
     public String toString() {
-      if (myString == null) {
-        myString = getClass().getSimpleName() + "(sleepTime = " + sleepTime + ")";
-      }
-      return myString;
+      return getClass().getSimpleName() + "(sleepTime = " + sleepTime + ")";
     }
   }
-  /**
-   * Retry up to maxAttempts.
-   * The actual sleep time of the n-th retry is f(n, sleepTime),
-   * where f is a function provided by the subclass implementation.
-   *
-   * The object of the subclasses should be immutable;
-   * otherwise, the subclass must override hashCode(), equals(..) and toString().
-   */
-  class RetryLimited implements RetryPolicy {
+
+  /** For any requests, keep retrying a limited number of attempts with a fixed sleep time between attempts. */
+  class RetryLimited extends RetryForeverWithSleep  {
     private final int maxAttempts;
-    private final TimeDuration sleepTime;
+    private final Supplier<String> myString;
 
-    private String myString;
+    private RetryLimited(int maxAttempts, TimeDuration sleepTime) {
+      super(sleepTime);
 
-    RetryLimited(int maxAttempts, TimeDuration sleepTime) {
       if (maxAttempts < 0) {
         throw new IllegalArgumentException("maxAttempts = " + maxAttempts+" < 0");
       }
-      if (sleepTime.isNegative()) {
-        throw new IllegalArgumentException(
-            "sleepTime = " + sleepTime.getDuration() + " < 0");
-      }
 
       this.maxAttempts = maxAttempts;
-      this.sleepTime = sleepTime;
-    }
-
-    @Override
-    public TimeDuration getSleepTime() {
-      return sleepTime;
+      this.myString = JavaUtils.memoize(() -> getClass().getSimpleName()
+          + "(maxAttempts=" + maxAttempts + ", sleepTime=" + sleepTime + ")");
     }
 
     public int getMaxAttempts() {
@@ -150,17 +122,13 @@ public interface RetryPolicies {
     }
 
     @Override
-    public boolean shouldRetry(int attemptCount) {
-      return attemptCount <= maxAttempts;
+    public Action handleAttemptFailure(Event event) {
+      return event.getAttemptCount() < maxAttempts? super.handleAttemptFailure(event): NO_RETRY_ACTION;
     }
 
     @Override
     public String toString() {
-      if (myString == null) {
-        myString = getClass().getSimpleName() + "(maxAttempts=" + maxAttempts
-            + ", sleepTime=" + sleepTime + ")";
-      }
-      return myString;
+      return myString.get();
     }
   }
 }
