@@ -18,6 +18,7 @@
 package org.apache.ratis.util;
 
 import java.util.*;
+import java.util.concurrent.ConcurrentMap;
 import java.util.concurrent.ThreadLocalRandom;
 import java.util.function.Function;
 import java.util.function.Supplier;
@@ -70,7 +71,7 @@ public interface CollectionUtils {
   static <INPUT, OUTPUT> Iterable<OUTPUT> as(
       Iterable<INPUT> iteration, Function<INPUT, OUTPUT> converter) {
     return () -> new Iterator<OUTPUT>() {
-      final Iterator<INPUT> i = iteration.iterator();
+      private final Iterator<INPUT> i = iteration.iterator();
       @Override
       public boolean hasNext() {
         return i.hasNext();
@@ -106,4 +107,24 @@ public interface CollectionUtils {
     Preconditions.assertTrue(removed,
         () -> "Entry not found for key " + key + " in map " + name.get());
   }
+
+  static <K, V> V computeIfAbsent(ConcurrentMap<K, V> map, K key, Supplier<V> supplier,
+      Runnable actionIfAbsent) {
+    V v = map.get(key);
+    if (v != null) {
+      return v;
+    }
+    V newValue = supplier.get();
+    v = map.putIfAbsent(key, newValue);
+    if (v != null) {
+      return v;
+    }
+    actionIfAbsent.run();
+    return newValue;
+  }
+
+  static <K, V> V computeIfAbsent(ConcurrentMap<K, V> map, K key, Supplier<V> supplier) {
+    return computeIfAbsent(map, key, supplier, () -> {
+    });
+  }
 }
