@@ -139,10 +139,18 @@ public interface JavaUtils {
   }
 
   /** Attempt to get a return value from the given supplier multiple times. */
-  static <RETURN, THROWABLE extends Throwable> RETURN attempt(
+  static <RETURN, THROWABLE extends Throwable> RETURN attemptRepeatedly(
       CheckedSupplier<RETURN, THROWABLE> supplier,
       int numAttempts, TimeDuration sleepTime, String name, Logger log)
       throws THROWABLE, InterruptedException {
+    return attempt(supplier, numAttempts, sleepTime, () -> name, log);
+  }
+
+  /** Attempt to get a return value from the given supplier multiple times. */
+  static <RETURN, THROWABLE extends Throwable> RETURN attempt(
+      CheckedSupplier<RETURN, THROWABLE> supplier,
+      int numAttempts, TimeDuration sleepTime, Supplier<?> name, Logger log)
+      throws THROWABLE, InterruptedException {
     Objects.requireNonNull(supplier, "supplier == null");
     Preconditions.assertTrue(numAttempts > 0, () -> "numAttempts = " + numAttempts + " <= 0");
     Preconditions.assertTrue(!sleepTime.isNegative(), () -> "sleepTime = " + sleepTime + " < 0");
@@ -155,7 +163,7 @@ public interface JavaUtils {
           throw t;
         }
         if (log != null && log.isWarnEnabled()) {
-          log.warn("FAILED \"" + name + "\", attempt #" + i + "/" + numAttempts
+          log.warn("FAILED \"" + name.get() + "\", attempt #" + i + "/" + numAttempts
               + ": " + t + ", sleep " + sleepTime + " and then retry.", t);
         }
       }
@@ -165,46 +173,26 @@ public interface JavaUtils {
     throw new IllegalStateException("BUG: this line should be unreachable.");
   }
 
-  /** @deprecated use {@link #attempt(CheckedRunnable, int, TimeDuration, String, Logger)} */
-  @Deprecated
-  static <THROWABLE extends Throwable> void attempt(
-      CheckedRunnable<THROWABLE> op, int numAttempts, long sleepMs, String name, Logger log)
-      throws THROWABLE, InterruptedException {
-    attempt(op, numAttempts, TimeDuration.valueOf(sleepMs, TimeUnit.MILLISECONDS), name, log);
-  }
-
   /** Attempt to run the given op multiple times. */
   static <THROWABLE extends Throwable> void attempt(
       CheckedRunnable<THROWABLE> runnable, int numAttempts, TimeDuration sleepTime, String name, Logger log)
       throws THROWABLE, InterruptedException {
-    attempt(CheckedRunnable.asCheckedSupplier(runnable), numAttempts, sleepTime, name, log);
+    attemptRepeatedly(CheckedRunnable.asCheckedSupplier(runnable), numAttempts, sleepTime, name, log);
   }
 
   /** Attempt to wait the given condition to return true multiple times. */
-  static void attempt(
+  static void attemptUntilTrue(
       BooleanSupplier condition, int numAttempts, TimeDuration sleepTime, String name, Logger log)
       throws InterruptedException {
     Objects.requireNonNull(condition, "condition == null");
-    Preconditions.assertTrue(numAttempts > 0, () -> "numAttempts = " + numAttempts + " <= 0");
-    Preconditions.assertTrue(!sleepTime.isNegative(), () -> "sleepTime = " + sleepTime + " < 0");
-
-    for(int i = 1; i <= numAttempts; i++) {
-      if (condition.getAsBoolean()) {
-        return;
+    attempt(() -> {
+      if (!condition.getAsBoolean()) {
+        throw new IllegalStateException("Condition " + name + " is false.");
       }
-      if (log != null && log.isWarnEnabled()) {
-        log.warn("FAILED " + name + " attempt #" + i + "/" + numAttempts
-            + ": sleep " + sleepTime + " and then retry.");
-      }
-
-      sleepTime.sleep();
-    }
-
-    if (!condition.getAsBoolean()) {
-      throw new IllegalStateException("Failed " + name + " for " + numAttempts + " attempts.");
-    }
+    }, numAttempts, sleepTime, name, log);
   }
 
+
   static Timer runRepeatedly(Runnable runnable, long delay, long period, TimeUnit unit) {
     final Timer timer = new Timer(true);
     timer.schedule(new TimerTask() {
@@ -249,4 +237,17 @@ public interface JavaUtils {
       throw new CompletionException(t);
     }
   }
+
+  static boolean sleep(long sleepMs, long thresholdMs) throws InterruptedException {
+    final Timestamp t = Timestamp.currentTime();
+    Thread.sleep(sleepMs);
+    final long elapsedMs = t.elapsedTimeMs();
+    if (elapsedMs - sleepMs > thresholdMs) {
+      LOG.warn("Unexpected long sleep: sleep({}ms) actually took {}ms which is over the threshold {}ms",
+          sleepMs, elapsedMs, thresholdMs);
+      return false;
+    }
+    return true;
+  }
+
 }
