@@ -19,14 +19,15 @@ package org.apache.ratis.util;
 
 import org.apache.ratis.proto.RaftProtos.CommitInfoProto;
 import org.apache.ratis.proto.RaftProtos.RaftGroupIdProto;
+import org.apache.ratis.proto.RaftProtos.RaftGroupMemberIdProto;
 import org.apache.ratis.proto.RaftProtos.RaftGroupProto;
 import org.apache.ratis.proto.RaftProtos.RaftPeerProto;
 import org.apache.ratis.proto.RaftProtos.RaftRpcReplyProto;
 import org.apache.ratis.proto.RaftProtos.RaftRpcRequestProto;
-import org.apache.ratis.proto.RaftProtos.RequestVoteReplyProto;
 import org.apache.ratis.proto.RaftProtos.SlidingWindowEntry;
 import org.apache.ratis.protocol.RaftGroup;
 import org.apache.ratis.protocol.RaftGroupId;
+import org.apache.ratis.protocol.RaftGroupMemberId;
 import org.apache.ratis.protocol.RaftPeer;
 import org.apache.ratis.protocol.RaftPeerId;
 import org.apache.ratis.thirdparty.com.google.protobuf.ByteString;
@@ -43,7 +44,7 @@ import java.util.stream.Collectors;
 public interface ProtoUtils {
   static ByteString writeObject2ByteString(Object obj) {
     final ByteString.Output byteOut = ByteString.newOutput();
-    try(final ObjectOutputStream objOut = new ObjectOutputStream(byteOut)) {
+    try(ObjectOutputStream objOut = new ObjectOutputStream(byteOut)) {
       objOut.writeObject(obj);
     } catch (IOException e) {
       throw new IllegalStateException(
@@ -74,18 +75,14 @@ public interface ProtoUtils {
     return new RaftPeer(RaftPeerId.valueOf(p.getId()), p.getAddress());
   }
 
-  static RaftPeer[] toRaftPeerArray(List<RaftPeerProto> protos) {
-    final RaftPeer[] peers = new RaftPeer[protos.size()];
-    for (int i = 0; i < peers.length; i++) {
-      peers[i] = toRaftPeer(protos.get(i));
-    }
-    return peers;
+  static List<RaftPeer> toRaftPeers(List<RaftPeerProto> protos) {
+    return protos.stream().map(ProtoUtils::toRaftPeer).collect(Collectors.toList());
   }
 
   static Iterable<RaftPeerProto> toRaftPeerProtos(
       final Collection<RaftPeer> peers) {
     return () -> new Iterator<RaftPeerProto>() {
-      final Iterator<RaftPeer> i = peers.iterator();
+      private final Iterator<RaftPeer> i = peers.iterator();
 
       @Override
       public boolean hasNext() {
@@ -108,7 +105,7 @@ public interface ProtoUtils {
   }
 
   static RaftGroup toRaftGroup(RaftGroupProto proto) {
-    return RaftGroup.valueOf(toRaftGroupId(proto.getGroupId()), toRaftPeerArray(proto.getPeersList()));
+    return RaftGroup.valueOf(toRaftGroupId(proto.getGroupId()), toRaftPeers(proto.getPeersList()));
   }
 
   static RaftGroupProto.Builder toRaftGroupProtoBuilder(RaftGroup group) {
@@ -117,6 +114,20 @@ public interface ProtoUtils {
         .addAllPeers(toRaftPeerProtos(group.getPeers()));
   }
 
+  static RaftGroupMemberId toRaftGroupMemberId(ByteString peerId, RaftGroupIdProto groupId) {
+    return RaftGroupMemberId.valueOf(RaftPeerId.valueOf(peerId), ProtoUtils.toRaftGroupId(groupId));
+  }
+
+  static RaftGroupMemberId toRaftGroupMemberId(RaftGroupMemberIdProto memberId) {
+    return toRaftGroupMemberId(memberId.getPeerId(), memberId.getGroupId());
+  }
+
+  static RaftGroupMemberIdProto.Builder toRaftGroupMemberIdProtoBuilder(RaftGroupMemberId memberId) {
+    return RaftGroupMemberIdProto.newBuilder()
+        .setPeerId(memberId.getPeerId().toByteString())
+        .setGroupId(toRaftGroupIdProtoBuilder(memberId.getGroupId()));
+  }
+
   static CommitInfoProto toCommitInfoProto(RaftPeer peer, long commitIndex) {
     return CommitInfoProto.newBuilder()
         .setServer(peer.getRaftPeerProto())
@@ -164,7 +175,4 @@ public interface ProtoUtils {
         + "#" + proto.getCallId() + ":"
         + (proto.getSuccess()? "OK": "FAIL");
   }
-  static String toString(RequestVoteReplyProto proto) {
-    return toString(proto.getServerReply()) + "-t" + proto.getTerm();
-  }
 }
