@@ -25,6 +25,7 @@ import java.io.Closeable;
 import java.util.Iterator;
 import java.util.SortedMap;
 import java.util.TreeMap;
+import java.util.concurrent.ConcurrentSkipListMap;
 import java.util.concurrent.TimeUnit;
 import java.util.function.Consumer;
 import java.util.function.LongFunction;
@@ -37,18 +38,22 @@ import java.util.function.LongFunction;
 public interface SlidingWindow {
   Logger LOG = LoggerFactory.getLogger(SlidingWindow.class);
 
+  static String getName(Class<?> clazz, Object name) {
+    return SlidingWindow.class.getSimpleName() +  "$" + clazz.getSimpleName() + ":" + name;
+  }
+
   interface Request<REPLY> {
     long getSeqNum();
 
     void setReply(REPLY reply);
 
     boolean hasReply();
+
+    void fail(Throwable e);
   }
 
   interface ClientSideRequest<REPLY> extends Request<REPLY> {
     void setFirstRequest();
-
-    void fail(Throwable e);
   }
 
   interface ServerSideRequest<REPLY> extends Request<REPLY> {
@@ -57,15 +62,15 @@ public interface SlidingWindow {
 
   /** A seqNum-to-request map, sorted by seqNum. */
   class RequestMap<REQUEST extends Request<REPLY>, REPLY> implements Iterable<REQUEST> {
-    static boolean LOG_REPEATEDLY = false;
+    private static boolean logRepeatedly = false;
     private final Object name;
     /** Request map: seqNum -> request */
-    private final SortedMap<Long, REQUEST> requests = new TreeMap<>();
+    private final SortedMap<Long, REQUEST> requests = new ConcurrentSkipListMap<>();
 
     RequestMap(Object name) {
       this.name = name;
-      if (LOG_REPEATEDLY && LOG.isDebugEnabled()) {
-        JavaUtils.runRepeatedly(() -> log(), 5, 10, TimeUnit.SECONDS);
+      if (logRepeatedly && LOG.isDebugEnabled()) {
+        JavaUtils.runRepeatedly(this::log, 5, 10, TimeUnit.SECONDS);
       }
     }
 
@@ -77,6 +82,10 @@ public interface SlidingWindow {
       return requests.isEmpty();
     }
 
+    private REQUEST get(long seqNum) {
+      return requests.get(seqNum);
+    }
+
     /**
      * If the request with the given seqNum is non-replied, return it.
      * Otherwise, return null.
@@ -86,7 +95,7 @@ public interface SlidingWindow {
      * (2) it does not has reply.
      */
     REQUEST getNonRepliedRequest(long seqNum, String op) {
-      final REQUEST request = requests.get(seqNum);
+      final REQUEST request = get(seqNum);
       if (request == null) {
         LOG.debug("{}: {}, seq={} not found in {}", getName(), op, seqNum, this);
         return null;
@@ -123,8 +132,8 @@ public interface SlidingWindow {
      *
      * @return true iff this method does set the reply for the request.
      */
-    boolean setReply(long seqNum, REPLY reply, String op) {
-      final REQUEST request = getNonRepliedRequest(seqNum, op);
+    boolean setReply(long seqNum, REPLY reply) {
+      final REQUEST request = getNonRepliedRequest(seqNum, "setReply");
       if (request == null) {
         LOG.debug("{}: DUPLICATED reply {} for seq={} in {}", getName(), reply, seqNum, this);
         return false;
@@ -135,12 +144,30 @@ public interface SlidingWindow {
       return true;
     }
 
-    synchronized void clear() {
+    void endOfRequests(long nextToProcess, REQUEST end, Consumer<REQUEST> replyMethod) {
+      final REQUEST nextToProcessRequest = requests.get(nextToProcess);
+      Preconditions.assertNull(nextToProcessRequest,
+          () -> "nextToProcessRequest = " + nextToProcessRequest + " != null, nextToProcess = " + nextToProcess);
+
+      final SortedMap<Long, REQUEST> tail = requests.tailMap(nextToProcess);
+      for (REQUEST r : tail.values()) {
+        final AlreadyClosedException e = new AlreadyClosedException(
+            getName() + " is closing: seq = " + r.getSeqNum() + " > nextToProcess = " + nextToProcess
+                + " will NEVER be processed; request = " + r);
+        r.fail(e);
+        replyMethod.accept(r);
+      }
+      tail.clear();
+
+      putNewRequest(end);
+    }
+
+    void clear() {
       LOG.debug("close {}", this);
       requests.clear();
     }
 
-    synchronized void log() {
+    void log() {
       LOG.debug(this.toString());
       for(REQUEST r : requests.values()) {
         LOG.debug("  {}: hasReply? {}", r.getSeqNum(), r.hasReply());
@@ -184,7 +211,7 @@ public interface SlidingWindow {
     private Throwable exception;
 
     public Client(Object name) {
-      this.requests = new RequestMap<REQUEST, REPLY>(name) {
+      this.requests = new RequestMap<REQUEST, REPLY>(getName(getClass(), name)) {
         @Override
         synchronized void log() {
           LOG.debug(toString());
@@ -284,7 +311,7 @@ public interface SlidingWindow {
      */
     public synchronized void receiveReply(
         long seqNum, REPLY reply, Consumer<REQUEST> sendMethod) {
-      if (!requests.setReply(seqNum, reply, "receiveReply")) {
+      if (!requests.setReply(seqNum, reply)) {
         return; // request already replied
       }
       if (seqNum == firstSeqNum) {
@@ -346,8 +373,11 @@ public interface SlidingWindow {
     }
 
     private void alreadyClosed(REQUEST request, Throwable e) {
-      request.fail(new AlreadyClosedException(SlidingWindow.class.getSimpleName() + "$" + getClass().getSimpleName()
-          + " " + requests.getName() + " is closed.", e));
+      request.fail(new AlreadyClosedException(requests.getName() + " is closed.", e));
+    }
+
+    public boolean isFirst(long seqNum) {
+      return seqNum == (firstSeqNum != -1 ? firstSeqNum : requests.firstSeqNum());
     }
   }
 
@@ -368,7 +398,7 @@ public interface SlidingWindow {
     private long nextToProcess = -1;
 
     public Server(Object name, REQUEST end) {
-      this.requests = new RequestMap<>(name);
+      this.requests = new RequestMap<>(getName(getClass(), name));
       this.end = end;
       Preconditions.assertTrue(end.getSeqNum() == Long.MAX_VALUE);
     }
@@ -393,30 +423,27 @@ public interface SlidingWindow {
 
     private void processRequestsFromHead(Consumer<REQUEST> processingMethod) {
       for(REQUEST r : requests) {
-        if (r.getSeqNum() != nextToProcess) {
+        if (r.getSeqNum() > nextToProcess) {
           return;
+        } else if (r.getSeqNum() == nextToProcess) {
+          processingMethod.accept(r);
+          nextToProcess++;
         }
-        processingMethod.accept(r);
-        nextToProcess++;
       }
     }
 
     /**
      * Receives a reply for the given seqNum (may out-of-order) from the processor.
-     * It may trigger sending replies to client or processing more requests.
+     * It may trigger sending replies to client.
      */
-    public synchronized void receiveReply(
-        long seqNum, REPLY reply, Consumer<REQUEST> replyMethod, Consumer<REQUEST> processingMethod) {
-      if (!requests.setReply(seqNum, reply, "receiveReply")) {
+    public synchronized void receiveReply(long seqNum, REPLY reply, Consumer<REQUEST> replyMethod) {
+      if (!requests.setReply(seqNum, reply)) {
         return; // request already replied
       }
       sendRepliesFromHead(replyMethod);
-      processRequestsFromHead(processingMethod);
     }
 
-    private void sendRepliesFromHead(
-        Consumer<REQUEST> replyMethod
-    ) {
+    private void sendRepliesFromHead(Consumer<REQUEST> replyMethod) {
       for(final Iterator<REQUEST> i = requests.iterator(); i.hasNext(); i.remove()) {
         final REQUEST r = i.next();
         if (!r.hasReply()) {
@@ -433,14 +460,14 @@ public interface SlidingWindow {
      * Signal the end of requests.
      * @return true if no more outstanding requests.
      */
-    public synchronized boolean endOfRequests() {
+    public synchronized boolean endOfRequests(Consumer<REQUEST> replyMethod) {
       if (requests.isEmpty()) {
         return true;
-      } else {
-        LOG.debug("{}: put end-of-request in {}", requests.getName(), this);
-        requests.putNewRequest(end);
-        return false;
       }
+
+      LOG.debug("{}: put end-of-request in {}", requests.getName(), this);
+      requests.endOfRequests(nextToProcess, end, replyMethod);
+      return false;
     }
 
     @Override
