@@ -21,21 +21,78 @@ import org.apache.ratis.util.function.CheckedRunnable;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
-import java.util.concurrent.Executors;
-import java.util.concurrent.ScheduledExecutorService;
+import java.io.Closeable;
+import java.util.Collection;
+import java.util.Optional;
+import java.util.concurrent.ScheduledFuture;
 import java.util.concurrent.ScheduledThreadPoolExecutor;
+import java.util.concurrent.ThreadFactory;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicReference;
 import java.util.function.Consumer;
 import java.util.function.Supplier;
 
-public final class TimeoutScheduler {
+public final class TimeoutScheduler implements Closeable {
   public static final Logger LOG = LoggerFactory.getLogger(TimeoutScheduler.class);
 
-  private static final TimeDuration DEFAULT_GRACE_PERIOD = TimeDuration.valueOf(1, TimeUnit.MINUTES);
+  static final TimeDuration DEFAULT_GRACE_PERIOD = TimeDuration.valueOf(1, TimeUnit.MINUTES);
 
-  public static TimeoutScheduler newInstance(int numThreads) {
-    return new TimeoutScheduler(numThreads);
+  private static final Supplier<TimeoutScheduler> INSTANCE = JavaUtils.memoize(TimeoutScheduler::new);
+
+  public static TimeoutScheduler getInstance() {
+    return INSTANCE.get();
+  }
+
+  static TimeoutScheduler newInstance() {
+    return new TimeoutScheduler();
+  }
+
+  static class ShutdownTask {
+    private final int sid;
+    private final ScheduledFuture<?> future;
+
+    ShutdownTask(int sid, ScheduledFuture<?> future) {
+      this.sid = sid;
+      this.future = future;
+    }
+
+    int getSid() {
+      return sid;
+    }
+
+    void cancel() {
+      future.cancel(false);
+    }
+  }
+
+  private static class Scheduler {
+    private final AtomicReference<ScheduledThreadPoolExecutor> executor = new AtomicReference<>();
+
+    boolean hasExecutor() {
+      return executor.get() != null;
+    }
+
+    int getQueueSize() {
+      return Optional.ofNullable(executor.get())
+          .map(ScheduledThreadPoolExecutor::getQueue)
+          .map(Collection::size).orElse(0);
+    }
+
+    ScheduledFuture<?> schedule(Runnable task, Supplier<String> name, TimeDuration time) {
+      return executor.updateAndGet(e -> Optional.ofNullable(e).orElseGet(Scheduler::newExecutor))
+          .schedule(LogUtils.newRunnable(LOG, task, name), time.getDuration(), time.getUnit());
+    }
+
+    private static ScheduledThreadPoolExecutor newExecutor() {
+      LOG.debug("new ScheduledThreadPoolExecutor");
+      final ScheduledThreadPoolExecutor e = new ScheduledThreadPoolExecutor(0, (ThreadFactory) Daemon::new);
+      e.setRemoveOnCancelPolicy(true);
+      return e;
+    }
+
+    void shutdown() {
+      Optional.ofNullable(executor.getAndSet(null)).ifPresent(ScheduledThreadPoolExecutor::shutdown);
+    }
   }
 
   /** When there is no tasks, the time period to wait before shutting down the scheduler. */
@@ -46,16 +103,15 @@ public final class TimeoutScheduler {
   /** The scheduleID for each task */
   private int scheduleID = 0;
 
-  private final int numThreads;
-  private volatile ScheduledExecutorService scheduler = null;
+  private ShutdownTask shutdownTask = null;
 
-  private TimeoutScheduler(int numThreads) {
-    this.numThreads = numThreads;
+  private final Scheduler scheduler = new Scheduler();
+
+  private TimeoutScheduler() {
   }
 
-  public int getNumThreads() {
-    final ScheduledExecutorService s = scheduler;
-    return s instanceof ScheduledThreadPoolExecutor? ((ScheduledThreadPoolExecutor)s).getCorePoolSize(): numThreads;
+  int getQueueSize() {
+    return scheduler.getQueueSize();
   }
 
   TimeDuration getGracePeriod() {
@@ -66,8 +122,8 @@ public final class TimeoutScheduler {
     this.gracePeriod.set(gracePeriod);
   }
 
-  synchronized boolean hasScheduler() {
-    return scheduler != null;
+  boolean hasScheduler() {
+    return scheduler.hasExecutor();
   }
 
   /**
@@ -92,30 +148,32 @@ public final class TimeoutScheduler {
   }
 
   private synchronized void onTimeout(TimeDuration timeout, Consumer<Integer> toSchedule) {
-    if (scheduler == null) {
-      Preconditions.assertTrue(numTasks == 0);
-      LOG.debug("Initialize scheduler");
-      scheduler = Executors.newScheduledThreadPool(numThreads, Daemon::new);
-    }
     numTasks++;
     final int sid = scheduleID++;
 
     LOG.debug("schedule a task: timeout {}, sid {}", timeout, sid);
-    schedule(scheduler, () -> toSchedule.accept(sid), () -> "task #" + sid, timeout);
+    scheduler.schedule(() -> toSchedule.accept(sid), () -> "task #" + sid, timeout);
   }
 
   private synchronized void onTaskCompleted() {
-    if (--numTasks == 0) {
-      final int sid = scheduleID;
-      final TimeDuration grace = getGracePeriod();
-      LOG.debug("Schedule a shutdown task: grace {}, sid {}", grace, sid);
-      schedule(scheduler, () -> tryShutdownScheduler(sid), () -> "shutdown task #" + sid, grace);
+    if (--numTasks > 0) {
+      return;
+    }
+    final int sid = scheduleID;
+    if (shutdownTask != null) {
+      if (shutdownTask.getSid() == sid) {
+        // the shutdown task is still valid
+        return;
+      }
+      // the shutdown task is invalid
+      shutdownTask.cancel();
     }
-  }
 
-  static void schedule(ScheduledExecutorService service, Runnable task, Supplier<String> name, TimeDuration timeDuration) {
-    service.schedule(LogUtils.newRunnable(LOG, task, name),
-        timeDuration.getDuration(), timeDuration.getUnit());
+    final TimeDuration grace = getGracePeriod();
+    LOG.debug("Schedule a shutdown task: grace {}, sid {}", grace, sid);
+    final ScheduledFuture<?> future = scheduler.schedule(() -> tryShutdownScheduler(sid),
+        () -> "shutdown task #" + sid, grace);
+    shutdownTask = new ShutdownTask(sid, future);
   }
 
   private synchronized void tryShutdownScheduler(int sid) {
@@ -123,7 +181,6 @@ public final class TimeoutScheduler {
       // No new tasks submitted, shutdown the scheduler.
       LOG.debug("shutdown scheduler: sid {}", sid);
       scheduler.shutdown();
-      scheduler = null;
     } else {
       LOG.debug("shutdown cancelled: scheduleID has changed from {} to {}", sid, scheduleID);
     }
@@ -133,4 +190,9 @@ public final class TimeoutScheduler {
   public void onTimeout(TimeDuration timeout, CheckedRunnable<?> task, Logger log, Supplier<String> errorMessage) {
     onTimeout(timeout, task, t -> log.error(errorMessage.get(), t));
   }
+
+  @Override
+  public synchronized void close() {
+    tryShutdownScheduler(scheduleID);
+  }
 }
