@@ -22,7 +22,7 @@ import org.apache.ratis.conf.ConfUtils;
 import org.apache.ratis.util.ExitUtils;
 import org.apache.ratis.util.FileUtils;
 import org.apache.ratis.util.JavaUtils;
-import org.apache.ratis.util.LogUtils;
+import org.apache.ratis.util.Log4jUtils;
 import org.apache.ratis.util.TimeDuration;
 import org.apache.ratis.util.function.CheckedRunnable;
 import org.junit.After;
@@ -37,25 +37,41 @@ import java.io.File;
 import java.io.IOException;
 import java.util.Objects;
 import java.util.concurrent.CompletableFuture;
+import java.util.concurrent.Future;
 import java.util.concurrent.ThreadLocalRandom;
 import java.util.concurrent.TimeUnit;
+import java.util.concurrent.atomic.AtomicReference;
 import java.util.function.Supplier;
 
 public abstract class BaseTest {
   public final Logger LOG = LoggerFactory.getLogger(getClass());
 
   public static final TimeDuration HUNDRED_MILLIS = TimeDuration.valueOf(100, TimeUnit.MILLISECONDS);
-  public static final TimeDuration ONE_SECOND = TimeDuration.valueOf(1, TimeUnit.SECONDS);
+  public static final TimeDuration ONE_SECOND = TimeDuration.ONE_SECOND;
+  public static final TimeDuration FIVE_SECONDS = TimeDuration.valueOf(5, TimeUnit.SECONDS);
 
   {
-    LogUtils.setLogLevel(ConfUtils.LOG, Level.WARN);
-    LogUtils.setLogLevel(FileUtils.LOG, Level.TRACE);
+    Log4jUtils.setLogLevel(ConfUtils.LOG, Level.WARN);
+    Log4jUtils.setLogLevel(FileUtils.LOG, Level.TRACE);
 
     ExitUtils.disableSystemExit();
   }
 
+  private final AtomicReference<Throwable> firstException = new AtomicReference<>();
+
+  public void setFirstException(Throwable e) {
+    if (firstException.compareAndSet(null, e)) {
+      LOG.error("Set firstException", e);
+    }
+  }
+
   @After
-  public void assertNotTerminated() {
+  public void assertNoFailures() {
+    final Throwable e = firstException.get();
+    if (e != null) {
+      throw new IllegalStateException("Failed: first exception was set", e);
+    }
+
     ExitUtils.assertNotTerminated();
   }
 
@@ -117,6 +133,9 @@ public abstract class BaseTest {
       String description, CheckedRunnable<?> testCode,
       Class<? extends Throwable> expectedThrowableClass, Logger log,
       Class<? extends Throwable>... expectedCauseClasses) {
+    if (log != null) {
+      log.info("run '{}'", description);
+    }
     try {
       testCode.run();
     } catch (Throwable t) {
@@ -139,6 +158,9 @@ public abstract class BaseTest {
       String description, Supplier<CompletableFuture<?>> testCode,
       Class<? extends Throwable> expectedThrowableClass, Logger log,
       Class<? extends Throwable>... expectedCauseClasses) {
+    if (log != null) {
+      log.info("run '{}'", description);
+    }
     try {
       testCode.get().join();
     } catch (Throwable t) {
@@ -155,4 +177,8 @@ public abstract class BaseTest {
       Class<? extends Throwable>... expectedCauseClasses) {
     return testFailureCaseAsync(description, testCode, expectedThrowableClass, LOG, expectedCauseClasses);
   }
+
+  static <T> T getWithDefaultTimeout(Future<T> future) throws Exception {
+    return future.get(FIVE_SECONDS.getDuration(), FIVE_SECONDS.getUnit());
+  }
 }
