@@ -63,12 +63,11 @@ public abstract class Client extends SubCommandBase {
 
     RaftClientConfigKeys.Rpc.setRequestTimeout(raftProperties,
         TimeDuration.valueOf(50000, TimeUnit.MILLISECONDS));
-    RaftClientConfigKeys.Async.setSchedulerThreads(raftProperties, 10);
     RaftClientConfigKeys.Async.setMaxOutstandingRequests(raftProperties, 1000);
 
 
-    final RaftGroup raftGroup = RaftGroup.valueOf(RaftGroupId.valueOf(ByteString.copyFromUtf8(raftGroupId)),
-        parsePeers(peers));
+    final RaftGroup raftGroup = RaftGroup.valueOf(RaftGroupId.valueOf(ByteString.copyFromUtf8(getRaftGroupId())),
+            getPeers());
 
     RaftClient.Builder builder =
         RaftClient.newBuilder().setProperties(raftProperties);
