@@ -26,7 +26,10 @@ import java.util.List;
 /**
  * This class enumerates all the commands enqueued by FileStore state machine.
  */
-public class FileStore {
+public final class FileStore {
+  private FileStore() {
+  }
+
   public static List<SubCommandBase> getSubCommands() {
     List<SubCommandBase> commands = new ArrayList<>();
     commands.add(new Server());
