@@ -31,6 +31,7 @@ import org.apache.ratis.proto.RaftProtos.GroupManagementRequestProto;
 import org.apache.ratis.proto.RaftProtos.RaftClientReplyProto;
 import org.apache.ratis.proto.RaftProtos.RaftClientRequestProto;
 import org.apache.ratis.proto.RaftProtos.SetConfigurationRequestProto;
+import org.apache.ratis.protocol.LeaderNotReadyException;
 import org.apache.ratis.protocol.TimeoutIOException;
 import org.apache.ratis.thirdparty.io.grpc.netty.GrpcSslContexts;
 import org.apache.ratis.thirdparty.io.netty.handler.ssl.SslContextBuilder;
@@ -65,6 +66,7 @@ import java.util.Map;
 import java.util.Optional;
 import java.util.concurrent.CompletableFuture;
 import java.util.concurrent.ConcurrentHashMap;
+import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicReference;
 import java.util.function.Consumer;
 import java.util.function.Function;
@@ -78,7 +80,8 @@ public class GrpcClientProtocolClient implements Closeable {
   private final ManagedChannel channel;
 
   private final TimeDuration requestTimeoutDuration;
-  private final TimeoutScheduler scheduler = TimeoutScheduler.newInstance(3);
+  private final TimeDuration watchRequestTimeoutDuration;
+  private final TimeoutScheduler scheduler = TimeoutScheduler.getInstance();
 
   private final RaftClientProtocolServiceBlockingStub blockingStub;
   private final RaftClientProtocolServiceStub asyncStub;
@@ -98,15 +101,23 @@ public class GrpcClientProtocolClient implements Closeable {
 
     if (tlsConf!= null) {
       SslContextBuilder sslContextBuilder = GrpcSslContexts.forClient();
-      if (tlsConf.getTrustStore() != null) {
+      if (tlsConf.isFileBasedConfig()) {
+        sslContextBuilder.trustManager(tlsConf.getTrustStoreFile());
+      } else {
         sslContextBuilder.trustManager(tlsConf.getTrustStore());
       }
       if (tlsConf.getMtlsEnabled()) {
-        sslContextBuilder.keyManager(tlsConf.getCertChain(),
-            tlsConf.getPrivateKey());
+        if (tlsConf.isFileBasedConfig()) {
+          sslContextBuilder.keyManager(tlsConf.getCertChainFile(),
+              tlsConf.getPrivateKeyFile());
+        } else {
+          sslContextBuilder.keyManager(tlsConf.getPrivateKey(),
+              tlsConf.getCertChain());
+        }
       }
       try {
-        channelBuilder.useTransportSecurity().sslContext(sslContextBuilder.build());
+        channelBuilder.useTransportSecurity().sslContext(
+            sslContextBuilder.build());
       } catch (Exception ex) {
         throw new RuntimeException(ex);
       }
@@ -120,6 +131,8 @@ public class GrpcClientProtocolClient implements Closeable {
     asyncStub = RaftClientProtocolServiceGrpc.newStub(channel);
     adminBlockingStub = AdminProtocolServiceGrpc.newBlockingStub(channel);
     this.requestTimeoutDuration = RaftClientConfigKeys.Rpc.requestTimeout(properties);
+    this.watchRequestTimeoutDuration =
+        RaftClientConfigKeys.Rpc.watchRequestTimeout(properties);
   }
 
   String getName() {
@@ -130,7 +143,13 @@ public class GrpcClientProtocolClient implements Closeable {
   public void close() {
     Optional.ofNullable(orderedStreamObservers.getAndSet(null)).ifPresent(AsyncStreamObservers::close);
     Optional.ofNullable(unorderedStreamObservers.getAndSet(null)).ifPresent(AsyncStreamObservers::close);
-    channel.shutdownNow();
+    channel.shutdown();
+    try {
+      channel.awaitTermination(5, TimeUnit.SECONDS);
+    } catch (Exception e) {
+      LOG.error("Unexpected exception while waiting for channel termination", e);
+    }
+    scheduler.close();
   }
 
   RaftClientReplyProto groupAdd(GroupManagementRequestProto request) throws IOException {
@@ -180,12 +199,12 @@ public class GrpcClientProtocolClient implements Closeable {
 
   AsyncStreamObservers getOrderedStreamObservers() {
     return orderedStreamObservers.updateAndGet(
-        a -> a != null? a : new AsyncStreamObservers(orderedStreamObservers, this::ordered));
+        a -> a != null? a : new AsyncStreamObservers(this::ordered));
   }
 
   AsyncStreamObservers getUnorderedAsyncStreamObservers() {
     return unorderedStreamObservers.updateAndGet(
-        a -> a != null? a : new AsyncStreamObservers(unorderedStreamObservers, asyncStub::unordered));
+        a -> a != null? a : new AsyncStreamObservers(asyncStub::unordered));
   }
 
   public RaftPeer getTarget() {
@@ -251,12 +270,17 @@ public class GrpcClientProtocolClient implements Closeable {
         final long callId = proto.getRpcReply().getCallId();
         try {
           final RaftClientReply reply = ClientProtoUtils.toRaftClientReply(proto);
-          LOG.info("{}: receive {}", getName(), reply);
+          LOG.trace("{}: receive {}", getName(), reply);
           final NotLeaderException nle = reply.getNotLeaderException();
           if (nle != null) {
             completeReplyExceptionally(nle, NotLeaderException.class.getName());
             return;
           }
+          final LeaderNotReadyException lnre = reply.getLeaderNotReadyException();
+          if (lnre != null) {
+            completeReplyExceptionally(lnre, NotLeaderException.class.getName());
+            return;
+          }
           handleReplyFuture(callId, f -> f.complete(reply));
         } catch (Throwable t) {
           handleReplyFuture(callId, f -> f.completeExceptionally(t));
@@ -275,36 +299,41 @@ public class GrpcClientProtocolClient implements Closeable {
       }
     };
     private final RequestStreamer requestStreamer;
-    private final AtomicReference<AsyncStreamObservers> ref;
 
-    AsyncStreamObservers(AtomicReference<AsyncStreamObservers> ref,
-        Function<StreamObserver<RaftClientReplyProto>, StreamObserver<RaftClientRequestProto>> f) {
+    AsyncStreamObservers(Function<StreamObserver<RaftClientReplyProto>, StreamObserver<RaftClientRequestProto>> f) {
       this.requestStreamer = new RequestStreamer(f.apply(replyStreamObserver));
-      this.ref = ref;
     }
 
     CompletableFuture<RaftClientReply> onNext(RaftClientRequest request) {
-      final CompletableFuture<RaftClientReply> f = replies.putNew(request.getCallId());
+      final long callId = request.getCallId();
+      final CompletableFuture<RaftClientReply> f = replies.putNew(callId);
       if (f == null) {
         return JavaUtils.completeExceptionally(new AlreadyClosedException(getName() + " is closed."));
       }
       try {
         if (!requestStreamer.onNext(ClientProtoUtils.toRaftClientRequestProto(request))) {
-          throw new AlreadyClosedException(getName() + ": the stream is closed.");
+          return JavaUtils.completeExceptionally(new AlreadyClosedException(getName() + ": the stream is closed."));
         }
       } catch(Throwable t) {
         handleReplyFuture(request.getCallId(), future -> future.completeExceptionally(t));
         return f;
       }
 
-      scheduler.onTimeout(requestTimeoutDuration, () -> timeoutCheck(request), LOG,
-          () -> "Timeout check failed for client request: " + request);
+      if (RaftClientRequestProto.TypeCase.WATCH.equals(request.getType().getTypeCase())) {
+        scheduler.onTimeout(watchRequestTimeoutDuration, () ->
+                timeoutCheck(callId, watchRequestTimeoutDuration), LOG,
+            () -> "Timeout check failed for client request #" + callId);
+      } else {
+        scheduler.onTimeout(requestTimeoutDuration,
+            () -> timeoutCheck(callId, requestTimeoutDuration), LOG,
+            () -> "Timeout check failed for client request #" + callId);
+      }
       return f;
     }
 
-    private void timeoutCheck(RaftClientRequest request) {
-      handleReplyFuture(request.getCallId(), f -> f.completeExceptionally(
-          new TimeoutIOException("Request timeout " + requestTimeoutDuration + ": " + request)));
+    private void timeoutCheck(long callId, TimeDuration timeOutDuration) {
+      handleReplyFuture(callId, f -> f.completeExceptionally(
+          new TimeoutIOException("Request #" + callId + " timeout " + timeOutDuration)));
     }
 
     private void handleReplyFuture(long callId, Consumer<CompletableFuture<RaftClientReply>> handler) {
@@ -317,8 +346,6 @@ public class GrpcClientProtocolClient implements Closeable {
     }
 
     private void completeReplyExceptionally(Throwable t, String event) {
-      ref.compareAndSet(this, null);
-
       final Map<Long, CompletableFuture<RaftClientReply>> map = replies.getAndSetNull();
       if (map == null) {
         return;
