@@ -34,12 +34,14 @@ import org.slf4j.LoggerFactory;
 
 import java.io.IOException;
 import java.util.HashMap;
+import java.util.Iterator;
 import java.util.Map;
 import java.util.concurrent.CompletableFuture;
 import java.util.concurrent.CompletionException;
 import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.atomic.AtomicBoolean;
 import java.util.concurrent.atomic.AtomicInteger;
+import java.util.concurrent.atomic.AtomicReference;
 import java.util.function.Consumer;
 import java.util.function.Supplier;
 
@@ -48,24 +50,31 @@ public class GrpcClientProtocolService extends RaftClientProtocolServiceImplBase
 
   private static class PendingOrderedRequest implements SlidingWindow.ServerSideRequest<RaftClientReply> {
     private final RaftClientRequest request;
-    private volatile RaftClientReply reply;
+    private final AtomicReference<RaftClientReply> reply = new AtomicReference<>();
 
     PendingOrderedRequest(RaftClientRequest request) {
       this.request = request;
     }
 
     @Override
-    public boolean hasReply() {
-      return reply != null || this == COMPLETED;
+    public void fail(Throwable t) {
+      Preconditions.assertTrue(t instanceof RaftException, () -> "Requires RaftException but " + t);
+      setReply(new RaftClientReply(request, (RaftException) t, null));
     }
 
     @Override
-    public void setReply(RaftClientReply reply) {
-      this.reply = reply;
+    public boolean hasReply() {
+      return getReply() != null || this == COMPLETED;
+    }
+
+    @Override
+    public void setReply(RaftClientReply r) {
+      final boolean set = reply.compareAndSet(null, r);
+      Preconditions.assertTrue(set, () -> "Reply is already set: request=" + request + ", reply=" + reply);
     }
 
     RaftClientReply getReply() {
-      return reply;
+      return reply.get();
     }
 
     RaftClientRequest getRequest() {
@@ -100,12 +109,17 @@ public class GrpcClientProtocolService extends RaftClientProtocolServiceImplBase
       CollectionUtils.removeExisting(so.getId(), so, map, () -> getClass().getSimpleName());
     }
 
-    void closeAllExisting() {
+    void closeAllExisting(RaftGroupId groupId) {
       // Iteration not synchronized:
       // Okay if an existing object is removed by another mean during the iteration since it must be already closed.
       // Also okay if a new object is added during the iteration since this method closes only the existing objects.
-      for(OrderedRequestStreamObserver so : map.values()) {
-        so.close();
+      for(Iterator<Map.Entry<Integer, OrderedRequestStreamObserver>> i = map.entrySet().iterator(); i.hasNext(); ) {
+        final OrderedRequestStreamObserver so = i.next().getValue();
+        final RaftGroupId gid = so.getGroupId();
+        if (gid == null || gid.equals(groupId)) {
+          so.close(true);
+          i.remove();
+        }
       }
     }
   }
@@ -139,9 +153,9 @@ public class GrpcClientProtocolService extends RaftClientProtocolServiceImplBase
     return so;
   }
 
-  public void closeAllOrderedRequestStreamObservers() {
+  public void closeAllOrderedRequestStreamObservers(RaftGroupId groupId) {
     LOG.debug("{}: closeAllOrderedRequestStreamObservers", getId());
-    orderedStreamObservers.closeAllExisting();
+    orderedStreamObservers.closeAllExisting(groupId);
   }
 
   @Override
@@ -201,15 +215,20 @@ public class GrpcClientProtocolService extends RaftClientProtocolServiceImplBase
       return isClosed.compareAndSet(false, true);
     }
 
+    boolean isClosed() {
+      return isClosed.get();
+    }
+
     CompletableFuture<Void> processClientRequest(RaftClientRequest request, Consumer<RaftClientReply> replyHandler) {
       try {
+        String errMsg = LOG.isDebugEnabled() ? "processClientRequest for " + request.toString() : "";
         return protocol.submitClientRequestAsync(request
         ).thenAcceptAsync(replyHandler
         ).exceptionally(exception -> {
           // TODO: the exception may be from either raft or state machine.
           // Currently we skip all the following responses when getting an
           // exception from the state machine.
-          responseError(exception, () -> "processClientRequest for " + request);
+          responseError(exception, () -> errMsg);
           return null;
         });
       } catch (IOException e) {
@@ -296,26 +315,50 @@ public class GrpcClientProtocolService extends RaftClientProtocolServiceImplBase
   private class OrderedRequestStreamObserver extends RequestStreamObserver {
     private final SlidingWindow.Server<PendingOrderedRequest, RaftClientReply> slidingWindow
         = new SlidingWindow.Server<>(getName(), COMPLETED);
+    /** The {@link RaftGroupId} for this observer. */
+    private final AtomicReference<RaftGroupId> groupId = new AtomicReference<>();
 
     OrderedRequestStreamObserver(StreamObserver<RaftClientReplyProto> responseObserver) {
       super(responseObserver);
     }
 
+    RaftGroupId getGroupId() {
+      return groupId.get();
+    }
+
     void processClientRequest(PendingOrderedRequest pending) {
       final long seq = pending.getSeqNum();
       processClientRequest(pending.getRequest(),
-          reply -> slidingWindow.receiveReply(seq, reply, this::sendReply, this::processClientRequest));
+          reply -> slidingWindow.receiveReply(seq, reply, this::sendReply));
     }
 
     @Override
     void processClientRequest(RaftClientRequest r) {
-      slidingWindow.receivedRequest(new PendingOrderedRequest(r), this::processClientRequest);
+      if (isClosed()) {
+        final AlreadyClosedException exception = new AlreadyClosedException(getName() + ": the stream is closed");
+        responseError(exception, () -> "processClientRequest (stream already closed) for " + r);
+      }
+
+      final RaftGroupId requestGroupId = r.getRaftGroupId();
+      // use the group id in the first request as the group id of this observer
+      final RaftGroupId updated = groupId.updateAndGet(g -> g != null ? g: requestGroupId);
+      final PendingOrderedRequest pending = new PendingOrderedRequest(r);
+
+      if (!requestGroupId.equals(updated)) {
+        final GroupMismatchException exception = new GroupMismatchException(getId()
+            + ": The group (" + requestGroupId + ") of " + r.getClientId()
+            + " does not match the group (" + updated + ") of the " + getClass().getSimpleName());
+        responseError(exception, () -> "processClientRequest (Group mismatched) for " + r);
+        return;
+      }
+
+      slidingWindow.receivedRequest(pending, this::processClientRequest);
     }
 
     private void sendReply(PendingOrderedRequest ready) {
       Preconditions.assertTrue(ready.hasReply());
       if (ready == COMPLETED) {
-        close();
+        close(true);
       } else {
         LOG.debug("{}: sendReply seq={}, {}", getName(), ready.getSeqNum(), ready.getReply());
         responseNext(ClientProtoUtils.toRaftClientReplyProto(ready.getReply()));
@@ -326,29 +369,35 @@ public class GrpcClientProtocolService extends RaftClientProtocolServiceImplBase
     public void onError(Throwable t) {
       // for now we just log a msg
       GrpcUtil.warn(LOG, () -> getName() + ": onError", t);
-      slidingWindow.close();
+      close(false);
     }
 
     @Override
     public void onCompleted() {
-      if (slidingWindow.endOfRequests()) {
-        close();
+      if (slidingWindow.endOfRequests(this::sendReply)) {
+        close(true);
       }
     }
 
-    private void close() {
+    private void close(boolean complete) {
       if (setClose()) {
         LOG.debug("{}: close", getName());
-        responseCompleted();
-        slidingWindow.close();
-        orderedStreamObservers.removeExisting(this);
+        if (complete) {
+          responseCompleted();
+        }
+        cleanup();
       }
     }
 
+    private void cleanup() {
+      slidingWindow.close();
+      orderedStreamObservers.removeExisting(this);
+    }
+
     @Override
     boolean responseError(Throwable t, Supplier<String> message) {
       if (super.responseError(t, message)) {
-        slidingWindow.close();
+        cleanup();
         return true;
       }
       return false;
