@@ -19,12 +19,14 @@ package org.apache.ratis.grpc.server;
 
 import org.apache.ratis.grpc.GrpcConfigKeys;
 import org.apache.ratis.grpc.GrpcUtil;
+import org.apache.ratis.grpc.metrics.GrpcServerMetrics;
 import org.apache.ratis.server.RaftServerConfigKeys;
 import org.apache.ratis.server.impl.FollowerInfo;
 import org.apache.ratis.server.impl.LeaderState;
 import org.apache.ratis.server.impl.LogAppender;
 import org.apache.ratis.server.impl.RaftServerImpl;
 import org.apache.ratis.server.impl.ServerProtoUtils;
+import org.apache.ratis.server.protocol.TermIndex;
 import org.apache.ratis.thirdparty.io.grpc.stub.StreamObserver;
 import org.apache.ratis.proto.RaftProtos.AppendEntriesReplyProto;
 import org.apache.ratis.proto.RaftProtos.AppendEntriesRequestProto;
@@ -40,6 +42,8 @@ import java.util.*;
 import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.atomic.AtomicBoolean;
 
+import com.codahale.metrics.Timer;
+
 /**
  * A new log appender implementation using grpc bi-directional stream API.
  */
@@ -47,17 +51,19 @@ public class GrpcLogAppender extends LogAppender {
   public static final Logger LOG = LoggerFactory.getLogger(GrpcLogAppender.class);
 
   private final GrpcService rpcService;
-  private final Map<Long, AppendEntriesRequestProto> pendingRequests;
+  private final RequestMap pendingRequests = new RequestMap();
   private final int maxPendingRequestsNum;
   private long callId = 0;
   private volatile boolean firstResponseReceived = false;
   private final boolean installSnapshotEnabled;
 
   private final TimeDuration requestTimeoutDuration;
-  private final TimeoutScheduler scheduler = TimeoutScheduler.newInstance(1);
+  private final TimeoutScheduler scheduler = TimeoutScheduler.getInstance();
 
   private volatile StreamObserver<AppendEntriesRequestProto> appendLogRequestObserver;
 
+  private final GrpcServerMetrics grpcServerMetrics;
+
   public GrpcLogAppender(RaftServerImpl server, LeaderState leaderState,
                          FollowerInfo f) {
     super(server, leaderState, f);
@@ -67,24 +73,29 @@ public class GrpcLogAppender extends LogAppender {
     maxPendingRequestsNum = GrpcConfigKeys.Server.leaderOutstandingAppendsMax(
         server.getProxy().getProperties());
     requestTimeoutDuration = RaftServerConfigKeys.Rpc.requestTimeout(server.getProxy().getProperties());
-    pendingRequests = new ConcurrentHashMap<>();
-    installSnapshotEnabled = GrpcConfigKeys.LogAppender.installSnapshotEnabled(
+    installSnapshotEnabled = RaftServerConfigKeys.Log.Appender.installSnapshotEnabled(
         server.getProxy().getProperties());
+    grpcServerMetrics = new GrpcServerMetrics(server.getMemberId().toString());
+    grpcServerMetrics.addPendingRequestsCount(server.getMemberId().toString(),
+        () -> pendingRequests.logRequestsSize());
   }
 
   private GrpcServerProtocolClient getClient() throws IOException {
-    return rpcService.getProxies().getProxy(follower.getPeer().getId());
+    return rpcService.getProxies().getProxy(getFollowerId());
   }
 
-  private synchronized void resetClient(AppendEntriesRequestProto request) {
-    rpcService.getProxies().resetProxy(follower.getPeer().getId());
+  private synchronized void resetClient(AppendEntriesRequest request) {
+    rpcService.getProxies().resetProxy(getFollowerId());
     appendLogRequestObserver = null;
     firstResponseReceived = false;
 
     // clear the pending requests queue and reset the next index of follower
-    final long nextIndex = request != null && request.hasPreviousLog()?
-        request.getPreviousLog().getIndex() + 1: raftLog.getStartIndex();
-    clearPendingRequests(nextIndex);
+    final long nextIndex = 1 + Optional.ofNullable(request)
+        .map(AppendEntriesRequest::getPreviousLog)
+        .map(TermIndex::getIndex)
+        .orElseGet(follower::getMatchIndex);
+    pendingRequests.clear();
+    follower.decreaseNextIndex(nextIndex);
   }
 
   @Override
@@ -99,8 +110,14 @@ public class GrpcLogAppender extends LogAppender {
             installSnapshot(snapshot);
             shouldAppendLog = false;
           }
+        } else {
+          TermIndex installSnapshotNotificationTermIndex = shouldNotifyToInstallSnapshot();
+          if (installSnapshotNotificationTermIndex != null) {
+            installSnapshot(installSnapshotNotificationTermIndex);
+            shouldAppendLog = false;
+          }
         }
-        if (shouldAppendLog && !shouldWait()) {
+        if (shouldHeartbeat() || (shouldAppendLog && !shouldWait())) {
           // keep appending log entries or sending heartbeats
           appendLog();
         }
@@ -144,7 +161,7 @@ public class GrpcLogAppender extends LogAppender {
 
   /** @return true iff not received first response or queue is full. */
   private boolean shouldWait() {
-    final int size = pendingRequests.size();
+    final int size = pendingRequests.logRequestsSize();
     if (size == 0) {
       return false;
     }
@@ -153,6 +170,7 @@ public class GrpcLogAppender extends LogAppender {
 
   private void appendLog() throws IOException {
     final AppendEntriesRequestProto pending;
+    final AppendEntriesRequest request;
     final StreamObserver<AppendEntriesRequestProto> s;
     synchronized (this) {
       // prepare and enqueue the append request. note changes on follower's
@@ -162,8 +180,11 @@ public class GrpcLogAppender extends LogAppender {
       if (pending == null) {
         return;
       }
-      pendingRequests.put(pending.getServerRequest().getCallId(), pending);
-      updateNextIndex(pending);
+      grpcServerMetrics.onRequestCreate();
+      request = new AppendEntriesRequest(pending,
+          grpcServerMetrics.getGrpcLogAppenderLatencyTimer(getFollowerId().toString()));
+      pendingRequests.put(request);
+      increaseNextIndex(pending);
       if (appendLogRequestObserver == null) {
         appendLogRequestObserver = getClient().appendEntries(new AppendLogResponseHandler());
       }
@@ -171,59 +192,68 @@ public class GrpcLogAppender extends LogAppender {
     }
 
     if (isAppenderRunning()) {
-      sendRequest(pending, s);
+      sendRequest(request, pending, s);
     }
   }
 
-  private void sendRequest(AppendEntriesRequestProto request,
-      StreamObserver<AppendEntriesRequestProto> s) {
+  private void sendRequest(AppendEntriesRequest request, AppendEntriesRequestProto proto, StreamObserver<AppendEntriesRequestProto> s) {
     CodeInjectionForTesting.execute(GrpcService.GRPC_SEND_SERVER_REQUEST,
-        server.getId(), null, request);
-
-    s.onNext(request);
-    scheduler.onTimeout(requestTimeoutDuration, () -> timeoutAppendRequest(request), LOG,
-        () -> "Timeout check failed for append entry request: " + request);
+        server.getId(), null, proto);
+    request.startRequestTimer();
+    s.onNext(proto);
+    scheduler.onTimeout(requestTimeoutDuration,
+        () -> timeoutAppendRequest(request.getCallId(), request.isHeartbeat()),
+        LOG, () -> "Timeout check failed for append entry request: " + request);
     follower.updateLastRpcSendTime();
   }
 
-  private void timeoutAppendRequest(AppendEntriesRequestProto request) {
-    AppendEntriesRequestProto pendingRequest = pendingRequests.remove(request.getServerRequest().getCallId());
-    if (pendingRequest != null) {
-      LOG.warn( "{}: appendEntries Timeout, request={}", this, ServerProtoUtils.toString(pendingRequest));
+  private void timeoutAppendRequest(long cid, boolean heartbeat) {
+    final AppendEntriesRequest pending = pendingRequests.handleTimeout(cid, heartbeat);
+    if (pending != null) {
+      LOG.warn("{}: {} appendEntries Timeout, request={}", this, heartbeat ? "HEARTBEAT" : "", pending);
+      grpcServerMetrics.onRequestTimeout(getFollowerId().toString());
     }
   }
 
-  private void updateNextIndex(AppendEntriesRequestProto request) {
+  private void increaseNextIndex(AppendEntriesRequestProto request) {
     final int count = request.getEntriesCount();
     if (count > 0) {
-      follower.updateNextIndex(request.getEntries(count - 1).getIndex() + 1);
+      follower.increaseNextIndex(request.getEntries(count - 1).getIndex() + 1);
     }
   }
 
   /**
    * StreamObserver for handling responses from the follower
    */
-  private class AppendLogResponseHandler
-      implements StreamObserver<AppendEntriesReplyProto> {
+  private class AppendLogResponseHandler implements StreamObserver<AppendEntriesReplyProto> {
+    private final String name = follower.getName() + "-" + getClass().getSimpleName();
+
     /**
      * After receiving a appendEntries reply, do the following:
      * 1. If the reply is success, update the follower's match index and submit
      *    an event to leaderState
      * 2. If the reply is NOT_LEADER, step down
-     * 3. If the reply is INCONSISTENCY, decrease the follower's next index
-     *    based on the response
+     * 3. If the reply is INCONSISTENCY, increase/ decrease the follower's next
+     *    index based on the response
      */
     @Override
     public void onNext(AppendEntriesReplyProto reply) {
+      AppendEntriesRequest request = pendingRequests.remove(reply);
+      if (request != null) {
+        request.stopRequestTimer(); // Update completion time
+      }
+
       if (LOG.isDebugEnabled()) {
-        LOG.debug("{}<-{}: received {} reply {} ", server.getId(), follower.getPeer(),
-            (!firstResponseReceived? "the first": "a"), ServerProtoUtils.toString(reply));
+        LOG.debug("{}: received {} reply {}, request={}",
+            this, firstResponseReceived? "a": "the first",
+            ServerProtoUtils.toString(reply), request);
       }
 
       try {
         onNextImpl(reply);
       } catch(Throwable t) {
-        LOG.error("Failed onNext " + reply, t);
+        LOG.error("Failed onNext request=" + request
+            + ", reply=" + ServerProtoUtils.toString(reply), t);
       }
     }
 
@@ -234,18 +264,27 @@ public class GrpcLogAppender extends LogAppender {
       if (!firstResponseReceived) {
         firstResponseReceived = true;
       }
+
       switch (reply.getResult()) {
         case SUCCESS:
-          onSuccess(reply);
+          grpcServerMetrics.onRequestSuccess(getFollowerId().toString());
+          updateCommitIndex(reply.getFollowerCommit());
+          if (follower.updateMatchIndex(reply.getMatchIndex())) {
+            submitEventOnSuccessAppend();
+          }
           break;
         case NOT_LEADER:
-          onNotLeader(reply);
+          grpcServerMetrics.onRequestNotLeader(getFollowerId().toString());
+          if (checkResponseTerm(reply.getTerm())) {
+            return;
+          }
           break;
         case INCONSISTENCY:
-          onInconsistency(reply);
+          grpcServerMetrics.onRequestInconsistency(getFollowerId().toString());
+          updateNextIndex(reply.getNextIndex());
           break;
         default:
-          break;
+          throw new IllegalStateException("Unexpected reply result: " + reply.getResult());
       }
       notifyAppend();
     }
@@ -259,78 +298,31 @@ public class GrpcLogAppender extends LogAppender {
         LOG.info("{} is stopped", GrpcLogAppender.this);
         return;
       }
-      GrpcUtil.warn(LOG, () -> server.getId() + ": Failed appendEntries to " + follower.getPeer(), t);
-
-      long callId = GrpcUtil.getCallId(t);
-      resetClient(pendingRequests.get(callId));
+      GrpcUtil.warn(LOG, () -> this + ": Failed appendEntries", t);
+      grpcServerMetrics.onRequestRetry(); // Update try counter
+      AppendEntriesRequest request = pendingRequests.remove(GrpcUtil.getCallId(t), GrpcUtil.isHeartbeat(t));
+      resetClient(request);
     }
 
     @Override
     public void onCompleted() {
-      LOG.info("{}: follower {} response Completed", server.getId(), follower);
+      LOG.info("{}: follower responses appendEntries COMPLETED", this);
       resetClient(null);
     }
+
+    @Override
+    public String toString() {
+      return name;
+    }
   }
 
-  private void clearPendingRequests(long newNextIndex) {
+  private synchronized void updateNextIndex(long replyNextIndex) {
     pendingRequests.clear();
-    follower.decreaseNextIndex(newNextIndex);
+    follower.updateNextIndex(replyNextIndex);
   }
 
-  protected synchronized void onSuccess(AppendEntriesReplyProto reply) {
-    AppendEntriesRequestProto request = pendingRequests.remove(reply.getServerReply().getCallId());
-    if (request == null) {
-      // If reply comes after timeout, the reply is ignored.
-      LOG.warn("{}: Request not found, ignoring reply: {}", this, ServerProtoUtils.toString(reply));
-      return;
-    }
-    updateCommitIndex(reply.getFollowerCommit());
-
-    final long replyNextIndex = reply.getNextIndex();
-    final long lastIndex = replyNextIndex - 1;
-    final boolean updateMatchIndex;
-
-    if (request.getEntriesCount() == 0) {
-      Preconditions.assertTrue(!request.hasPreviousLog() ||
-              lastIndex == request.getPreviousLog().getIndex(),
-          "reply's next index is %s, request's previous is %s",
-          replyNextIndex, request.getPreviousLog());
-      updateMatchIndex = request.hasPreviousLog() && follower.getMatchIndex() < lastIndex;
-    } else {
-      // check if the reply and the pending request is consistent
-      final long lastEntryIndex = request
-          .getEntries(request.getEntriesCount() - 1).getIndex();
-      Preconditions.assertTrue(lastIndex == lastEntryIndex,
-          "reply's next index is %s, request's last entry index is %s",
-          replyNextIndex, lastEntryIndex);
-      updateMatchIndex = true;
-    }
-    if (updateMatchIndex) {
-      follower.updateMatchIndex(lastIndex);
-      submitEventOnSuccessAppend();
-    }
-  }
-
-  private void onNotLeader(AppendEntriesReplyProto reply) {
-    checkResponseTerm(reply.getTerm());
-    // the running loop will end and the connection will onComplete
-  }
-
-  private synchronized void onInconsistency(AppendEntriesReplyProto reply) {
-    AppendEntriesRequestProto request = pendingRequests.remove(reply.getServerReply().getCallId());
-    if (request == null) {
-      // If reply comes after timeout, the reply is ignored.
-      LOG.warn("{}: Ignoring {}", server.getId(), reply);
-      return;
-    }
-    Preconditions.assertTrue(request.hasPreviousLog());
-    if (request.getPreviousLog().getIndex() >= reply.getNextIndex()) {
-      clearPendingRequests(reply.getNextIndex());
-    }
-  }
-
-  private class InstallSnapshotResponseHandler
-      implements StreamObserver<InstallSnapshotReplyProto> {
+  private class InstallSnapshotResponseHandler implements StreamObserver<InstallSnapshotReplyProto> {
+    private final String name = follower.getName() + "-" + getClass().getSimpleName();
     private final Queue<Integer> pending;
     private final AtomicBoolean done = new AtomicBoolean(false);
 
@@ -339,11 +331,12 @@ public class GrpcLogAppender extends LogAppender {
     }
 
     synchronized void addPending(InstallSnapshotRequestProto request) {
-      pending.offer(request.getRequestIndex());
+      pending.offer(request.getSnapshotChunk().getRequestIndex());
     }
 
     synchronized void removePending(InstallSnapshotReplyProto reply) {
-      int index = pending.poll();
+      final Integer index = pending.poll();
+      Objects.requireNonNull(index, "index == null");
       Preconditions.assertTrue(index == reply.getRequestIndex());
     }
 
@@ -362,9 +355,10 @@ public class GrpcLogAppender extends LogAppender {
 
     @Override
     public void onNext(InstallSnapshotReplyProto reply) {
-      LOG.debug("{} received {} response from {}", server.getId(),
-          (!firstResponseReceived ? "the first" : "a"),
-          follower.getPeer());
+      if (LOG.isInfoEnabled()) {
+        LOG.info("{}: received {} reply {}", this, firstResponseReceived ? "a" : "the first",
+            ServerProtoUtils.toString(reply));
+      }
 
       // update the last rpc time
       follower.updateLastRpcResponseTime();
@@ -375,11 +369,22 @@ public class GrpcLogAppender extends LogAppender {
 
       switch (reply.getResult()) {
         case SUCCESS:
+        case IN_PROGRESS:
+          removePending(reply);
+          break;
+        case ALREADY_INSTALLED:
+          final long followerSnapshotIndex = reply.getSnapshotIndex();
+          LOG.info("{}: set follower snapshotIndex to {}.", this, followerSnapshotIndex);
+          follower.setSnapshotIndex(followerSnapshotIndex);
           removePending(reply);
           break;
         case NOT_LEADER:
           checkResponseTerm(reply.getTerm());
           break;
+        case CONF_MISMATCH:
+          LOG.error("{}: Configuration Mismatch ({}): Leader {} has it set to {} but follower {} has it set to {}",
+              this, RaftServerConfigKeys.Log.Appender.INSTALL_SNAPSHOT_ENABLED_KEY,
+              server.getId(), installSnapshotEnabled, getFollowerId(), !installSnapshotEnabled);
         case UNRECOGNIZED:
           break;
       }
@@ -388,28 +393,33 @@ public class GrpcLogAppender extends LogAppender {
     @Override
     public void onError(Throwable t) {
       if (!isAppenderRunning()) {
-        LOG.info("{} is stopped", GrpcLogAppender.this);
+        LOG.info("{} is stopped", this);
         return;
       }
-      LOG.info("{} got error when installing snapshot to {}, exception: {}",
-          server.getId(), follower.getPeer(), t);
+      LOG.error("{}: Failed installSnapshot: {}", this, t);
       resetClient(null);
       close();
     }
 
     @Override
     public void onCompleted() {
-      LOG.info("{} stops sending snapshots to follower {}", server.getId(),
-          follower);
+      LOG.info("{}: follower responses installSnapshot COMPLETED", this);
       close();
     }
+
+    @Override
+    public String toString() {
+      return name;
+    }
   }
 
+  /**
+   * Send installSnapshot request to Follower with a snapshot.
+   * @param snapshot the snapshot to be sent to Follower
+   */
   private void installSnapshot(SnapshotInfo snapshot) {
-    LOG.info("{}: follower {}'s next index is {}," +
-            " log's start index is {}, need to install snapshot",
-        server.getId(), follower.getPeer(), follower.getNextIndex(),
-        raftLog.getStartIndex());
+    LOG.info("{}: followerNextIndex = {} but logStartIndex = {}, send snapshot {} to follower",
+        this, follower.getNextIndex(), raftLog.getStartIndex(), snapshot);
 
     final InstallSnapshotResponseHandler responseHandler = new InstallSnapshotResponseHandler();
     StreamObserver<InstallSnapshotRequestProto> snapshotRequestObserver = null;
@@ -427,9 +437,9 @@ public class GrpcLogAppender extends LogAppender {
         }
       }
       snapshotRequestObserver.onCompleted();
+      grpcServerMetrics.onInstallSnapshot();
     } catch (Exception e) {
-      LOG.warn("{} failed to install snapshot {}. Exception: {}", this,
-          snapshot.getFiles(), e);
+      LOG.warn("{}: failed to install snapshot {}: {}", this, snapshot.getFiles(), e);
       if (snapshotRequestObserver != null) {
         snapshotRequestObserver.onError(e);
       }
@@ -447,8 +457,141 @@ public class GrpcLogAppender extends LogAppender {
 
     if (responseHandler.hasAllResponse()) {
       follower.setSnapshotIndex(snapshot.getTermIndex().getIndex());
-      LOG.info("{}: install snapshot-{} successfully on follower {}",
-          server.getId(), snapshot.getTermIndex().getIndex(), follower.getPeer());
+      LOG.info("{}: installed snapshot {} successfully", this, snapshot);
+    }
+  }
+
+  /**
+   * Send installSnapshot request to Follower with only a notification that a snapshot needs to be installed.
+   * @param firstAvailableLogTermIndex the first available log's index on the Leader
+   */
+  private void installSnapshot(TermIndex firstAvailableLogTermIndex) {
+    LOG.info("{}: followerNextIndex = {} but logStartIndex = {}, notify follower to install snapshot-{}",
+        this, follower.getNextIndex(), raftLog.getStartIndex(), firstAvailableLogTermIndex);
+
+    final InstallSnapshotResponseHandler responseHandler = new InstallSnapshotResponseHandler();
+    StreamObserver<InstallSnapshotRequestProto> snapshotRequestObserver = null;
+    // prepare and enqueue the notify install snapshot request.
+    final InstallSnapshotRequestProto request = createInstallSnapshotNotificationRequest(firstAvailableLogTermIndex);
+    if (LOG.isInfoEnabled()) {
+      LOG.info("{}: send {}", this, ServerProtoUtils.toString(request));
+    }
+    try {
+      snapshotRequestObserver = getClient().installSnapshot(responseHandler);
+      snapshotRequestObserver.onNext(request);
+      follower.updateLastRpcSendTime();
+      responseHandler.addPending(request);
+      snapshotRequestObserver.onCompleted();
+    } catch (Exception e) {
+      GrpcUtil.warn(LOG, () -> this + ": Failed to notify follower to install snapshot.", e);
+      if (snapshotRequestObserver != null) {
+        snapshotRequestObserver.onError(e);
+      }
+      return;
+    }
+
+    synchronized (this) {
+      if (isAppenderRunning() && !responseHandler.isDone()) {
+        try {
+          wait();
+        } catch (InterruptedException ignored) {
+        }
+      }
+    }
+  }
+
+  /**
+   * Should the Leader notify the Follower to install the snapshot through
+   * its own State Machine.
+   * @return the first available log's start term index
+   */
+  private TermIndex shouldNotifyToInstallSnapshot() {
+    if (follower.getNextIndex() < raftLog.getStartIndex()) {
+      // The Leader does not have the logs from the Follower's last log
+      // index onwards. And install snapshot is disabled. So the Follower
+      // should be notified to install the latest snapshot through its
+      // State Machine.
+      return raftLog.getTermIndex(raftLog.getStartIndex());
+    }
+    return null;
+  }
+
+  static class AppendEntriesRequest {
+    private final Timer timer;
+    private volatile Timer.Context timerContext;
+
+    private final long callId;
+    private final TermIndex previousLog;
+    private final int entriesCount;
+
+    private final TermIndex lastEntry;
+
+    AppendEntriesRequest(AppendEntriesRequestProto proto, Timer timer) {
+      this.callId = proto.getServerRequest().getCallId();
+      this.previousLog = proto.hasPreviousLog()? ServerProtoUtils.toTermIndex(proto.getPreviousLog()): null;
+      this.entriesCount = proto.getEntriesCount();
+      this.lastEntry = entriesCount > 0? ServerProtoUtils.toTermIndex(proto.getEntries(entriesCount - 1)): null;
+
+      this.timer = timer;
+    }
+
+    long getCallId() {
+      return callId;
+    }
+
+    TermIndex getPreviousLog() {
+      return previousLog;
+    }
+
+    void startRequestTimer() {
+      timerContext = timer.time();
+    }
+
+    void stopRequestTimer() {
+      timerContext.stop();
+    }
+
+    boolean isHeartbeat() {
+      return entriesCount == 0;
+    }
+
+    @Override
+    public String toString() {
+      return getClass().getSimpleName() + ":cid=" + callId + ",entriesCount=" + entriesCount + ",lastEntry=" + lastEntry;
+    }
+  }
+
+  static class RequestMap {
+    private final Map<Long, AppendEntriesRequest> logRequests = new ConcurrentHashMap<>();
+    private final Map<Long, AppendEntriesRequest> heartbeats = new ConcurrentHashMap<>();
+
+    int logRequestsSize() {
+      return logRequests.size();
+    }
+
+    void clear() {
+      logRequests.clear();
+      heartbeats.clear();
+    }
+
+    void put(AppendEntriesRequest request) {
+      if (request.isHeartbeat()) {
+        heartbeats.put(request.getCallId(), request);
+      } else {
+        logRequests.put(request.getCallId(), request);
+      }
+    }
+
+    AppendEntriesRequest remove(AppendEntriesReplyProto reply) {
+      return remove(reply.getServerReply().getCallId(), reply.getIsHearbeat());
+    }
+
+    AppendEntriesRequest remove(long cid, boolean isHeartbeat) {
+      return isHeartbeat ? heartbeats.remove(cid): logRequests.remove(cid);
+    }
+
+    public AppendEntriesRequest handleTimeout(long callId, boolean heartbeat) {
+      return heartbeat ? heartbeats.remove(callId) : logRequests.get(callId);
     }
   }
 }
