@@ -30,6 +30,7 @@ import org.apache.ratis.logservice.api.LogServiceConfiguration;
 import org.apache.ratis.logservice.api.LogStream;
 import org.apache.ratis.logservice.api.LogWriter;
 import org.apache.ratis.logservice.api.RecordListener;
+import org.apache.ratis.logservice.proto.LogServiceProtos;
 import org.apache.ratis.logservice.proto.LogServiceProtos.GetLogLastCommittedIndexReplyProto;
 import org.apache.ratis.logservice.proto.LogServiceProtos.GetLogLengthReplyProto;
 import org.apache.ratis.logservice.proto.LogServiceProtos.GetLogSizeReplyProto;
@@ -47,34 +48,34 @@ public class LogStreamImpl implements LogStream {
   /*
    * Log stream listeners
    */
-  List<RecordListener> listeners;
+  private List<RecordListener> listeners;
   /*
    * Log stream name
    */
-  LogName name;
+  private LogName name;
   /*
    * Parent log service instance
    */
-  RaftClient raftClient;
+  private RaftClient raftClient;
   /*
    * Log stream configuration
    */
-  LogServiceConfiguration config;
+  private LogServiceConfiguration config;
   /*
    * State
    */
-  LogStream.State state;
+  private LogStream.State state;
 
   /*
    * Length
    */
-  long length;
+  private long length;
 
 
   public LogStreamImpl(LogName name, RaftClient raftClient) {
     this.raftClient = raftClient;
     this.name = name;
-    this.config = new LogServiceConfiguration();
+    this.config = LogServiceConfiguration.create();
     init();
   }
 
@@ -87,8 +88,8 @@ public class LogStreamImpl implements LogStream {
 
   private void init() {
     // TODO create new state machine. etc
-    state = State.OPEN;
-    listeners = Collections.synchronizedList(new ArrayList<RecordListener>());
+    this.state = State.OPEN;
+    this.listeners = Collections.synchronizedList(new ArrayList<RecordListener>());
   }
 
   @Override
@@ -96,9 +97,12 @@ public class LogStreamImpl implements LogStream {
     return name;
   }
 
-  @Override
-  public State getState() {
-    return state;
+  @Override public State getState() throws IOException {
+    RaftClientReply reply = raftClient.sendReadOnly(
+        Message.valueOf(LogServiceProtoUtil.toGetStateRequestProto(name).toByteString()));
+    LogServiceProtos.GetStateReplyProto proto =
+        LogServiceProtos.GetStateReplyProto.parseFrom(reply.getMessage().getContent());
+    return State.valueOf(proto.getState().name());
   }
 
   @Override
@@ -130,7 +134,7 @@ public class LogStreamImpl implements LogStream {
   }
 
   @Override
-  public LogReader createReader() {
+  public LogReader createReader() throws IOException{
     return new LogReaderImpl(this);
   }
 
