@@ -19,11 +19,7 @@
 package org.apache.ratis.logservice.server;
 
 import java.io.IOException;
-import java.util.Collection;
-import java.util.HashSet;
-import java.util.List;
-import java.util.Map;
-import java.util.Set;
+import java.util.*;
 import java.util.concurrent.CompletableFuture;
 import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.PriorityBlockingQueue;
@@ -32,16 +28,19 @@ import java.util.stream.Collectors;
 import java.util.stream.IntStream;
 import java.util.stream.StreamSupport;
 
+import com.codahale.metrics.Timer;
 import org.apache.ratis.client.RaftClient;
 import org.apache.ratis.conf.RaftProperties;
 import org.apache.ratis.logservice.api.LogInfo;
 import org.apache.ratis.logservice.api.LogName;
+import org.apache.ratis.logservice.api.LogStream;
+import org.apache.ratis.logservice.common.Constants;
 import org.apache.ratis.logservice.common.LogAlreadyExistException;
 import org.apache.ratis.logservice.common.LogNotFoundException;
 import org.apache.ratis.logservice.common.NoEnoughWorkersException;
+import org.apache.ratis.logservice.metrics.LogServiceMetricsRegistry;
+import org.apache.ratis.logservice.proto.LogServiceProtos;
 import org.apache.ratis.logservice.proto.MetaServiceProtos;
-import org.apache.ratis.logservice.proto.MetaServiceProtos.ArchiveLogReplyProto;
-import org.apache.ratis.logservice.proto.MetaServiceProtos.ArchiveLogRequestProto;
 import org.apache.ratis.logservice.proto.MetaServiceProtos.CreateLogRequestProto;
 import org.apache.ratis.logservice.proto.MetaServiceProtos.DeleteLogRequestProto;
 import org.apache.ratis.logservice.proto.MetaServiceProtos.LogServicePingRequestProto;
@@ -50,19 +49,18 @@ import org.apache.ratis.logservice.proto.MetaServiceProtos.LogServiceUnregisterL
 import org.apache.ratis.logservice.proto.MetaServiceProtos.MetaSMRequestProto;
 import org.apache.ratis.logservice.util.LogServiceProtoUtil;
 import org.apache.ratis.logservice.util.MetaServiceProtoUtil;
+import org.apache.ratis.metrics.RatisMetricRegistry;
 import org.apache.ratis.proto.RaftProtos;
-import org.apache.ratis.protocol.ClientId;
-import org.apache.ratis.protocol.Message;
-import org.apache.ratis.protocol.RaftClientRequest;
-import org.apache.ratis.protocol.RaftGroup;
-import org.apache.ratis.protocol.RaftGroupId;
-import org.apache.ratis.protocol.RaftPeer;
+import org.apache.ratis.protocol.*;
+
+
 import org.apache.ratis.server.RaftServer;
 import org.apache.ratis.server.storage.RaftStorage;
 import org.apache.ratis.statemachine.TransactionContext;
 import org.apache.ratis.statemachine.impl.BaseStateMachine;
 import org.apache.ratis.thirdparty.com.google.protobuf.InvalidProtocolBufferException;
 import org.apache.ratis.util.AutoCloseableLock;
+import org.apache.ratis.util.Daemon;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
@@ -76,7 +74,7 @@ import org.slf4j.LoggerFactory;
 
 public class MetaStateMachine extends BaseStateMachine {
 
-    Logger LOG = LoggerFactory.getLogger(MetaStateMachine.class);
+    private static final Logger LOG = LoggerFactory.getLogger(MetaStateMachine.class);
 
 
     //Persisted map between log and RaftGroup
@@ -87,10 +85,15 @@ public class MetaStateMachine extends BaseStateMachine {
     // keep a copy of raftServer to get group information.
     private RaftServer raftServer;
 
+    private Map<RaftPeer, Set<LogName>> peerLogs = new ConcurrentHashMap<>();
+
+    private Map<RaftPeer, Long> heartbeatInfo = new ConcurrentHashMap<>();
 
     private RaftGroup currentGroup = null;
 
+    private Daemon peerHealthChecker = null;
     // MinHeap queue for load balancing groups across the peers
+    private long failureDetectionPeriod = Constants.DEFAULT_PEER_FAILURE_DETECTION_PERIOD;
     private PriorityBlockingQueue<PeerGroups> avail = new PriorityBlockingQueue<PeerGroups>();
 
     //Properties
@@ -100,16 +103,23 @@ public class MetaStateMachine extends BaseStateMachine {
 
     private RaftGroupId metadataGroupId;
     private RaftGroupId logServerGroupId;
+    private RatisMetricRegistry metricRegistry;
 
-    public MetaStateMachine(RaftGroupId metadataGroupId, RaftGroupId logServerGroupId) {
+    public MetaStateMachine(RaftGroupId metadataGroupId, RaftGroupId logServerGroupId,
+                            long failureDetectionPeriod) {
       this.metadataGroupId = metadataGroupId;
       this.logServerGroupId = logServerGroupId;
+      this.failureDetectionPeriod = failureDetectionPeriod;
     }
 
     @Override
     public void initialize(RaftServer server, RaftGroupId groupId, RaftStorage storage) throws IOException {
         this.raftServer = server;
+        this.metricRegistry = LogServiceMetricsRegistry
+            .createMetricRegistryForLogServiceMetaData(server.getId().toString());
         super.initialize(server, groupId, storage);
+        peerHealthChecker = new Daemon(new PeerHealthChecker(),"peer-Health-Checker");
+        peerHealthChecker.start();
     }
 
     @Override
@@ -126,7 +136,19 @@ public class MetaStateMachine extends BaseStateMachine {
                 LogServiceRegisterLogRequestProto r = req.getRegisterRequest();
                 LogName logname = LogServiceProtoUtil.toLogName(r.getLogname());
                 RaftGroup rg = MetaServiceProtoUtil.toRaftGroup(r.getRaftGroup());
+                rg.getPeers().stream().forEach(raftPeer -> {
+                    Set<LogName> logNames;
+                    if(!peerLogs.containsKey(raftPeer)) {
+                        logNames = new HashSet<>();
+                        peerLogs.put(raftPeer, logNames);
+                    } else {
+                        logNames = peerLogs.get(raftPeer);
+                    }
+                    logNames.add(logname);
+
+                });
                 map.put(logname, rg);
+
                 LOG.info("Log {} registered at {} with group {} ", logname, getId(), rg );
                 break;
             case UNREGISTERREQUEST:
@@ -137,14 +159,18 @@ public class MetaStateMachine extends BaseStateMachine {
             case PINGREQUEST:
                 LogServicePingRequestProto pingRequest = req.getPingRequest();
                 RaftPeer peer = MetaServiceProtoUtil.toRaftPeer(pingRequest.getPeer());
-                if (peers.contains(peer)) {
-                    //Do Nothing, that's just heartbeat
-                } else {
+                //If Set<RaftPeer> contains peer then do nothing as that's just heartbeat else add the peer to the set.
+                if (!peers.contains(peer)) {
                     peers.add(peer);
                     avail.add(new PeerGroups(peer));
+                    heartbeatInfo.put(peer,  System.currentTimeMillis());
                 }
                 break;
-
+            case HEARTBEATREQUEST:
+                MetaServiceProtos.LogServiceHeartbeatRequestProto heartbeatRequest = req.getHeartbeatRequest();
+                RaftPeer heartbeatPeer = MetaServiceProtoUtil.toRaftPeer(heartbeatRequest.getPeer());
+                heartbeatInfo.put(heartbeatPeer,  System.currentTimeMillis());
+                break;
             default:
         }
         return super.applyTransactionSerial(trx);
@@ -172,27 +198,31 @@ public class MetaStateMachine extends BaseStateMachine {
 
     @Override
     public CompletableFuture<Message> query(Message request) {
-        if (currentGroup == null) {
-            try {
-                List<RaftGroup> x = StreamSupport.stream(raftServer.getGroups().spliterator(), false)
-                    .filter(group -> group.getGroupId().equals(metadataGroupId))
-                    .collect(Collectors.toList());
-                if (x.size() == 1) {
-                    currentGroup = x.get(0);
+        Timer.Context timerContext = null;
+        MetaServiceProtos.MetaServiceRequestProto.TypeCase type = null;
+        try {
+            if (currentGroup == null) {
+                try {
+                    List<RaftGroup> x =
+                        StreamSupport.stream(raftServer.getGroups().spliterator(), false)
+                            .filter(group -> group.getGroupId().equals(metadataGroupId)).collect(Collectors.toList());
+                    if (x.size() == 1) {
+                        currentGroup = x.get(0);
+                    }
+                } catch (IOException e) {
+                    e.printStackTrace();
                 }
-            } catch (IOException e) {
+            }
+
+            MetaServiceProtos.MetaServiceRequestProto req = null;
+            try {
+                req = MetaServiceProtos.MetaServiceRequestProto.parseFrom(request.getContent());
+            } catch (InvalidProtocolBufferException e) {
                 e.printStackTrace();
             }
-        }
-        RaftProperties properties = new RaftProperties();
-        MetaServiceProtos.MetaServiceRequestProto req = null;
-        try {
-            req =  MetaServiceProtos.MetaServiceRequestProto.parseFrom(request.getContent());
-        } catch (InvalidProtocolBufferException e) {
-            e.printStackTrace();
-        }
-        MetaServiceProtos.MetaServiceRequestProto.TypeCase type = req.getTypeCase();
-        switch (type) {
+            type = req.getTypeCase();
+            timerContext = metricRegistry.timer(type.name()).time();
+            switch (type) {
 
             case CREATELOG:
                 return processCreateLogRequest(req);
@@ -200,14 +230,17 @@ public class MetaStateMachine extends BaseStateMachine {
                 return processListLogsRequest();
             case GETLOG:
                 return processGetLogRequest(req);
-            case ARCHIVELOG:
-                return processArchiveLog(req);
             case DELETELOG:
                 return processDeleteLog(req);
-                default:
+            default:
+            }
+            CompletableFuture<Message> reply = super.query(request);
+            return reply;
+        }finally{
+            if (timerContext != null) {
+                timerContext.stop();
+            }
         }
-        CompletableFuture<Message> reply = super.query(request);
-        return reply;
     }
 
 
@@ -222,8 +255,8 @@ public class MetaStateMachine extends BaseStateMachine {
                     MetaServiceProtoUtil.toDeleteLogExceptionReplyProto(
                             new LogNotFoundException(logName.getName())).build().toByteString()));
         } else {
-            Collection<RaftPeer> peers = raftGroup.getPeers();
-            peers.stream().forEach(peer -> {
+            Collection<RaftPeer> raftPeers = raftGroup.getPeers();
+            raftPeers.stream().forEach(peer -> {
                 RaftClient client = RaftClient.newBuilder()
                         .setProperties(properties)
                         .setClientId(ClientId.randomId())
@@ -247,6 +280,8 @@ public class MetaStateMachine extends BaseStateMachine {
                                         .setLogname(LogServiceProtoUtil.toLogNameProto(logName)))
                         .build().toByteString());
             } catch (IOException e) {
+                LOG.error(
+                    "Exception while unregistring raft group with Metadata Service during deletion of log");
                 e.printStackTrace();
             }
         }
@@ -254,36 +289,11 @@ public class MetaStateMachine extends BaseStateMachine {
                 MetaServiceProtoUtil.toDeleteLogReplyProto().toByteString()));
     }
 
-//    private CompletableFuture<Message> processCloseLog(MetaServiceProtos.MetaServiceRequestProto logServiceRequestProto) {
-//        CloseLogRequestProto closeLog = logServiceRequestProto.getCloseLog();
-//        LogName logName = LogServiceProtoUtil.toLogName(closeLog.getLogName());
-//        // Need to check whether the file is opened if opened close it.
-//        // TODO need to handle exceptions while operating with files.
-//        return CompletableFuture.completedFuture(Message
-//                .valueOf(CloseLogReplyProto.newBuilder().build().toByteString()));
-//    }
 
-    private CompletableFuture<Message>
-    processArchiveLog(MetaServiceProtos.MetaServiceRequestProto logServiceRequestProto) {
-        ArchiveLogRequestProto archiveLog = logServiceRequestProto.getArchiveLog();
-        LogName logName = LogServiceProtoUtil.toLogName(archiveLog.getLogName());
-        // Handle log archiving.
-        return CompletableFuture.completedFuture(Message
-                .valueOf(ArchiveLogReplyProto.newBuilder().build().toByteString()));
-    }
-
-//    private CompletableFuture<Message> processGetStateRequest(
-//            MetaServiceProtos.MetaServiceRequestProto logServiceRequestProto) {
-//        MetaServiceProtos.GetStateRequestProto getState = logServiceRequestProto.getGetState();
-//        LogName logName = LogServiceProtoUtil.toLogName(getState.getLogName());
-//        return CompletableFuture.completedFuture(Message.valueOf(LogServiceProtoUtil
-//                .toGetStateReplyProto(true).toByteString()));
-//    }
-//
     private CompletableFuture<Message> processCreateLogRequest(
             MetaServiceProtos.MetaServiceRequestProto logServiceRequestProto) {
         LogName name;
-        try (final AutoCloseableLock writeLock = writeLock()) {
+        try (AutoCloseableLock writeLock = writeLock()) {
             CreateLogRequestProto createLog = logServiceRequestProto.getCreateLog();
             name = LogServiceProtoUtil.toLogName(createLog.getLogName());
             if(map.containsKey(name)) {
@@ -299,14 +309,16 @@ public class MetaStateMachine extends BaseStateMachine {
                         .build()
                         .toByteString()));
             } else {
-                List<PeerGroups> peerGroup = IntStream.range(0, 3).mapToObj(i -> avail.poll()).collect(Collectors.toList());
-                List<RaftPeer> peers = peerGroup.stream().map(obj -> obj.getPeer()).collect(Collectors.toList());
-                RaftGroup raftGroup = RaftGroup.valueOf(RaftGroupId.randomId(), peers);
+                List<PeerGroups> peerGroup = IntStream.range(0, 3).mapToObj(i -> avail.poll())
+                    .collect(Collectors.toList());
+                List<RaftPeer> peersFromGroup =
+                    peerGroup.stream().map(obj -> obj.getPeer()).collect(Collectors.toList());
+                RaftGroup raftGroup = RaftGroup.valueOf(RaftGroupId.randomId(), peersFromGroup);
                 peerGroup.stream().forEach(pg -> {
                     pg.getGroups().add(raftGroup);
                     avail.add(pg);
                 });
-                peers.forEach(i -> {
+                peersFromGroup.forEach(i -> {
                     RaftClient client = RaftClient.newBuilder().setProperties(properties)
                       .setRaftGroup(RaftGroup.valueOf(logServerGroupId, i)).build();
                     try {
@@ -322,13 +334,14 @@ public class MetaStateMachine extends BaseStateMachine {
                         .build();
                 try {
                     client.send(() -> MetaServiceProtos.MetaSMRequestProto.newBuilder()
-                            .setRegisterRequest(LogServiceRegisterLogRequestProto
-                                    .newBuilder()
+                            .setRegisterRequest(LogServiceRegisterLogRequestProto.newBuilder()
                                     .setLogname(LogServiceProtoUtil.toLogNameProto(name))
                                     .setRaftGroup(MetaServiceProtoUtil
                                             .toRaftGroupProto(raftGroup)))
                             .build().toByteString());
                 } catch (IOException e) {
+                    LOG.error(
+                        "Exception while registering raft group with Metadata Service during creation of log");
                     e.printStackTrace();
                 }
                 return CompletableFuture.completedFuture(Message.valueOf(MetaServiceProtoUtil
@@ -369,12 +382,11 @@ public class MetaStateMachine extends BaseStateMachine {
 
 
     class PeerGroups implements Comparable{
-        RaftPeer peer;
-        Set<RaftGroup> groups = new HashSet<>();
+        private RaftPeer peer;
+        private Set<RaftGroup> groups = new HashSet<>();
 
-        public PeerGroups(RaftPeer peer) {
+        PeerGroups(RaftPeer peer) {
             this.peer = peer;
-
         }
 
         public Set<RaftGroup> getGroups () {
@@ -390,4 +402,90 @@ public class MetaStateMachine extends BaseStateMachine {
             return groups.size() - ((PeerGroups) o).groups.size();
         }
     }
+
+    private class PeerHealthChecker implements Runnable {
+        @Override
+        public void run() {
+            while(true) {
+                try {
+                    Thread.sleep(1000);
+                    long now = System.currentTimeMillis();
+                    heartbeatInfo.keySet().stream().forEach(raftPeer -> {
+                        Long heartbeatTimestamp = heartbeatInfo.get(raftPeer);
+                        // Introduce configuration for period to detect the failure.
+                        if((now - heartbeatTimestamp) > failureDetectionPeriod) {
+                            // Close the logs serve by peer if any.
+                            if (peerLogs.containsKey(raftPeer)) {
+                                LOG.warn("Closing all logs hosted by peer {} because last heartbeat ({}ms) exceeds " +
+                                    "the threshold ({}ms)", raftPeer, now - heartbeatTimestamp, failureDetectionPeriod);
+                                peers.remove(raftPeer);
+                                Set<LogName> logNames = peerLogs.get(raftPeer);
+                                Iterator<LogName> itr = logNames.iterator();
+                                while(itr.hasNext()) {
+                                    LogName logName = itr.next();
+                                    RaftGroup group = map.get(logName);
+                                    RaftClient client = RaftClient.newBuilder().
+                                            setRaftGroup(group).setProperties(properties).build();
+                                    try {
+                                        LOG.warn(String.format("Peer %s in the group %s went down." +
+                                                        " Hence closing the log %s serve by the group.",
+                                                raftPeer.toString(), group.toString(), logName.toString()));
+                                        RaftClientReply reply = client.send(
+                                                () -> LogServiceProtoUtil.
+                                                        toChangeStateRequestProto(logName, LogStream.State.CLOSED, true)
+                                                        .toByteString());
+                                        LogServiceProtos.ChangeStateReplyProto message =
+                                                LogServiceProtos.ChangeStateReplyProto.parseFrom(
+                                                    reply.getMessage().getContent());
+                                        if(message.hasException()) {
+                                            throw new IOException(message.getException().getErrorMsg());
+                                        }
+                                        itr.remove();
+                                        client.close();
+                                    } catch (IOException e) {
+                                        LOG.warn(String.format("Failed to close log %s on peer %s failure.",
+                                                logName, raftPeer.toString()), e);
+                                    }
+                                }
+                                if(logNames.isEmpty()) {
+                                    peerLogs.remove(raftPeer);
+                                    heartbeatInfo.remove(raftPeer);
+                                } // else retry closing failed logs on next period.
+                            }
+                            final List<PeerGroups> peerGroupsToRemove = new ArrayList<>();
+                            // remove peer groups from avail.
+                            avail.stream().forEach(peerGroup -> {
+                                if(peerGroup.getPeer().equals(raftPeer)) {
+                                    peerGroupsToRemove.add(peerGroup);
+                                }
+                            });
+                            for(PeerGroups peerGroups: peerGroupsToRemove) {
+                                avail.remove(peerGroups);
+                            }
+                        }
+                    });
+                } catch (Exception e) {
+                    LOG.error(
+                            "Exception while closing logs and removing peer" +
+                                    " from raft groups with Metadata Service on node failure", e);
+                }
+            }
+        }
+    }
+
+
+    // This method need to be used for testing only.
+    public boolean checkPeersAreSame() {
+        if(!peers.equals(peerLogs.keySet()) || !peers.equals(heartbeatInfo.keySet())) {
+            return false;
+        }
+        Set<RaftPeer> availPeers = new HashSet<>();
+        avail.stream().forEach(peerGroups -> {
+            availPeers.add(peerGroups.getPeer());
+        });
+        if(!peers.equals(availPeers)) {
+            return false;
+        }
+        return true;
+    }
 }
