@@ -20,9 +20,9 @@ package org.apache.ratis.logservice.server;
 
 import com.beust.jcommander.JCommander;
 import org.apache.ratis.conf.RaftProperties;
-import org.apache.ratis.grpc.GrpcConfigKeys;
+import org.apache.ratis.logservice.api.LogServiceConfiguration;
+import org.apache.ratis.logservice.common.Constants;
 import org.apache.ratis.logservice.util.LogServiceUtils;
-import org.apache.ratis.netty.NettyConfigKeys;
 import org.apache.ratis.protocol.*;
 import org.apache.ratis.server.RaftServer;
 import org.apache.ratis.server.RaftServerConfigKeys;
@@ -55,6 +55,10 @@ public class MetadataServer extends BaseServer {
 
     private LifeCycle lifeCycle;
 
+    public StateMachine getMetaStateMachine() {
+      return metaStateMachine;
+    }
+
     public MetadataServer(ServerOpts opts) {
       super(opts);
       LOG.debug("Metadata Server options: {}", opts);
@@ -70,14 +74,30 @@ public class MetadataServer extends BaseServer {
         this.lifeCycle = new LifeCycle(this.id);
         RaftProperties properties = new RaftProperties();
         if(opts.getWorkingDir() != null) {
-            RaftServerConfigKeys.setStorageDirs(properties, Collections.singletonList(new File(opts.getWorkingDir())));
+            RaftServerConfigKeys.setStorageDirs(properties,
+              Collections.singletonList(new File(opts.getWorkingDir())));
+        } else {
+          String path = getConfig().get(Constants.META_SERVER_WORKDIR_KEY);
+          if (path != null) {
+            RaftServerConfigKeys.setStorageDirs(properties,
+              Collections.singletonList(new File(path)));
+          }
         }
-        GrpcConfigKeys.Server.setPort(properties, opts.getPort());
-        NettyConfigKeys.Server.setPort(properties, opts.getPort());
+
+        // Set properties common to all log service state machines
+        setRaftProperties(properties);
+        long failureDetectionPeriod = getConfig().
+                getLong(Constants.LOG_SERVICE_PEER_FAILURE_DETECTION_PERIOD_KEY,
+                Constants.DEFAULT_PEER_FAILURE_DETECTION_PERIOD);
         Set<RaftPeer> peers = getPeersFromQuorum(opts.getMetaQuorum());
         RaftGroupId raftMetaGroupId = RaftGroupId.valueOf(opts.getMetaGroupId());
         RaftGroup metaGroup = RaftGroup.valueOf(raftMetaGroupId, peers);
-        metaStateMachine = new MetaStateMachine(raftMetaGroupId, RaftGroupId.valueOf(opts.getLogServerGroupId()));
+        metaStateMachine = new MetaStateMachine(raftMetaGroupId, RaftGroupId.valueOf(opts.getLogServerGroupId()),
+                failureDetectionPeriod);
+
+        // Make sure that we aren't setting any invalid/harmful properties
+        validateRaftProperties(properties);
+
         server = RaftServer.newBuilder()
                 .setGroup(metaGroup)
                 .setServerId(RaftPeerId.valueOf(id))
@@ -99,6 +119,9 @@ public class MetadataServer extends BaseServer {
                 .addObject(opts)
                 .build()
                 .parse(args);
+        // Add config from log service configuration file
+        LogServiceConfiguration config = LogServiceConfiguration.create();
+        opts = config.addMetaServerOpts(opts);
 
         try (MetadataServer master = new MetadataServer(opts)) {
           master.start();
@@ -143,4 +166,8 @@ public class MetadataServer extends BaseServer {
             return new MetadataServer(getOpts());
         }
     }
+
+    public RaftServer getServer() {
+        return server;
+    }
 }
