@@ -22,15 +22,17 @@ import java.util.HashMap;
 import java.util.Map;
 import java.util.Objects;
 
+import org.apache.ratis.logservice.shell.commands.ArchiveLogCommand;
 import org.apache.ratis.logservice.shell.commands.CreateLogCommand;
 import org.apache.ratis.logservice.shell.commands.DeleteLogCommand;
 import org.apache.ratis.logservice.shell.commands.ExitCommand;
+import org.apache.ratis.logservice.shell.commands.ExportLogCommand;
 import org.apache.ratis.logservice.shell.commands.HelpCommand;
 import org.apache.ratis.logservice.shell.commands.ListLogsCommand;
 import org.apache.ratis.logservice.shell.commands.PutToLogCommand;
 import org.apache.ratis.logservice.shell.commands.ReadLogCommand;
 
-public class CommandFactory {
+public final class CommandFactory {
   private static final Map<String,Command> KNOWN_COMMANDS = cacheCommands();
 
   private static Map<String,Command> cacheCommands() {
@@ -46,6 +48,8 @@ public class CommandFactory {
     commands.put("quit", exitCommand);
     commands.put("help", new HelpCommand());
     commands.put("list", new ListLogsCommand());
+    commands.put("archive", new ArchiveLogCommand());
+    commands.put("export", new ExportLogCommand());
 
     return Collections.unmodifiableMap(commands);
   }
