@@ -36,7 +36,9 @@ import java.util.stream.Collectors;
 
 import static org.apache.ratis.logservice.util.LogServiceProtoUtil.toLogNameProto;
 
-public class MetaServiceProtoUtil {
+public final class MetaServiceProtoUtil {
+    private MetaServiceProtoUtil() {
+    }
 
     public static RaftPeerProto toRaftPeerProto(RaftPeer peer) {
         RaftPeerProto.Builder builder = RaftPeerProto.newBuilder()
@@ -81,6 +83,15 @@ public class MetaServiceProtoUtil {
                                 .setPeer(MetaServiceProtoUtil.toRaftPeerProto(peer)).build()).build();
     }
 
+    public static MetaSMRequestProto toHeartbeatRequestProto(RaftPeer peer) {
+        return MetaServiceProtos.MetaSMRequestProto
+                .newBuilder()
+                .setHeartbeatRequest(
+                        MetaServiceProtos.LogServiceHeartbeatRequestProto
+                                .newBuilder()
+                                .setPeer(MetaServiceProtoUtil.toRaftPeerProto(peer)).build()).build();
+    }
+
     public static MetaServiceRequestProto toCreateLogRequestProto(LogName logName) {
         LogServiceProtos.LogNameProto logNameProto = LogServiceProtos.LogNameProto.newBuilder()
                 .setName(logName.getName())
@@ -101,15 +112,6 @@ public class MetaServiceProtoUtil {
         return MetaServiceRequestProto.newBuilder().setGetLog(getLog).build();
     }
 
-    public static MetaServiceRequestProto toArchiveLogRequestProto(LogName logName) {
-        LogServiceProtos.LogNameProto logNameProto = LogServiceProtos.LogNameProto.newBuilder()
-                .setName(logName.getName())
-                .build();
-        ArchiveLogRequestProto archiveLog =
-                ArchiveLogRequestProto.newBuilder().setLogName(logNameProto).build();
-        return MetaServiceRequestProto.newBuilder().setArchiveLog(archiveLog).build();
-    }
-
     public static MetaServiceRequestProto toDeleteLogRequestProto(LogName logName) {
         LogServiceProtos.LogNameProto logNameProto = LogServiceProtos.LogNameProto.newBuilder()
                 .setName(logName.getName())
@@ -188,7 +190,7 @@ public class MetaServiceProtoUtil {
     static Iterable<RaftPeerProto> toRaftPeerProtos(
             final Collection<RaftPeer> peers) {
         return () -> new Iterator<RaftPeerProto>() {
-            final Iterator<RaftPeer> i = peers.iterator();
+            private final Iterator<RaftPeer> i = peers.iterator();
 
             @Override
             public boolean hasNext() {
@@ -202,7 +204,6 @@ public class MetaServiceProtoUtil {
         };
     }
 
-
     public static DeleteLogReplyProto toDeleteLogReplyProto() {
         return DeleteLogReplyProto.newBuilder().build();
     }
