@@ -21,14 +21,18 @@ import static org.junit.Assert.assertEquals;
 import static org.junit.Assert.assertTrue;
 
 import java.io.IOException;
+import java.lang.management.ManagementFactory;
 import java.nio.ByteBuffer;
 import java.nio.charset.StandardCharsets;
 import java.util.ArrayList;
+import java.util.concurrent.TimeUnit;
 import java.util.stream.Collectors;
 import java.util.stream.LongStream;
 import java.util.Iterator;
 import java.util.List;
 
+import javax.management.ObjectName;
+
 import org.apache.ratis.BaseTest;
 import org.apache.ratis.MiniRaftCluster;
 import org.apache.ratis.RaftTestUtil;
@@ -40,9 +44,12 @@ import org.apache.ratis.logservice.api.LogStream;
 import org.apache.ratis.logservice.api.LogStream.State;
 import org.apache.ratis.logservice.api.LogWriter;
 import org.apache.ratis.logservice.impl.LogStreamImpl;
+import org.apache.ratis.logservice.metrics.LogServiceMetricsRegistry;
 import org.apache.ratis.logservice.server.LogStateMachine;
 import org.apache.ratis.logservice.util.TestUtils;
+import org.apache.ratis.metrics.JVMMetrics;
 import org.apache.ratis.statemachine.StateMachine;
+import org.apache.ratis.util.TimeDuration;
 import org.junit.After;
 import org.junit.Before;
 import org.junit.Test;
@@ -54,6 +61,10 @@ public abstract class LogServiceReadWriteBase<CLUSTER extends MiniRaftCluster>
     implements MiniRaftCluster.Factory.Get<CLUSTER> {
   public static final Logger LOG = LoggerFactory.getLogger(LogServiceReadWriteBase.class);
 
+  static {
+    JVMMetrics.initJvmMetrics(TimeDuration.valueOf(10, TimeUnit.SECONDS));
+  }
+
   {
     final RaftProperties p = getProperties();
     p.setClass(MiniRaftCluster.STATEMACHINE_CLASS_KEY,
@@ -64,9 +75,50 @@ public abstract class LogServiceReadWriteBase<CLUSTER extends MiniRaftCluster>
   static final int NUM_PEERS = 3;
   CLUSTER cluster;
 
+  class MetricLogStream extends LogStreamImpl{
+    Long startRecordIdCount = 0l;
+    Long getStateCount = 0l;
+    Long getLastRecordIdCount = 0l;
+    Long getLengthCount = 0l;
+    Long getSizeCount = 0l;
+
+    public MetricLogStream(LogName name, RaftClient raftClient) {
+      super(name, raftClient);
+    }
+
+    @Override public long getStartRecordId() throws IOException {
+      startRecordIdCount++;
+      return super.getStartRecordId();
+    }
+
+    @Override public State getState() throws IOException {
+      getStateCount++;
+      return super.getState();
+    }
+
+    @Override public long getLastRecordId() throws IOException {
+      getLastRecordIdCount++;
+      return super.getLastRecordId();
+    }
+
+    @Override public long getLength() throws IOException {
+      getLengthCount++;
+      return super.getLength();
+    }
+
+    @Override public long getSize() throws IOException {
+      getSizeCount++;
+      return super.getSize();
+    }
+
+    @Override public LogName getName() {
+      return super.getName();
+    }
+  }
   @Before
   public void setUpCluster() throws IOException, InterruptedException {
-    cluster = newCluster(NUM_PEERS);
+    RaftProperties raftProperties = getProperties();
+    cluster = getFactory().newCluster(NUM_PEERS, raftProperties);
     cluster.start();
     RaftTestUtil.waitForLeader(cluster);
   }
@@ -78,11 +130,12 @@ public abstract class LogServiceReadWriteBase<CLUSTER extends MiniRaftCluster>
             .build();
     LogName logName = LogName.of("log1");
     // TODO need API to circumvent metadata service for testing
-    try (LogStream logStream = new LogStreamImpl(logName, raftClient)) {
+    try (LogStream logStream = new MetricLogStream(logName, raftClient)) {
       assertEquals("log1", logStream.getName().getName());
       assertEquals(State.OPEN, logStream.getState());
       assertEquals(0, logStream.getSize());
       assertEquals(0, logStream.getLength());
+      testJMXMetrics(logStream);
 
       LogReader reader = logStream.createReader();
       LogWriter writer = logStream.createWriter();
@@ -120,6 +173,9 @@ public abstract class LogServiceReadWriteBase<CLUSTER extends MiniRaftCluster>
         ByteBuffer actual = actualIter.next();
         assertEquals(expected, actual);
       }
+      testJMXMetrics(logStream);
+      assertEquals(logStream.getState(),State.OPEN);
+
     }
   }
 
@@ -131,7 +187,7 @@ public abstract class LogServiceReadWriteBase<CLUSTER extends MiniRaftCluster>
     final LogName logName = LogName.of("log1");
     final int numRecords = 25;
     // TODO need API to circumvent metadata service for testing
-    try (LogStream logStream = new LogStreamImpl(logName, raftClient)) {
+    try (LogStream logStream = new MetricLogStream(logName, raftClient)) {
       try (LogWriter writer = logStream.createWriter()) {
         LOG.info("Writing {} records", numRecords);
         // Write records 0 through 99 (inclusive)
@@ -172,7 +228,7 @@ public abstract class LogServiceReadWriteBase<CLUSTER extends MiniRaftCluster>
     final LogName logName = LogName.of("log1");
     final int numRecords = 100;
     // TODO need API to circumvent metadata service for testing
-    try (LogStream logStream = new LogStreamImpl(logName, raftClient)) {
+    try (LogStream logStream = new MetricLogStream(logName, raftClient)) {
       try (LogWriter writer = logStream.createWriter()) {
         LOG.info("Writing {} records", numRecords);
         // Write records 0 through 99 (inclusive)
@@ -208,7 +264,7 @@ public abstract class LogServiceReadWriteBase<CLUSTER extends MiniRaftCluster>
             .build();
     final LogName logName = LogName.of("log1");
     final int numRecords = 10;
-    try (LogStream logStream = new LogStreamImpl(logName, raftClient)) {
+    try (LogStream logStream = new MetricLogStream(logName, raftClient)) {
       final List<Long> recordIds;
       try (LogWriter writer = logStream.createWriter()) {
         LOG.info("Writing {} records", numRecords);
@@ -258,4 +314,20 @@ public abstract class LogServiceReadWriteBase<CLUSTER extends MiniRaftCluster>
     System.arraycopy(bb.array(), bb.arrayOffset(), bytes, 0, bb.remaining());
     return Integer.parseInt(new String(bytes, StandardCharsets.UTF_8));
   }
+
+  private void testJMXMetrics(LogStream logStream) throws Exception {
+    assertEquals(((MetricLogStream) logStream).getLengthCount,
+        getJMXCount(cluster.getGroup().getGroupId().toString(), "lengthQueryTime"));
+    assertEquals(((MetricLogStream) logStream).getSizeCount,
+        getJMXCount(cluster.getGroup().getGroupId().toString(), "sizeRequestTime"));
+
+  }
+
+  private Long getJMXCount(String groupId, String metricName) throws Exception {
+    ObjectName oname = new ObjectName(LogServiceMetricsRegistry.RATIS_LOG_SERVICE_METRICS, "name",
+        LogServiceMetricsRegistry
+            .getMetricRegistryForLogService(groupId, cluster.getLeader().getId().toString())
+            .getMetricRegistryInfo().getName() + "." + metricName);
+    return (Long) ManagementFactory.getPlatformMBeanServer().getAttribute(oname, "Count");
+  }
 }
