@@ -26,8 +26,10 @@ import org.slf4j.LoggerFactory;
 import java.io.File;
 import java.util.Collections;
 import java.util.List;
+import java.util.Optional;
 import java.util.concurrent.TimeUnit;
 import java.util.function.Consumer;
+import java.util.function.Function;
 
 import static org.apache.ratis.conf.ConfUtils.*;
 
@@ -48,6 +50,16 @@ public interface RaftServerConfigKeys {
     setFiles(properties::setFiles, STORAGE_DIR_KEY, storageDir);
   }
 
+  String SLEEP_DEVIATION_THRESHOLD_KEY = PREFIX + ".sleep.deviation.threshold";
+  int SLEEP_DEVIATION_THRESHOLD_DEFAULT = 300;
+  static int sleepDeviationThreshold(RaftProperties properties) {
+    return getInt(properties::getInt, SLEEP_DEVIATION_THRESHOLD_KEY,
+        SLEEP_DEVIATION_THRESHOLD_DEFAULT, getDefaultLog());
+  }
+  static void setSleepDeviationThreshold(RaftProperties properties, int thresholdMs) {
+    setInt(properties::setInt, SLEEP_DEVIATION_THRESHOLD_KEY, thresholdMs);
+  }
+
   /**
    * When bootstrapping a new peer, If the gap between the match index of the
    * peer and the leader's latest committed index is less than this gap, we
@@ -63,42 +75,63 @@ public interface RaftServerConfigKeys {
     setInt(properties::setInt, STAGING_CATCHUP_GAP_KEY, stagingCatchupGap);
   }
 
-  /**
-   * Timeout for leader election after which the statemachine of the server is notified
-   * about leader election pending for a long time.
-   */
-  String LEADER_ELECTION_TIMEOUT_KEY = PREFIX + ".leader.election.timeout";
-  TimeDuration LEADER_ELECTION_TIMEOUT_DEFAULT = TimeDuration.valueOf(60, TimeUnit.SECONDS);
-  static TimeDuration leaderElectionTimeout(RaftProperties properties) {
-    return getTimeDuration(properties.getTimeDuration(LEADER_ELECTION_TIMEOUT_DEFAULT.getUnit()),
-        LEADER_ELECTION_TIMEOUT_KEY, LEADER_ELECTION_TIMEOUT_DEFAULT, getDefaultLog());
-  }
-  static void setLeaderElectionTimeout(RaftProperties properties, TimeDuration leaderElectionTimeout) {
-    setTimeDuration(properties::setTimeDuration, LEADER_ELECTION_TIMEOUT_KEY, leaderElectionTimeout);
 
+  interface Write {
+    String PREFIX = RaftServerConfigKeys.PREFIX + ".write";
+
+    String ELEMENT_LIMIT_KEY = PREFIX + ".element-limit";
+    int ELEMENT_LIMIT_DEFAULT = 4096;
+
+    static int elementLimit(RaftProperties properties) {
+      return getInt(properties::getInt, ELEMENT_LIMIT_KEY, ELEMENT_LIMIT_DEFAULT, getDefaultLog(), requireMin(1));
+    }
+    static void setElementLimit(RaftProperties properties, int limit) {
+      setInt(properties::setInt, ELEMENT_LIMIT_KEY, limit, requireMin(1));
+    }
+
+    String BYTE_LIMIT_KEY = PREFIX + ".byte-limit";
+    SizeInBytes BYTE_LIMIT_DEFAULT = SizeInBytes.valueOf("64MB");
+    static SizeInBytes byteLimit(RaftProperties properties) {
+      return getSizeInBytes(properties::getSizeInBytes,
+          BYTE_LIMIT_KEY, BYTE_LIMIT_DEFAULT, getDefaultLog());
+    }
+    static void setByteLimit(RaftProperties properties, int byteLimit) {
+      setInt(properties::setInt, BYTE_LIMIT_KEY, byteLimit, requireMin(1));
+    }
   }
 
-  String WATCH_TIMEOUT_DENOMINATION_KEY = PREFIX + ".watch.timeout.denomination";
-  TimeDuration WATCH_TIMEOUT_DENOMINATION_DEFAULT = TimeDuration.valueOf(1, TimeUnit.SECONDS);
-  static TimeDuration watchTimeoutDenomination(RaftProperties properties) {
-    return getTimeDuration(properties.getTimeDuration(WATCH_TIMEOUT_DENOMINATION_DEFAULT.getUnit()),
-        WATCH_TIMEOUT_DENOMINATION_KEY, WATCH_TIMEOUT_DENOMINATION_DEFAULT, getDefaultLog(), requirePositive());
-  }
-  static void setWatchTimeoutDenomination(RaftProperties properties, TimeDuration watchTimeout) {
-    setTimeDuration(properties::setTimeDuration, WATCH_TIMEOUT_DENOMINATION_KEY, watchTimeout);
-  }
+  interface Watch {
+    String PREFIX = RaftServerConfigKeys.PREFIX + ".watch";
 
-  /**
-   * Timeout for watch requests.
-   */
-  String WATCH_TIMEOUT_KEY = PREFIX + ".watch.timeout";
-  TimeDuration WATCH_TIMEOUT_DEFAULT = TimeDuration.valueOf(10, TimeUnit.SECONDS);
-  static TimeDuration watchTimeout(RaftProperties properties) {
-    return getTimeDuration(properties.getTimeDuration(WATCH_TIMEOUT_DEFAULT.getUnit()),
-        WATCH_TIMEOUT_KEY, WATCH_TIMEOUT_DEFAULT, getDefaultLog(), requirePositive());
-  }
-  static void setWatchTimeout(RaftProperties properties, TimeDuration watchTimeout) {
-    setTimeDuration(properties::setTimeDuration, WATCH_TIMEOUT_KEY, watchTimeout);
+    String ELEMENT_LIMIT_KEY = PREFIX + ".element-limit";
+    int ELEMENT_LIMIT_DEFAULT = 65536;
+    static int elementLimit(RaftProperties properties) {
+      return getInt(properties::getInt, ELEMENT_LIMIT_KEY, ELEMENT_LIMIT_DEFAULT, getDefaultLog(), requireMin(1));
+    }
+    static void setElementLimit(RaftProperties properties, int limit) {
+      setInt(properties::setInt, ELEMENT_LIMIT_KEY, limit, requireMin(1));
+    }
+
+    String TIMEOUT_DENOMINATION_KEY = PREFIX + ".timeout.denomination";
+    TimeDuration TIMEOUT_DENOMINATION_DEFAULT = TimeDuration.valueOf(1, TimeUnit.SECONDS);
+    static TimeDuration timeoutDenomination(RaftProperties properties) {
+      return getTimeDuration(properties.getTimeDuration(TIMEOUT_DENOMINATION_DEFAULT.getUnit()),
+          TIMEOUT_DENOMINATION_KEY, TIMEOUT_DENOMINATION_DEFAULT, getDefaultLog(), requirePositive());
+    }
+    static void setTimeoutDenomination(RaftProperties properties, TimeDuration watchTimeout) {
+      setTimeDuration(properties::setTimeDuration, TIMEOUT_DENOMINATION_KEY, watchTimeout);
+    }
+
+    /** Timeout for watch requests. */
+    String TIMEOUT_KEY = PREFIX + ".timeout";
+    TimeDuration TIMEOUT_DEFAULT = TimeDuration.valueOf(10, TimeUnit.SECONDS);
+    static TimeDuration timeout(RaftProperties properties) {
+      return getTimeDuration(properties.getTimeDuration(TIMEOUT_DEFAULT.getUnit()),
+          TIMEOUT_KEY, TIMEOUT_DEFAULT, getDefaultLog(), requirePositive());
+    }
+    static void setTimeout(RaftProperties properties, TimeDuration watchTimeout) {
+      setTimeDuration(properties::setTimeDuration, TIMEOUT_KEY, watchTimeout);
+    }
   }
 
   interface Log {
@@ -119,7 +152,7 @@ public interface RaftServerConfigKeys {
       return getInt(properties::getInt, QUEUE_ELEMENT_LIMIT_KEY, QUEUE_ELEMENT_LIMIT_DEFAULT, getDefaultLog(),
           requireMin(1));
     }
-    static void setElementLimit(RaftProperties properties, int queueSize) {
+    static void setQueueElementLimit(RaftProperties properties, int queueSize) {
       setInt(properties::setInt, QUEUE_ELEMENT_LIMIT_KEY, queueSize, requireMin(1));
     }
 
@@ -129,10 +162,19 @@ public interface RaftServerConfigKeys {
       return getSizeInBytes(properties::getSizeInBytes,
           QUEUE_BYTE_LIMIT_KEY, QUEUE_BYTE_LIMIT_DEFAULT, getDefaultLog());
     }
-    static void setByteLimit(RaftProperties properties, int queueSize) {
+    static void setQueueByteLimit(RaftProperties properties, int queueSize) {
       setInt(properties::setInt, QUEUE_BYTE_LIMIT_KEY, queueSize, requireMin(1));
     }
 
+    String PURGE_GAP_KEY = PREFIX + ".purge.gap";
+    int PURGE_GAP_DEFAULT = 1024;
+    static int purgeGap(RaftProperties properties) {
+      return getInt(properties::getInt, PURGE_GAP_KEY, PURGE_GAP_DEFAULT, getDefaultLog(), requireMin(1));
+    }
+    static void setPurgeGap(RaftProperties properties, int purgeGap) {
+      setInt(properties::setInt, PURGE_GAP_KEY, purgeGap, requireMin(1));
+    }
+
     String SEGMENT_SIZE_MAX_KEY = PREFIX + ".segment.size.max";
     SizeInBytes SEGMENT_SIZE_MAX_DEFAULT = SizeInBytes.valueOf("8MB");
     static SizeInBytes segmentSizeMax(RaftProperties properties) {
@@ -186,6 +228,32 @@ public interface RaftServerConfigKeys {
       setInt(properties::setInt, FORCE_SYNC_NUM_KEY, forceSyncNum);
     }
 
+    /** The policy to handle corrupted raft log. */
+    enum CorruptionPolicy {
+      /** Rethrow the exception. */
+      EXCEPTION,
+      /** Print a warn log message and return all uncorrupted log entries up to the corruption. */
+      WARN_AND_RETURN;
+
+      public static CorruptionPolicy getDefault() {
+        return EXCEPTION;
+      }
+
+      public static <T> CorruptionPolicy get(T supplier, Function<T, CorruptionPolicy> getMethod) {
+        return Optional.ofNullable(supplier).map(getMethod).orElse(getDefault());
+      }
+    }
+
+    String CORRUPTION_POLICY_KEY = PREFIX + ".corruption.policy";
+    CorruptionPolicy CORRUPTION_POLICY_DEFAULT = CorruptionPolicy.getDefault();
+    static CorruptionPolicy corruptionPolicy(RaftProperties properties) {
+      return get(properties::getEnum,
+          CORRUPTION_POLICY_KEY, CORRUPTION_POLICY_DEFAULT, getDefaultLog());
+    }
+    static void setCorruptionPolicy(RaftProperties properties, CorruptionPolicy corruptionPolicy) {
+      set(properties::setEnum, CORRUPTION_POLICY_KEY, corruptionPolicy);
+    }
+
     interface StateMachineData {
       String PREFIX = Log.PREFIX + ".statemachine.data";
 
@@ -267,6 +335,16 @@ public interface RaftServerConfigKeys {
       static void setSnapshotChunkSizeMax(RaftProperties properties, SizeInBytes maxChunkSize) {
         setSizeInBytes(properties::set, SNAPSHOT_CHUNK_SIZE_MAX_KEY, maxChunkSize);
       }
+
+      String INSTALL_SNAPSHOT_ENABLED_KEY = PREFIX + ".install.snapshot.enabled";
+      boolean INSTALL_SNAPSHOT_ENABLED_DEFAULT = true;
+      static boolean installSnapshotEnabled(RaftProperties properties) {
+        return getBoolean(properties::getBoolean,
+            INSTALL_SNAPSHOT_ENABLED_KEY, INSTALL_SNAPSHOT_ENABLED_DEFAULT, getDefaultLog());
+      }
+      static void setInstallSnapshotEnabled(RaftProperties properties, boolean shouldInstallSnapshot) {
+        setBoolean(properties::setBoolean, INSTALL_SNAPSHOT_ENABLED_KEY, shouldInstallSnapshot);
+      }
     }
   }
 
@@ -295,6 +373,15 @@ public interface RaftServerConfigKeys {
     static void setAutoTriggerThreshold(RaftProperties properties, long autoTriggerThreshold) {
       setLong(properties::setLong, AUTO_TRIGGER_THRESHOLD_KEY, autoTriggerThreshold);
     }
+
+    String RETENTION_FILE_NUM_KEY = PREFIX + ".retention.file.num";
+    int RETENTION_FILE_NUM_DEFAULT = -1;
+    static int retentionFileNum(RaftProperties raftProperties) {
+      return getInt(raftProperties::getInt, RETENTION_FILE_NUM_KEY, RETENTION_FILE_NUM_DEFAULT, getDefaultLog());
+    }
+    static void setRetentionFileNum(RaftProperties properties, int numSnapshotFilesRetained) {
+      setInt(properties::setInt, RETENTION_FILE_NUM_KEY, numSnapshotFilesRetained);
+    }
   }
 
   /** server rpc timeout related */
@@ -367,6 +454,21 @@ public interface RaftServerConfigKeys {
     }
   }
 
+  interface Notification {
+    String PREFIX = RaftServerConfigKeys.PREFIX + "." + Notification.class.getSimpleName().toLowerCase();
+
+    /** Timeout value to notify the state machine when there is no leader. */
+    String NO_LEADER_TIMEOUT_KEY = PREFIX + ".no-leader.timeout";
+    TimeDuration NO_LEADER_TIMEOUT_DEFAULT = TimeDuration.valueOf(60, TimeUnit.SECONDS);
+    static TimeDuration noLeaderTimeout(RaftProperties properties) {
+      return getTimeDuration(properties.getTimeDuration(NO_LEADER_TIMEOUT_DEFAULT.getUnit()),
+          NO_LEADER_TIMEOUT_KEY, NO_LEADER_TIMEOUT_DEFAULT, getDefaultLog());
+    }
+    static void setNoLeaderTimeout(RaftProperties properties, TimeDuration leaderElectionTimeout) {
+      setTimeDuration(properties::setTimeDuration, NO_LEADER_TIMEOUT_KEY, leaderElectionTimeout);
+    }
+  }
+
   static void main(String[] args) {
     printAll(RaftServerConfigKeys.class);
   }
