@@ -17,9 +17,10 @@
  */
 package org.apache.ratis.server.impl;
 
+import org.apache.ratis.protocol.RaftGroupMemberId;
 import org.apache.ratis.protocol.RaftPeer;
-import org.apache.ratis.protocol.RaftPeerId;
-import org.apache.ratis.server.storage.RaftLogIndex;
+import org.apache.ratis.server.raftlog.RaftLog;
+import org.apache.ratis.server.raftlog.RaftLogIndex;
 import org.apache.ratis.util.Timestamp;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
@@ -39,12 +40,11 @@ public class FollowerInfo {
   private final AtomicReference<Timestamp> lastRpcSendTime;
   private final RaftLogIndex nextIndex;
   private final RaftLogIndex matchIndex = new RaftLogIndex("matchIndex", 0L);
-  private final RaftLogIndex commitIndex = new RaftLogIndex("commitIndex", RaftServerConstants.INVALID_LOG_INDEX);
+  private final RaftLogIndex commitIndex = new RaftLogIndex("commitIndex", RaftLog.INVALID_LOG_INDEX);
   private volatile boolean attendVote;
   private final int rpcSlownessTimeoutMs;
 
-
-  FollowerInfo(RaftPeerId id, RaftPeer peer, Timestamp lastRpcTime, long nextIndex,
+  FollowerInfo(RaftGroupMemberId id, RaftPeer peer, Timestamp lastRpcTime, long nextIndex,
       boolean attendVote, int rpcSlownessTimeoutMs) {
     this.name = id + "->" + peer.getId();
     this.infoIndexChange = s -> LOG.info("{}: {}", name, s);
@@ -62,8 +62,8 @@ public class FollowerInfo {
     return matchIndex.get();
   }
 
-  public void updateMatchIndex(long newMatchIndex) {
-    matchIndex.updateIncreasingly(newMatchIndex, debugIndexChange);
+  public boolean updateMatchIndex(long newMatchIndex) {
+    return matchIndex.updateToMax(newMatchIndex, debugIndexChange);
   }
 
   /** @return the commit index acked by the follower. */
@@ -79,7 +79,7 @@ public class FollowerInfo {
     return nextIndex.get();
   }
 
-  public void updateNextIndex(long newNextIndex) {
+  public void increaseNextIndex(long newNextIndex) {
     nextIndex.updateIncreasingly(newNextIndex, debugIndexChange);
   }
 
@@ -87,11 +87,19 @@ public class FollowerInfo {
     nextIndex.updateUnconditionally(old -> old <= 0L? old: Math.min(old - 1, newNextIndex), infoIndexChange);
   }
 
+  public void updateNextIndex(long newNextIndex) {
+    nextIndex.updateUnconditionally(old -> newNextIndex >= 0 ? newNextIndex : old, infoIndexChange);
+  }
+
   public void setSnapshotIndex(long snapshotIndex) {
     matchIndex.setUnconditionally(snapshotIndex, infoIndexChange);
     nextIndex.setUnconditionally(snapshotIndex + 1, infoIndexChange);
   }
 
+  public String getName() {
+    return name;
+  }
+
   @Override
   public String toString() {
     return name + "(c" + getCommitIndex() + ",m" + getMatchIndex() + ",n" + getNextIndex()
@@ -104,7 +112,7 @@ public class FollowerInfo {
     attendVote = true;
   }
 
-  public boolean isAttendingVote() {
+  boolean isAttendingVote() {
     return attendVote;
   }
 
@@ -117,7 +125,7 @@ public class FollowerInfo {
     lastRpcResponseTime.set(Timestamp.currentTime());
   }
 
-  public Timestamp getLastRpcResponseTime() {
+  Timestamp getLastRpcResponseTime() {
     return lastRpcResponseTime.get();
   }
 
@@ -130,7 +138,7 @@ public class FollowerInfo {
     return Timestamp.latest(lastRpcResponseTime.get(), lastRpcSendTime.get());
   }
 
-  public boolean isSlow() {
+  boolean isSlow() {
     return lastRpcResponseTime.get().elapsedTimeMs() > rpcSlownessTimeoutMs;
   }
 }
