@@ -23,11 +23,11 @@ import org.apache.ratis.protocol.RaftPeer;
 import org.apache.ratis.protocol.RaftPeerId;
 import org.apache.ratis.server.protocol.TermIndex;
 import org.apache.ratis.statemachine.SnapshotInfo;
+import org.apache.ratis.thirdparty.com.google.common.annotations.VisibleForTesting;
 import org.apache.ratis.util.Daemon;
 import org.apache.ratis.util.LifeCycle;
 import org.apache.ratis.util.LogUtils;
 import org.apache.ratis.util.Preconditions;
-import org.apache.ratis.util.ProtoUtils;
 import org.apache.ratis.util.TimeDuration;
 import org.apache.ratis.util.Timestamp;
 import org.slf4j.Logger;
@@ -36,7 +36,9 @@ import org.slf4j.LoggerFactory;
 import java.io.IOException;
 import java.util.ArrayList;
 import java.util.Collection;
+import java.util.HashMap;
 import java.util.List;
+import java.util.Map;
 import java.util.concurrent.Callable;
 import java.util.concurrent.ExecutionException;
 import java.util.concurrent.ExecutorCompletionService;
@@ -46,14 +48,18 @@ import java.util.concurrent.Future;
 import java.util.concurrent.atomic.AtomicInteger;
 import java.util.stream.Collectors;
 
+import static org.apache.ratis.util.LifeCycle.State.NEW;
+import static org.apache.ratis.util.LifeCycle.State.RUNNING;
+import static org.apache.ratis.util.LifeCycle.State.STARTING;
+
 class LeaderElection implements Runnable {
   public static final Logger LOG = LoggerFactory.getLogger(LeaderElection.class);
 
   private ResultAndTerm logAndReturn(Result result,
-      List<RequestVoteReplyProto> responses,
+      Map<RaftPeerId, RequestVoteReplyProto> responses,
       List<Exception> exceptions, long newTerm) {
     LOG.info(this + ": Election " + result + "; received " + responses.size() + " response(s) "
-        + responses.stream().map(ProtoUtils::toString).collect(Collectors.toList())
+        + responses.values().stream().map(ServerProtoUtils::toString).collect(Collectors.toList())
         + " and " + exceptions.size() + " exception(s); " + server.getState());
     int i = 0;
     for(Exception e : exceptions) {
@@ -109,48 +115,68 @@ class LeaderElection implements Runnable {
   private final RaftServerImpl server;
 
   LeaderElection(RaftServerImpl server) {
-    this.name = server.getId() + ":" + server.getGroupId() + ":" + getClass().getSimpleName() + COUNT.incrementAndGet();
+    this.name = server.getMemberId() + "-" + getClass().getSimpleName() + COUNT.incrementAndGet();
     this.lifeCycle = new LifeCycle(this);
     this.daemon = new Daemon(this);
     this.server = server;
   }
 
   void start() {
-    lifeCycle.startAndTransition(daemon::start);
+    startIfNew(daemon::start);
+  }
+
+  @VisibleForTesting
+  void startInForeground() {
+    startIfNew(this);
+  }
+
+  private void startIfNew(Runnable starter) {
+    if (lifeCycle.compareAndTransition(NEW, STARTING)) {
+      starter.run();
+    } else {
+      final LifeCycle.State state = lifeCycle.getCurrentState();
+      LOG.info("{}: skip starting since this is already {}", this, state);
+    }
   }
 
   void shutdown() {
     lifeCycle.checkStateAndClose();
   }
 
+  @VisibleForTesting
+  LifeCycle.State getCurrentState() {
+    return lifeCycle.getCurrentState();
+  }
+
   @Override
   public void run() {
+    if (!lifeCycle.compareAndTransition(STARTING, RUNNING)) {
+      final LifeCycle.State state = lifeCycle.getCurrentState();
+      LOG.info("{}: skip running since this is already {}", this, state);
+      return;
+    }
+
+    Timestamp electionStartTime = Timestamp.currentTime();
     try {
       askForVotes();
-    } catch (InterruptedException e) {
-      // the leader election thread is interrupted. The peer may already step
-      // down to a follower. The leader election should skip.
-      LOG.info("{} thread is interrupted gracefully; server={}", this, server);
     } catch(Throwable e) {
       final LifeCycle.State state = lifeCycle.getCurrentState();
-      final String message = "Failed " + this + ", state=" + state;
-
       if (state.isClosingOrClosed()) {
-        LogUtils.infoOrTrace(LOG, message, e);
         LOG.info("{}: {} is safely ignored since this is already {}",
-            this, e.getClass().getSimpleName(), state);
+            this, e.getClass().getSimpleName(), state, e);
       } else {
         if (!server.isAlive()) {
-          LogUtils.infoOrTrace(LOG, message, e);
           LOG.info("{}: {} is safely ignored since the server is not alive: {}",
-              this, e.getClass().getSimpleName(), server);
+              this, e.getClass().getSimpleName(), server, e);
         } else {
-          LOG.error(message, e);
+          LOG.error("{}: Failed, state={}", this, state, e);
         }
         shutdown();
       }
     } finally {
-      lifeCycle.transition(LifeCycle.State.CLOSED);
+      // Update leader election completion metric(s).
+      server.getLeaderElectionMetrics().onLeaderElectionCompletion(electionStartTime.elapsedTimeMs());
+      lifeCycle.checkStateAndClose(() -> {});
     }
   }
 
@@ -242,7 +268,7 @@ class LeaderElection implements Runnable {
   private ResultAndTerm waitForResults(final long electionTerm, final int submitted,
       RaftConfiguration conf, Executor voteExecutor) throws InterruptedException {
     final Timestamp timeout = Timestamp.currentTime().addTimeMs(server.getRandomTimeoutMs());
-    final List<RequestVoteReplyProto> responses = new ArrayList<>();
+    final Map<RaftPeerId, RequestVoteReplyProto> responses = new HashMap<>();
     final List<Exception> exceptions = new ArrayList<>();
     int waitForNum = submitted;
     Collection<RaftPeerId> votedPeers = new ArrayList<>();
@@ -259,7 +285,15 @@ class LeaderElection implements Runnable {
         }
 
         final RequestVoteReplyProto r = future.get();
-        responses.add(r);
+        final RaftPeerId replierId = RaftPeerId.valueOf(r.getServerReply().getReplyId());
+        final RequestVoteReplyProto previous = responses.putIfAbsent(replierId, r);
+        if (previous != null) {
+          if (LOG.isWarnEnabled()) {
+            LOG.warn("{} received duplicated replies from {}, the 2nd reply is ignored: 1st={}, 2nd={}",
+                this, replierId, ServerProtoUtils.toString(previous), ServerProtoUtils.toString(r));
+          }
+          continue;
+        }
         if (r.getShouldShutdown()) {
           return logAndReturn(Result.SHUTDOWN, responses, exceptions, -1);
         }
@@ -268,7 +302,7 @@ class LeaderElection implements Runnable {
               exceptions, r.getTerm());
         }
         if (r.getServerReply().getSuccess()) {
-          votedPeers.add(RaftPeerId.valueOf(r.getServerReply().getReplyId()));
+          votedPeers.add(replierId);
           if (conf.hasMajority(votedPeers, server.getId())) {
             return logAndReturn(Result.PASSED, responses, exceptions, -1);
           }
