@@ -21,8 +21,9 @@ import org.apache.ratis.conf.RaftProperties;
 import org.apache.ratis.proto.RaftProtos.ReplicationLevel;
 import org.apache.ratis.protocol.*;
 import org.apache.ratis.server.RaftServerConfigKeys;
+import org.apache.ratis.server.metrics.LogAppenderMetrics;
 import org.apache.ratis.server.protocol.TermIndex;
-import org.apache.ratis.server.storage.RaftLog;
+import org.apache.ratis.server.raftlog.RaftLog;
 import org.apache.ratis.proto.RaftProtos.CommitInfoProto;
 import org.apache.ratis.proto.RaftProtos.AppendEntriesRequestProto;
 import org.apache.ratis.proto.RaftProtos.LogEntryProto;
@@ -98,13 +99,14 @@ public class LeaderState {
   }
 
   private class EventQueue {
+    private final String name = server.getMemberId() + "-" + getClass().getSimpleName();
     private final BlockingQueue<StateUpdateEvent> queue = new ArrayBlockingQueue<>(4096);
 
     void submit(StateUpdateEvent event) {
       try {
         queue.put(event);
       } catch (InterruptedException e) {
-        LOG.info("{}: Interrupted when submitting {} ", server.getId(), event);
+        LOG.info("{}: Interrupted when submitting {} ", this, event);
       }
     }
 
@@ -113,7 +115,7 @@ public class LeaderState {
       try {
         e = queue.poll(server.getMaxTimeoutMs(), TimeUnit.MILLISECONDS);
       } catch(InterruptedException ie) {
-        String s = server.getId() + ": " + getClass().getSimpleName() + " thread is interrupted";
+        String s = this + ": poll() is interrupted";
         if (!running) {
           LOG.info(s + " gracefully");
           return null;
@@ -128,6 +130,11 @@ public class LeaderState {
       }
       return e;
     }
+
+    @Override
+    public String toString() {
+      return name;
+    }
   }
 
   /**
@@ -138,12 +145,8 @@ public class LeaderState {
   static class SenderList {
     private final List<LogAppender> senders;
 
-    SenderList(LogAppender[] senders) {
-      this.senders = new CopyOnWriteArrayList<>(senders);
-    }
-
-    int size() {
-      return senders.size();
+    SenderList() {
+      this.senders = new CopyOnWriteArrayList<>();
     }
 
     Stream<LogAppender> stream() {
@@ -154,8 +157,17 @@ public class LeaderState {
       senders.forEach(action);
     }
 
-    boolean addAll(Collection<LogAppender> c) {
-      return senders.addAll(c);
+    void addAll(Collection<LogAppender> newSenders) {
+      if (newSenders.isEmpty()) {
+        return;
+      }
+
+      Preconditions.assertUnique(
+          CollectionUtils.as(senders, LogAppender::getFollowerId),
+          CollectionUtils.as(newSenders, LogAppender::getFollowerId));
+
+      final boolean changed = senders.addAll(newSenders);
+      Preconditions.assertTrue(changed);
     }
 
     boolean removeAll(Collection<LogAppender> c) {
@@ -168,6 +180,7 @@ public class LeaderState {
   private final StateUpdateEvent CHECK_STAGING_EVENT =
       new StateUpdateEvent(StateUpdateEvent.Type.CHECK_STAGING, -1, this::checkStaging);
 
+  private final String name;
   private final RaftServerImpl server;
   private final RaftLog raftLog;
   private final long currentTerm;
@@ -179,7 +192,7 @@ public class LeaderState {
    * The list is protected by the RaftServer's lock.
    */
   private final SenderList senders;
-  private final EventQueue eventQueue = new EventQueue();
+  private final EventQueue eventQueue;
   private final EventProcessor processor;
   private final PendingRequests pendingRequests;
   private final WatchRequests watchRequests;
@@ -188,8 +201,11 @@ public class LeaderState {
   private final int stagingCatchupGap;
   private final TimeDuration syncInterval;
   private final long placeHolderIndex;
+  private final RaftServerMetrics raftServerMetrics;
+  private final LogAppenderMetrics logAppenderMetrics;
 
   LeaderState(RaftServerImpl server, RaftProperties properties) {
+    this.name = server.getMemberId() + "-" + getClass().getSimpleName();
     this.server = server;
 
     stagingCatchupGap = RaftServerConfigKeys.stagingCatchupGap(properties);
@@ -198,19 +214,20 @@ public class LeaderState {
     final ServerState state = server.getState();
     this.raftLog = state.getLog();
     this.currentTerm = state.getCurrentTerm();
+
+    this.eventQueue = new EventQueue();
     processor = new EventProcessor();
-    this.pendingRequests = new PendingRequests(server.getId());
-    this.watchRequests = new WatchRequests(server.getId(), properties);
+    raftServerMetrics = server.getRaftServerMetrics();
+    logAppenderMetrics = new LogAppenderMetrics(server.getMemberId());
+    this.pendingRequests = new PendingRequests(server.getMemberId(), properties, raftServerMetrics);
+    this.watchRequests = new WatchRequests(server.getMemberId(), properties);
 
     final RaftConfiguration conf = server.getRaftConf();
-    Collection<RaftPeer> others = conf.getOtherPeers(state.getSelfId());
-    final Timestamp t = Timestamp.currentTime().addTimeMs(-server.getMaxTimeoutMs());
+    Collection<RaftPeer> others = conf.getOtherPeers(server.getId());
     placeHolderIndex = raftLog.getNextIndex();
 
-    senders = new SenderList(others.stream().map(
-        p -> server.newLogAppender(this, p, t, placeHolderIndex, true))
-        .toArray(LogAppender[]::new));
-
+    senders = new SenderList();
+    addSenders(others, placeHolderIndex, true);
     voterLists = divideFollowers(conf);
   }
 
@@ -223,7 +240,6 @@ public class LeaderState {
     CodeInjectionForTesting.execute(APPEND_PLACEHOLDER,
         server.getId().toString(), null);
     raftLog.append(placeHolder);
-
     processor.start();
     senders.forEach(LogAppender::startAppender);
     return placeHolder;
@@ -244,9 +260,10 @@ public class LeaderState {
       server.getStateMachine().notifyNotLeader(transactions);
       watchRequests.failWatches(nle);
     } catch (IOException e) {
-      LOG.warn(server.getId() + ": Caught exception in sendNotLeaderResponses", e);
+      LOG.warn("{}: Caught exception in sendNotLeaderResponses", this, e);
     }
-    server.getServerRpc().notifyNotLeader();
+    server.getServerRpc().notifyNotLeader(server.getMemberId().getGroupId());
+    logAppenderMetrics.unregister();
   }
 
   void notifySenders() {
@@ -269,17 +286,17 @@ public class LeaderState {
    * Start bootstrapping new peers
    */
   PendingRequest startSetConfiguration(SetConfigurationRequest request) {
+    LOG.info("{}: startSetConfiguration {}", this, request);
     Preconditions.assertTrue(running && !inStagingState());
 
-    RaftPeer[] peersInNewConf = request.getPeersInNewConf();
-    Collection<RaftPeer> peersToBootStrap = RaftConfiguration
-        .computeNewPeers(peersInNewConf, server.getRaftConf());
+    final List<RaftPeer> peersInNewConf = request.getPeersInNewConf();
+    final Collection<RaftPeer> peersToBootStrap = server.getRaftConf().filterNotContainedInConf(peersInNewConf);
 
     // add the request to the pending queue
     final PendingRequest pending = pendingRequests.addConfRequest(request);
 
     ConfigurationStagingState stagingState = new ConfigurationStagingState(
-        peersToBootStrap, new PeerConfiguration(Arrays.asList(peersInNewConf)));
+        peersToBootStrap, new PeerConfiguration(peersInNewConf));
     Collection<RaftPeer> newPeers = stagingState.getNewPeers();
     // set the staging state
     this.stagingState = stagingState;
@@ -288,21 +305,25 @@ public class LeaderState {
       applyOldNewConf();
     } else {
       // update the LeaderState's sender list
-      addSenders(newPeers);
+      addAndStartSenders(newPeers);
     }
     return pending;
   }
 
-  PendingRequest addPendingRequest(RaftClientRequest request, TransactionContext entry) {
+  PendingRequests.Permit tryAcquirePendingRequest(Message message) {
+    return pendingRequests.tryAcquire(message);
+  }
+
+  PendingRequest addPendingRequest(PendingRequests.Permit permit, RaftClientRequest request, TransactionContext entry) {
     if (LOG.isDebugEnabled()) {
-      LOG.debug("{}: addPendingRequest at {}, entry=", server.getId(), request,
+      LOG.debug("{}: addPendingRequest at {}, entry={}", this, request,
           ServerProtoUtils.toLogEntryString(entry.getLogEntry()));
     }
-    return pendingRequests.add(request, entry);
+    return pendingRequests.add(permit, request, entry);
   }
 
   CompletableFuture<RaftClientReply> addWatchReqeust(RaftClientRequest request) {
-    LOG.debug("{}: addWatchRequest {}", server.getId(), request);
+    LOG.debug("{}: addWatchRequest {}", this, request);
     return watchRequests.add(request)
         .thenApply(v -> new RaftClientReply(request, server.getCommitInfos()))
         .exceptionally(e -> {
@@ -319,7 +340,7 @@ public class LeaderState {
 
   void commitIndexChanged() {
     getMajorityMin(FollowerInfo::getCommitIndex, raftLog::getLastCommittedIndex).ifPresent(m -> {
-      // Normally, leader commit index is always ahead followers.
+      // Normally, leader commit index is always ahead of followers.
       // However, after a leader change, the new leader commit index may
       // be behind some followers in the beginning.
       watchRequests.update(ReplicationLevel.ALL_COMMITTED, m.min);
@@ -355,24 +376,29 @@ public class LeaderState {
   AppendEntriesRequestProto newAppendEntriesRequestProto(RaftPeerId targetId,
       TermIndex previous, List<LogEntryProto> entries, boolean initializing,
       long callId) {
-    return ServerProtoUtils.toAppendEntriesRequestProto(server.getId(), targetId,
-        server.getGroupId(), currentTerm, entries, raftLog.getLastCommittedIndex(),
+    return ServerProtoUtils.toAppendEntriesRequestProto(server.getMemberId(), targetId,
+        currentTerm, entries, raftLog.getLastCommittedIndex(),
         initializing, previous, server.getCommitInfos(), callId);
   }
 
   /**
-   * After receiving a setConfiguration request, the leader should update its
-   * RpcSender list.
+   * Update sender list for setConfiguration request
    */
-  void addSenders(Collection<RaftPeer> newMembers) {
-    final Timestamp t = Timestamp.currentTime().addTimeMs(-server.getMaxTimeoutMs());
-    final long nextIndex = raftLog.getNextIndex();
+  void addAndStartSenders(Collection<RaftPeer> newPeers) {
+    addSenders(newPeers, raftLog.getNextIndex(), false).forEach(LogAppender::startAppender);
+  }
 
-    senders.addAll(newMembers.stream().map(peer -> {
-      LogAppender sender = server.newLogAppender(this, peer, t, nextIndex, false);
-      sender.startAppender();
-      return sender;
-    }).collect(Collectors.toList()));
+  Collection<LogAppender> addSenders(Collection<RaftPeer> newPeers, long nextIndex, boolean attendVote) {
+    final Timestamp t = Timestamp.currentTime().addTimeMs(-server.getMaxTimeoutMs());
+    final List<LogAppender> newAppenders = newPeers.stream()
+        .map(peer -> {
+          LogAppender logAppender = server.newLogAppender(this, peer, t, nextIndex, attendVote);
+          raftServerMetrics.addFollower(logAppender.getFollower().getPeer());
+          logAppenderMetrics.addFollowerGauges(logAppender.getFollower());
+          return logAppender;
+        }).collect(Collectors.toList());
+    senders.addAll(newAppenders);
+    return newAppenders;
   }
 
   void stopAndRemoveSenders(Predicate<LogAppender> predicate) {
@@ -381,6 +407,13 @@ public class LeaderState {
     senders.removeAll(toStop);
   }
 
+  void restartSender(LogAppender sender) {
+    final FollowerInfo follower = sender.getFollower();
+    LOG.info("{}: Restarting {} for {}", this, sender.getClass().getSimpleName(), follower.getName());
+    senders.removeAll(Collections.singleton(sender));
+    addAndStartSenders(Collections.singleton(follower.getPeer()));
+  }
+
   /**
    * Update the RpcSender list based on the current configuration
    */
@@ -401,7 +434,7 @@ public class LeaderState {
     try {
       server.changeToFollowerAndPersistMetadata(term, "stepDown");
     } catch(IOException e) {
-      final String s = server.getId() + ": Failed to persist metadata for term " + term;
+      final String s = this + ": Failed to persist metadata for term " + term;
       LOG.warn(s, e);
       // the failure should happen while changing the state to follower
       // thus the in-memory state should have been updated
@@ -465,8 +498,7 @@ public class LeaderState {
     final Timestamp progressTime = Timestamp.currentTime().addTimeMs(-server.getMaxTimeoutMs());
     final Timestamp timeoutTime = Timestamp.currentTime().addTimeMs(-3*server.getMaxTimeoutMs());
     if (follower.getLastRpcResponseTime().compareTo(timeoutTime) < 0) {
-      LOG.debug("{} detects a follower {} timeout for bootstrapping," +
-              " timeoutTime: {}", server.getId(), follower, timeoutTime);
+      LOG.debug("{} detects a follower {} timeout ({}) for bootstrapping", this, follower, timeoutTime);
       return BootStrapProgress.NOPROGRESS;
     } else if (follower.getMatchIndex() + stagingCatchupGap > committed
         && follower.getLastRpcResponseTime().compareTo(progressTime) > 0) {
@@ -497,8 +529,7 @@ public class LeaderState {
           .getLastCommittedIndex();
       Collection<BootStrapProgress> reports = checkAllProgress(committedIndex);
       if (reports.contains(BootStrapProgress.NOPROGRESS)) {
-        LOG.debug("{} fails the setConfiguration request", server.getId());
-        stagingState.fail();
+        stagingState.fail(BootStrapProgress.NOPROGRESS);
       } else if (!reports.contains(BootStrapProgress.PROGRESSING)) {
         // all caught up!
         applyOldNewConf();
@@ -547,7 +578,7 @@ public class LeaderState {
   }
 
   private void updateCommit() {
-    getMajorityMin(FollowerInfo::getMatchIndex, raftLog::getLatestFlushedIndex)
+    getMajorityMin(FollowerInfo::getMatchIndex, raftLog::getFlushIndex)
         .ifPresent(m -> updateCommit(m.majority, m.min));
   }
 
@@ -621,8 +652,7 @@ public class LeaderState {
         pendingRequests.replySetConfiguration(server::getCommitInfos);
         // if the leader is not included in the current configuration, step down
         if (!conf.containsInConf(server.getId())) {
-          LOG.info("{} is not included in the new configuration {}. Step down.",
-              server.getId(), conf);
+          LOG.info("{} is not included in the new configuration {}. Will shutdown server...", this, conf);
           try {
             // leave some time for all RPC senders to send out new conf entry
             Thread.sleep(server.getMinTimeoutMs());
@@ -706,6 +736,7 @@ public class LeaderState {
   }
 
   private class ConfigurationStagingState {
+    private final String name = server.getMemberId() + "-" + getClass().getSimpleName();
     private final Map<RaftPeerId, RaftPeer> newPeers;
     private final PeerConfiguration newConf;
 
@@ -736,14 +767,19 @@ public class LeaderState {
       return newPeers.containsKey(peerId);
     }
 
-    void fail() {
+    void fail(BootStrapProgress progress) {
+      final String message = this + ": Fail to set configuration " + newConf + " due to " + progress;
+      LOG.debug(message);
       stopAndRemoveSenders(s -> !s.getFollower().isAttendingVote());
 
       LeaderState.this.stagingState = null;
       // send back failure response to client's request
-      pendingRequests.failSetConfiguration(
-          new ReconfigurationTimeoutException("Fail to set configuration "
-              + newConf + ". Timeout when bootstrapping new peers."));
+      pendingRequests.failSetConfiguration(new ReconfigurationTimeoutException(message));
+    }
+
+    @Override
+    public String toString() {
+      return name;
     }
   }
 
@@ -759,4 +795,19 @@ public class LeaderState {
   Stream<LogAppender> getLogAppenders() {
     return senders.stream();
   }
+
+  /**
+   * Record Follower Heartbeat Elapsed Time.
+   * @param follower RaftPeer.
+   * @param elapsedTime Elapsed time in Nanos.
+   */
+  void recordFollowerHeartbeatElapsedTime(RaftPeer follower, long elapsedTime) {
+    raftServerMetrics.recordFollowerHeartbeatElapsedTime(follower,
+        elapsedTime);
+  }
+
+  @Override
+  public String toString() {
+    return name;
+  }
 }
