@@ -21,10 +21,10 @@ import org.apache.ratis.conf.RaftProperties;
 import org.apache.ratis.protocol.RaftPeerId;
 import org.apache.ratis.server.RaftServerConfigKeys;
 import org.apache.ratis.server.protocol.TermIndex;
-import org.apache.ratis.server.storage.RaftLog.EntryWithData;
 import org.apache.ratis.server.storage.FileInfo;
-import org.apache.ratis.server.storage.RaftLog;
-import org.apache.ratis.server.storage.RaftLogIOException;
+import org.apache.ratis.server.raftlog.RaftLog;
+import org.apache.ratis.server.raftlog.RaftLog.EntryWithData;
+import org.apache.ratis.server.raftlog.RaftLogIOException;
 import org.apache.ratis.thirdparty.com.google.protobuf.ByteString;
 import org.apache.ratis.proto.RaftProtos.*;
 import org.apache.ratis.statemachine.SnapshotInfo;
@@ -40,7 +40,7 @@ import java.nio.file.Path;
 import java.util.*;
 
 import static org.apache.ratis.server.impl.RaftServerConstants.DEFAULT_CALLID;
-import static org.apache.ratis.server.impl.RaftServerConstants.INVALID_LOG_INDEX;
+import static org.apache.ratis.server.metrics.RaftLogMetrics.LOG_APPENDER_INSTALL_SNAPSHOT_METRIC;
 import static org.apache.ratis.util.LifeCycle.State.CLOSED;
 import static org.apache.ratis.util.LifeCycle.State.CLOSING;
 import static org.apache.ratis.util.LifeCycle.State.EXCEPTION;
@@ -54,6 +54,72 @@ import static org.apache.ratis.util.LifeCycle.State.STARTING;
 public class LogAppender {
   public static final Logger LOG = LoggerFactory.getLogger(LogAppender.class);
 
+  class AppenderDaemon {
+    private final String name = LogAppender.this + "-" + getClass().getSimpleName();
+    private final LifeCycle lifeCycle = new LifeCycle(name);
+    private final Daemon daemon = new Daemon(this::run);
+
+    void start() {
+      // The life cycle state could be already closed due to server shutdown.
+      if (lifeCycle.compareAndTransition(NEW, STARTING)) {
+        daemon.start();
+      }
+    }
+
+    void run() {
+      synchronized (lifeCycle) {
+        if (!isRunning()) {
+          return;
+        }
+        lifeCycle.transition(RUNNING);
+      }
+      try {
+        runAppenderImpl();
+      } catch (InterruptedException | InterruptedIOException e) {
+        LOG.info(this + " was interrupted: " + e);
+      } catch (RaftLogIOException e) {
+        LOG.error(this + " failed RaftLog", e);
+        lifeCycle.transition(EXCEPTION);
+      } catch (IOException e) {
+        LOG.error(this + " failed IOException", e);
+        lifeCycle.transition(EXCEPTION);
+      } catch (Throwable e) {
+        LOG.error(this + " unexpected exception", e);
+        lifeCycle.transition(EXCEPTION);
+      } finally {
+        if (!lifeCycle.compareAndTransition(CLOSING, CLOSED)) {
+          lifeCycle.transitionIfNotEqual(EXCEPTION);
+        }
+        if (lifeCycle.getCurrentState() == EXCEPTION) {
+          leaderState.restartSender(LogAppender.this);
+        }
+      }
+    }
+
+    boolean isRunning() {
+      return !LifeCycle.States.CLOSING_OR_CLOSED_OR_EXCEPTION.contains(lifeCycle.getCurrentState());
+    }
+
+    void stop() {
+      synchronized (lifeCycle) {
+        if (!isRunning()) {
+          return;
+        }
+        if (lifeCycle.compareAndTransition(NEW, CLOSED)) {
+          return;
+        }
+        lifeCycle.transition(CLOSING);
+      }
+      daemon.interrupt();
+    }
+
+    @Override
+    public String toString() {
+      return name;
+    }
+  }
+
+  private final String name;
   protected final RaftServerImpl server;
   private final LeaderState leaderState;
   protected final RaftLog raftLog;
@@ -63,11 +129,11 @@ public class LogAppender {
   private final int snapshotChunkMaxSize;
   protected final long halfMinTimeoutMs;
 
-  private final LifeCycle lifeCycle;
-  private final Daemon daemon = new Daemon(this::runAppender);
+  private final AppenderDaemon daemon;
 
   public LogAppender(RaftServerImpl server, LeaderState leaderState, FollowerInfo f) {
     this.follower = f;
+    this.name = follower.getName() + "-" + getClass().getSimpleName();
     this.server = server;
     this.leaderState = leaderState;
     this.raftLog = server.getState().getLog();
@@ -79,79 +145,61 @@ public class LogAppender {
     final SizeInBytes bufferByteLimit = RaftServerConfigKeys.Log.Appender.bufferByteLimit(properties);
     final int bufferElementLimit = RaftServerConfigKeys.Log.Appender.bufferElementLimit(properties);
     this.buffer = new DataQueue<>(this, bufferByteLimit, bufferElementLimit, EntryWithData::getSerializedSize);
-    this.lifeCycle = new LifeCycle(this);
+    this.daemon = new AppenderDaemon();
   }
 
   @Override
   public String toString() {
-    return getClass().getSimpleName() + "(" + server.getId() + " -> " +
-        follower.getPeer().getId() + ")";
+    return name;
   }
 
   void startAppender() {
-    // The life cycle state could be already closed due to server shutdown.
-    if (lifeCycle.compareAndTransition(NEW, STARTING)) {
-      daemon.start();
-    }
+    daemon.start();
   }
 
-  private void runAppender() {
-    lifeCycle.transition(RUNNING);
-    try {
-      runAppenderImpl();
-    } catch (InterruptedException | InterruptedIOException e) {
-      LOG.info(this + " was interrupted: " + e);
-    } catch (IOException e) {
-      LOG.error(this + " hit IOException while loading raft log", e);
-      lifeCycle.transition(EXCEPTION);
-    } catch (Throwable e) {
-      LOG.error(this + " unexpected exception", e);
-      lifeCycle.transition(EXCEPTION);
-    } finally {
-      if (!lifeCycle.compareAndTransition(CLOSING, CLOSED)) {
-        lifeCycle.transitionIfNotEqual(EXCEPTION);
-      }
-    }
+  public boolean isAppenderRunning() {
+    return daemon.isRunning();
   }
 
-  protected boolean isAppenderRunning() {
-    return !lifeCycle.getCurrentState().isOneOf(CLOSING, CLOSED, EXCEPTION);
-  }
-
-  public void stopAppender() {
-    if (lifeCycle.compareAndTransition(NEW, CLOSED)) {
-      return;
-    }
-    lifeCycle.transition(CLOSING);
-    daemon.interrupt();
+  void stopAppender() {
+    daemon.stop();
   }
 
   public FollowerInfo getFollower() {
     return follower;
   }
 
-  RaftPeerId getFollowerId() {
+  protected RaftPeerId getFollowerId() {
     return getFollower().getPeer().getId();
   }
 
-  private TermIndex getPrevious() {
-    TermIndex previous = raftLog.getTermIndex(follower.getNextIndex() - 1);
-    if (previous == null) {
-      // if previous is null, nextIndex must be equal to the log start
-      // index (otherwise we will install snapshot).
-      Preconditions.assertTrue(follower.getNextIndex() == raftLog.getStartIndex(),
-          "%s: follower's next index %s, local log start index %s",
-          this, follower.getNextIndex(), raftLog.getStartIndex());
-      SnapshotInfo snapshot = server.getState().getLatestSnapshot();
-      previous = snapshot == null ? null : snapshot.getTermIndex();
+  private TermIndex getPrevious(long nextIndex) {
+    if (nextIndex == RaftLog.LEAST_VALID_LOG_INDEX) {
+      return null;
     }
-    return previous;
+
+    final long previousIndex = nextIndex - 1;
+    final TermIndex previous = raftLog.getTermIndex(previousIndex);
+    if (previous != null) {
+      return previous;
+    }
+
+    final SnapshotInfo snapshot = server.getState().getLatestSnapshot();
+    if (snapshot != null) {
+      final TermIndex snapshotTermIndex = snapshot.getTermIndex();
+      if (snapshotTermIndex.getIndex() == previousIndex) {
+        return snapshotTermIndex;
+      }
+    }
+
+    return null;
   }
 
   protected AppendEntriesRequestProto createRequest(long callId) throws RaftLogIOException {
-    final TermIndex previous = getPrevious();
+    final TermIndex previous = getPrevious(follower.getNextIndex());
     final long heartbeatRemainingMs = getHeartbeatRemainingTime();
     if (heartbeatRemainingMs <= 0L) {
+      // heartbeat
       return leaderState.newAppendEntriesRequestProto(
           getFollowerId(), previous, Collections.emptyList(), !follower.isAttendingVote(), callId);
     }
@@ -159,7 +207,9 @@ public class LogAppender {
     Preconditions.assertTrue(buffer.isEmpty(), () -> "buffer has " + buffer.getNumElements() + " elements.");
 
     final long leaderNext = raftLog.getNextIndex();
-    for (long next = follower.getNextIndex(); leaderNext > next; ) {
+    final long followerNext = follower.getNextIndex();
+    final long halfMs = heartbeatRemainingMs/2;
+    for (long next = followerNext; leaderNext > next && getHeartbeatRemainingTime() - halfMs > 0; ) {
       if (!buffer.offer(raftLog.getEntryWithData(next++))) {
         break;
       }
@@ -168,13 +218,30 @@ public class LogAppender {
       return null;
     }
 
-    final List<LogEntryProto> protos = buffer.pollList(heartbeatRemainingMs, EntryWithData::getEntry,
-        (entry, time, exception) -> LOG.warn(this + ": Failed get " + entry + " in " + time, exception));
+    final List<LogEntryProto> protos = buffer.pollList(getHeartbeatRemainingTime(), EntryWithData::getEntry,
+        (entry, time, exception) -> LOG.warn("{}: Failed to get {} in {}: {}",
+            follower.getName(), entry, time, exception));
     buffer.clear();
+    assertProtos(protos, followerNext, previous);
     return leaderState.newAppendEntriesRequestProto(
         getFollowerId(), previous, protos, !follower.isAttendingVote(), callId);
   }
 
+  private void assertProtos(List<LogEntryProto> protos, long nextIndex, TermIndex previous) {
+    if (protos.isEmpty()) {
+      return;
+    }
+    final long firstIndex = protos.get(0).getIndex();
+    Preconditions.assertTrue(firstIndex == nextIndex,
+        () -> follower.getName() + ": firstIndex = " + firstIndex + " != nextIndex = " + nextIndex);
+    if (firstIndex > RaftLog.LEAST_VALID_LOG_INDEX) {
+      Objects.requireNonNull(previous,
+          () -> follower.getName() + ": Previous TermIndex not found for firstIndex = " + firstIndex);
+      Preconditions.assertTrue(previous.getIndex() == firstIndex - 1,
+          () -> follower.getName() + ": Previous = " + previous + " but firstIndex = " + firstIndex);
+    }
+  }
+
   /** Send an appendEntries RPC; retry indefinitely. */
   private AppendEntriesReplyProto sendAppendEntriesWithRetries()
       throws InterruptedException, InterruptedIOException, RaftLogIOException {
@@ -187,11 +254,10 @@ public class LogAppender {
         }
 
         if (request == null) {
-          LOG.trace("{} need not send AppendEntries now." +
-              " Wait for more entries.", server.getId());
+          LOG.trace("{} no entries to send now, wait ...", this);
           return null;
         } else if (!isAppenderRunning()) {
-          LOG.debug("LogAppender {} has been stopped. Skip the request.", this);
+          LOG.info("{} is stopped. Skip appendEntries.", this);
           return null;
         }
 
@@ -308,7 +374,7 @@ public class LogAppender {
               } catch (IOException ignored) {
               }
             }
-            LOG.warn("Got exception when preparing InstallSnapshot request", e);
+            LOG.warn("{}: Failed to prepare installSnapshot request", LogAppender.this, e);
             throw new RuntimeException(e);
           }
         }
@@ -316,6 +382,11 @@ public class LogAppender {
     }
   }
 
+  protected InstallSnapshotRequestProto createInstallSnapshotNotificationRequest(
+      TermIndex firstLogStartTermIndex) {
+    return server.createInstallSnapshotRequest(getFollowerId(), firstLogStartTermIndex);
+  }
+
   private FileChunkProto readFileChunk(FileInfo fileInfo,
       FileInputStream in, byte[] buf, int length, long offset, int chunkIndex)
       throws IOException {
@@ -356,8 +427,8 @@ public class LogAppender {
 
     if (reply != null) {
       follower.setSnapshotIndex(snapshot.getTermIndex().getIndex());
-      LOG.info("{}:{} install snapshot-{} successfully on follower {}",
-          server.getId(), server.getGroupId(), snapshot.getTermIndex().getIndex(), follower.getPeer());
+      LOG.info("{}: installSnapshot {} successfully", this, snapshot);
+      server.getRaftServerMetrics().getCounter(LOG_APPENDER_INSTALL_SNAPSHOT_METRIC).inc();
     }
     return reply;
   }
@@ -370,7 +441,7 @@ public class LogAppender {
     if (follower.getNextIndex() < raftLog.getNextIndex()) {
       SnapshotInfo snapshot = server.getState().getLatestSnapshot();
       if (follower.getNextIndex() < logStartIndex ||
-          (logStartIndex == INVALID_LOG_INDEX && snapshot != null)) {
+          (logStartIndex == RaftLog.INVALID_LOG_INDEX && snapshot != null)) {
         return snapshot;
       }
     }
@@ -383,10 +454,8 @@ public class LogAppender {
       if (shouldSendRequest()) {
         SnapshotInfo snapshot = shouldInstallSnapshot();
         if (snapshot != null) {
-          LOG.info("{}:{} follower {}'s next index is {}," +
-              " log's start index is {}, need to install snapshot",
-              server.getId(), server.getGroupId(), follower.getPeer(), follower.getNextIndex(),
-              raftLog.getStartIndex());
+          LOG.info("{}: followerNextIndex = {} but logStartIndex = {}, send snapshot {} to follower",
+              this, follower.getNextIndex(), raftLog.getStartIndex(), snapshot);
 
           final InstallSnapshotReplyProto r = installSnapshot(snapshot);
           if (r != null && r.getResult() == InstallSnapshotResult.NOT_LEADER) {
@@ -425,7 +494,7 @@ public class LogAppender {
 
           if (nextIndex > oldNextIndex) {
             follower.updateMatchIndex(nextIndex - 1);
-            follower.updateNextIndex(nextIndex);
+            follower.increaseNextIndex(nextIndex);
             submitEventOnSuccessAppend();
           }
           break;
@@ -437,8 +506,7 @@ public class LogAppender {
           follower.decreaseNextIndex(reply.getNextIndex());
           break;
         case UNRECOGNIZED:
-          LOG.warn("{} received UNRECOGNIZED AppendResult from {}",
-              server.getId(), follower.getPeer().getId());
+          LOG.warn("{}: received {}", this, reply.getResult());
           break;
       }
     }
@@ -459,8 +527,10 @@ public class LogAppender {
 
   protected void checkSlowness() {
     if (follower.isSlow()) {
-      server.getStateMachine().notifySlowness(server.getGroup(), server.getRoleInfoProto());
+      server.getStateMachine().notifySlowness(server.getRoleInfoProto());
     }
+    leaderState.recordFollowerHeartbeatElapsedTime(follower.getPeer(),
+        follower.getLastRpcResponseTime().elapsedTime().getDuration());
   }
 
   public synchronized void notifyAppend() {
@@ -476,7 +546,7 @@ public class LogAppender {
     return followerIndex < raftLog.getNextIndex();
   }
 
-  private boolean shouldHeartbeat() {
+  protected boolean shouldHeartbeat() {
     return getHeartbeatRemainingTime() <= 0;
   }
 
@@ -487,12 +557,14 @@ public class LogAppender {
     return halfMinTimeoutMs - follower.getLastRpcTime().elapsedTimeMs();
   }
 
-  protected void checkResponseTerm(long responseTerm) {
+  protected boolean checkResponseTerm(long responseTerm) {
     synchronized (server) {
       if (isAppenderRunning() && follower.isAttendingVote()
           && responseTerm > leaderState.getCurrentTerm()) {
         leaderState.submitStepDownEvent(responseTerm);
+        return true;
       }
     }
+    return false;
   }
 }
