@@ -17,35 +17,109 @@
  */
 package org.apache.ratis.server.impl;
 
+import org.apache.ratis.conf.RaftProperties;
 import org.apache.ratis.proto.RaftProtos.CommitInfoProto;
-import org.apache.ratis.protocol.*;
 import org.apache.ratis.proto.RaftProtos.RaftClientRequestProto;
+import org.apache.ratis.protocol.Message;
+import org.apache.ratis.protocol.NotLeaderException;
+import org.apache.ratis.protocol.RaftClientReply;
+import org.apache.ratis.protocol.RaftClientRequest;
+import org.apache.ratis.protocol.RaftException;
+import org.apache.ratis.protocol.RaftGroupMemberId;
+import org.apache.ratis.protocol.SetConfigurationRequest;
+import org.apache.ratis.server.RaftServerConfigKeys;
 import org.apache.ratis.statemachine.TransactionContext;
 import org.apache.ratis.util.Preconditions;
+import org.apache.ratis.util.ResourceSemaphore;
+import org.apache.ratis.util.SizeInBytes;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
 
+import java.util.ArrayList;
 import java.util.Collection;
+import java.util.HashMap;
+import java.util.Iterator;
+import java.util.List;
+import java.util.Map;
 import java.util.concurrent.ConcurrentHashMap;
 import java.util.concurrent.ConcurrentMap;
 import java.util.function.Supplier;
-import java.util.stream.Collectors;
 
 class PendingRequests {
   public static final Logger LOG = LoggerFactory.getLogger(PendingRequests.class);
 
+  static class Permit {}
+
+  static class RequestLimits extends ResourceSemaphore.Group {
+    RequestLimits(int elementLimit, SizeInBytes byteLimit) {
+      super(elementLimit, byteLimit.getSizeInt());
+    }
+
+    int getElementCount() {
+      return get(0).used();
+    }
+
+    // TODO: add metrics
+    int getByteSize() {
+      return get(1).used();
+    }
+
+    boolean tryAcquire(Message message) {
+      return tryAcquire(1, Message.getSize(message));
+    }
+
+    void release(Message message) {
+      release(1, Message.getSize(message));
+    }
+  }
+
   private static class RequestMap {
     private final Object name;
     private final ConcurrentMap<Long, PendingRequest> map = new ConcurrentHashMap<>();
+    private final RaftServerMetrics raftServerMetrics;
 
-    RequestMap(Object name) {
+    /** Permits to put new requests, always synchronized. */
+    private final Map<Permit, Permit> permits = new HashMap<>();
+    /** Track and limit the number of requests and the total message size. */
+    private final RequestLimits resource;
+
+    RequestMap(Object name, int elementLimit, SizeInBytes byteLimit, RaftServerMetrics raftServerMetrics) {
       this.name = name;
+      this.resource = new RequestLimits(elementLimit, byteLimit);
+      this.raftServerMetrics = raftServerMetrics;
+
+      raftServerMetrics.addNumPendingRequestsGauge(resource::getElementCount);
     }
 
-    void put(long index, PendingRequest p) {
+    Permit tryAcquire(Message message) {
+      final boolean acquired = resource.tryAcquire(message);
+      LOG.trace("tryAcquire? {}", acquired);
+      if (!acquired) {
+        raftServerMetrics.onRequestQueueLimitHit();
+        return null;
+      }
+      return putPermit();
+    }
+
+    private synchronized Permit putPermit() {
+      if (resource.isClosed()) {
+        return null;
+      }
+      final Permit permit = new Permit();
+      permits.put(permit, permit);
+      return permit;
+    }
+
+    synchronized PendingRequest put(Permit permit, long index, PendingRequest p) {
       LOG.debug("{}: PendingRequests.put {} -> {}", name, index, p);
+      final Permit removed = permits.remove(permit);
+      if (removed == null) {
+        return null;
+      }
+      Preconditions.assertTrue(removed == permit);
       final PendingRequest previous = map.put(index, p);
       Preconditions.assertTrue(previous == null);
+      return p;
     }
 
     PendingRequest get(long index) {
@@ -57,17 +131,32 @@ class PendingRequests {
     PendingRequest remove(long index) {
       final PendingRequest r = map.remove(index);
       LOG.debug("{}: PendingRequests.remove {} returns {}", name, index, r);
+      if (r == null) {
+        return null;
+      }
+      resource.release(r.getRequest().getMessage());
+      LOG.trace("release");
       return r;
     }
 
     Collection<TransactionContext> setNotLeaderException(NotLeaderException nle, Collection<CommitInfoProto> commitInfos) {
+      synchronized (this) {
+        resource.close();
+        permits.clear();
+      }
+
       LOG.debug("{}: PendingRequests.setNotLeaderException", name);
-      try {
-        return map.values().stream()
-            .map(p -> p.setNotLeaderException(nle, commitInfos))
-            .collect(Collectors.toList());
-      } finally {
-        map.clear();
+      final List<TransactionContext> transactions = new ArrayList<>(map.size());
+      for(;;) {
+        final Iterator<Long> i = map.keySet().iterator();
+        if (!i.hasNext()) { // the map is empty
+          return transactions;
+        }
+
+        final PendingRequest pending = map.remove(i.next());
+        if (pending != null) {
+          transactions.add(pending.setNotLeaderException(nle, commitInfos));
+        }
       }
     }
   }
@@ -76,19 +165,25 @@ class PendingRequests {
   private final String name;
   private final RequestMap pendingRequests;
 
-  PendingRequests(RaftPeerId id) {
+  PendingRequests(RaftGroupMemberId id, RaftProperties properties, RaftServerMetrics raftServerMetrics) {
     this.name = id + "-" + getClass().getSimpleName();
-    this.pendingRequests = new RequestMap(id);
+    this.pendingRequests = new RequestMap(id,
+        RaftServerConfigKeys.Write.elementLimit(properties),
+        RaftServerConfigKeys.Write.byteLimit(properties),
+        raftServerMetrics);
   }
 
-  PendingRequest add(RaftClientRequest request, TransactionContext entry) {
+  Permit tryAcquire(Message message) {
+    return pendingRequests.tryAcquire(message);
+  }
+
+  PendingRequest add(Permit permit, RaftClientRequest request, TransactionContext entry) {
     // externally synced for now
     Preconditions.assertTrue(request.is(RaftClientRequestProto.TypeCase.WRITE));
     final long index = entry.getLogEntry().getIndex();
     LOG.debug("{}: addPendingRequest at index={}, request={}", name, index, request);
     final PendingRequest pending = new PendingRequest(index, request, entry);
-    pendingRequests.put(index, pending);
-    return pending;
+    return pendingRequests.put(permit, index, pending);
   }
 
   PendingRequest addConfRequest(SetConfigurationRequest request) {
