@@ -384,7 +384,9 @@ public class RaftServerProxy implements RaftServer {
         }, implExecutor)
         .whenComplete((_1, throwable) -> {
           if (throwable != null) {
-            impls.remove(newGroup.getGroupId());
+            if (!(throwable.getCause() instanceof AlreadyExistsException)) {
+              impls.remove(newGroup.getGroupId());
+            }
             LOG.warn(getId() + ": Failed groupAdd* " + request, throwable);
           }
         });
@@ -404,6 +406,7 @@ public class RaftServerProxy implements RaftServer {
     return f.thenApply(impl -> {
       final Collection<CommitInfoProto> commitInfos = impl.getCommitInfos();
       impl.shutdown(deleteDirectory);
+      impl.getStateMachine().notifyGroupRemove();
       return new RaftClientReply(request, commitInfos);
     });
   }
