@@ -22,7 +22,7 @@ import org.apache.ratis.proto.RaftProtos.*;
 import org.apache.ratis.proto.RaftProtos.AppendEntriesReplyProto.AppendResult;
 import org.apache.ratis.protocol.ClientId;
 import org.apache.ratis.protocol.RaftClientRequest;
-import org.apache.ratis.protocol.RaftGroupId;
+import org.apache.ratis.protocol.RaftGroupMemberId;
 import org.apache.ratis.protocol.RaftPeer;
 import org.apache.ratis.protocol.RaftPeerId;
 import org.apache.ratis.server.protocol.TermIndex;
@@ -34,6 +34,7 @@ import java.util.Arrays;
 import java.util.Collection;
 import java.util.List;
 import java.util.Optional;
+import java.util.function.Function;
 import java.util.stream.Collectors;
 
 import static org.apache.ratis.server.impl.RaftServerConstants.DEFAULT_CALLID;
@@ -65,17 +66,34 @@ public interface ServerProtoUtils {
   }
 
   static String toLogEntryString(LogEntryProto entry) {
+    return toLogEntryString(entry, null);
+  }
+
+  static String toStateMachineLogEntryString(StateMachineLogEntryProto smLog,
+                                             Function<StateMachineLogEntryProto, String> function) {
+    final ByteString clientId = smLog.getClientId();
+    String callIdString = (clientId.isEmpty() ? "<empty clientId>" : ClientId.valueOf(clientId))
+        + ", cid=" + smLog.getCallId();
+
+    String smString = "";
+    if (function != null) {
+      smString = "\n\t State Machine: " + function.apply(smLog);
+    }
+    return callIdString + smString;
+  }
+
+
+  static String toLogEntryString(LogEntryProto entry,
+                                 Function<StateMachineLogEntryProto, String> function) {
     if (entry == null) {
       return null;
     }
     final String s;
     if (entry.hasStateMachineLogEntry()) {
-      final StateMachineLogEntryProto smLog = entry.getStateMachineLogEntry();
-      final ByteString clientId = smLog.getClientId();
-      s = ", " + (clientId.isEmpty()? "<empty clientId>": ClientId.valueOf(clientId)) + ", cid=" + smLog.getCallId();
+      s = ", " + toStateMachineLogEntryString(entry.getStateMachineLogEntry(), function);
     } else if (entry.hasMetadataEntry()) {
       final MetadataProto metadata = entry.getMetadataEntry();
-      s = "(c" + metadata.getCommitIndex() + ")";
+      s = "(c:" + metadata.getCommitIndex() + ")";
     } else {
       s = "";
     }
@@ -94,6 +112,9 @@ public interface ServerProtoUtils {
         : "size=" + entries.size() + ", first=" + toLogEntryString(entries.get(0));
   }
   static String toString(AppendEntriesRequestProto proto) {
+    if (proto == null) {
+      return null;
+    }
     return ProtoUtils.toString(proto.getServerRequest()) + "-t" + proto.getLeaderTerm()
         + ", previous=" + toTermIndexString(proto.getPreviousLog())
         + ", leaderCommit=" + proto.getLeaderCommit()
@@ -101,14 +122,57 @@ public interface ServerProtoUtils {
         + ", entries: " + toShortString(proto.getEntriesList());
   }
   static String toString(AppendEntriesReplyProto reply) {
-    return toString(reply.getServerReply()) + "," + reply.getResult()
+    if (reply == null) {
+      return null;
+    }
+    return ProtoUtils.toString(reply.getServerReply()) + "," + reply.getResult()
         + ",nextIndex:" + reply.getNextIndex() + ",term:" + reply.getTerm()
         + ",followerCommit:" + reply.getFollowerCommit();
   }
 
-  static String toString(RaftRpcReplyProto reply) {
-    return reply.getRequestorId().toStringUtf8() + "->"
-        + reply.getReplyId().toStringUtf8() + "," + reply.getSuccess();
+  static String toString(RequestVoteReplyProto proto) {
+    if (proto == null) {
+      return null;
+    }
+    return ProtoUtils.toString(proto.getServerReply()) + "-t" + proto.getTerm();
+  }
+
+  static String toString(InstallSnapshotRequestProto proto) {
+    if (proto == null) {
+      return null;
+    }
+    final String s;
+    switch (proto.getInstallSnapshotRequestBodyCase()) {
+      case SNAPSHOTCHUNK:
+        final InstallSnapshotRequestProto.SnapshotChunkProto chunk = proto.getSnapshotChunk();
+        s = "chunk:" + chunk.getRequestId() + "," + chunk.getRequestIndex();
+        break;
+      case NOTIFICATION:
+        final InstallSnapshotRequestProto.NotificationProto notification = proto.getNotification();
+        s = "notify:" + toTermIndexString(notification.getFirstAvailableTermIndex());
+        break;
+      default:
+        throw new IllegalStateException("Unexpected body case in " + proto);
+    }
+    return ProtoUtils.toString(proto.getServerRequest()) + "-t" + proto.getLeaderTerm() + "," + s;
+  }
+
+  static String toString(InstallSnapshotReplyProto proto) {
+    if (proto == null) {
+      return null;
+    }
+    final String s;
+    switch (proto.getInstallSnapshotReplyBodyCase()) {
+      case REQUESTINDEX:
+        s = ",requestIndex=" + proto.getRequestIndex();
+        break;
+      case SNAPSHOTINDEX:
+        s = ",snapshotIndex=" + proto.getSnapshotIndex();
+        break;
+      default:
+        s = ""; // result is not SUCCESS
+    }
+    return ProtoUtils.toString(proto.getServerReply()) + "-t" + proto.getTerm() + "," + proto.getResult() + s;
   }
 
   static RaftConfigurationProto.Builder toRaftConfigurationProto(RaftConfiguration conf) {
@@ -121,10 +185,10 @@ public interface ServerProtoUtils {
     Preconditions.assertTrue(entry.hasConfigurationEntry());
     final RaftConfigurationProto proto = entry.getConfigurationEntry();
     final RaftConfiguration.Builder b = RaftConfiguration.newBuilder()
-        .setConf(ProtoUtils.toRaftPeerArray(proto.getPeersList()))
+        .setConf(ProtoUtils.toRaftPeers(proto.getPeersList()))
         .setLogEntryIndex(entry.getIndex());
     if (proto.getOldPeersCount() > 0) {
-      b.setOldConf(ProtoUtils.toRaftPeerArray(proto.getOldPeersList()));
+      b.setOldConf(ProtoUtils.toRaftPeers(proto.getOldPeersList()));
     }
     return b.build();
   }
@@ -242,31 +306,30 @@ public interface ServerProtoUtils {
   }
 
   static RaftRpcReplyProto.Builder toRaftRpcReplyProtoBuilder(
-      RaftPeerId requestorId, RaftPeerId replyId, RaftGroupId groupId, boolean success) {
+      RaftPeerId requestorId, RaftGroupMemberId replyId, boolean success) {
     return ClientProtoUtils.toRaftRpcReplyProtoBuilder(
-        requestorId.toByteString(), replyId.toByteString(), groupId, DEFAULT_CALLID, success);
+        requestorId.toByteString(), replyId.getPeerId().toByteString(), replyId.getGroupId(), DEFAULT_CALLID, success);
   }
 
   static RequestVoteReplyProto toRequestVoteReplyProto(
-      RaftPeerId requestorId, RaftPeerId replyId, RaftGroupId groupId,
-      boolean success, long term, boolean shouldShutdown) {
+      RaftPeerId requestorId, RaftGroupMemberId replyId, boolean success, long term, boolean shouldShutdown) {
     return RequestVoteReplyProto.newBuilder()
-        .setServerReply(toRaftRpcReplyProtoBuilder(requestorId, replyId, groupId, success))
+        .setServerReply(toRaftRpcReplyProtoBuilder(requestorId, replyId, success))
         .setTerm(term)
         .setShouldShutdown(shouldShutdown)
         .build();
   }
 
   static RaftRpcRequestProto.Builder toRaftRpcRequestProtoBuilder(
-      RaftPeerId requestorId, RaftPeerId replyId, RaftGroupId groupId) {
+      RaftGroupMemberId requestorId, RaftPeerId replyId) {
     return ClientProtoUtils.toRaftRpcRequestProtoBuilder(
-        requestorId.toByteString(), replyId.toByteString(), groupId, DEFAULT_CALLID, null);
+        requestorId.getPeerId().toByteString(), replyId.toByteString(), requestorId.getGroupId(), DEFAULT_CALLID, null);
   }
 
   static RequestVoteRequestProto toRequestVoteRequestProto(
-      RaftPeerId requestorId, RaftPeerId replyId, RaftGroupId groupId, long term, TermIndex lastEntry) {
+      RaftGroupMemberId requestorId, RaftPeerId replyId, long term, TermIndex lastEntry) {
     final RequestVoteRequestProto.Builder b = RequestVoteRequestProto.newBuilder()
-        .setServerRequest(toRaftRpcRequestProtoBuilder(requestorId, replyId, groupId))
+        .setServerRequest(toRaftRpcRequestProtoBuilder(requestorId, replyId))
         .setCandidateTerm(term);
     if (lastEntry != null) {
       b.setCandidateLastEntry(toTermIndexProto(lastEntry));
@@ -275,51 +338,96 @@ public interface ServerProtoUtils {
   }
 
   static InstallSnapshotReplyProto toInstallSnapshotReplyProto(
-      RaftPeerId requestorId, RaftPeerId replyId, RaftGroupId groupId,
-      long term, int requestIndex, InstallSnapshotResult result) {
+      RaftPeerId requestorId, RaftGroupMemberId replyId,
+      long currentTerm, int requestIndex, InstallSnapshotResult result) {
     final RaftRpcReplyProto.Builder rb = toRaftRpcReplyProtoBuilder(requestorId,
-        replyId, groupId, result == InstallSnapshotResult.SUCCESS);
+        replyId, result == InstallSnapshotResult.SUCCESS);
     final InstallSnapshotReplyProto.Builder builder = InstallSnapshotReplyProto
-        .newBuilder().setServerReply(rb).setTerm(term).setResult(result)
+        .newBuilder().setServerReply(rb).setTerm(currentTerm).setResult(result)
         .setRequestIndex(requestIndex);
     return builder.build();
   }
 
-  static InstallSnapshotRequestProto toInstallSnapshotRequestProto(
-      RaftPeerId requestorId, RaftPeerId replyId, RaftGroupId groupId, String requestId, int requestIndex,
-      long term, TermIndex lastTermIndex, List<FileChunkProto> chunks,
-      long totalSize, boolean done) {
-    return InstallSnapshotRequestProto.newBuilder()
-        .setServerRequest(toRaftRpcRequestProtoBuilder(requestorId, replyId, groupId))
-        .setRequestId(requestId)
-        .setRequestIndex(requestIndex)
-        // .setRaftConfiguration()  TODO: save and pass RaftConfiguration
-        .setLeaderTerm(term)
-        .setTermIndex(toTermIndexProto(lastTermIndex))
-        .addAllFileChunks(chunks)
-        .setTotalSize(totalSize)
-        .setDone(done).build();
+  static InstallSnapshotReplyProto toInstallSnapshotReplyProto(
+      RaftPeerId requestorId, RaftGroupMemberId replyId,
+      long currentTerm, InstallSnapshotResult result, long installedSnapshotIndex) {
+    final RaftRpcReplyProto.Builder rb = toRaftRpcReplyProtoBuilder(requestorId,
+        replyId, result == InstallSnapshotResult.SUCCESS);
+    final InstallSnapshotReplyProto.Builder builder = InstallSnapshotReplyProto
+        .newBuilder().setServerReply(rb).setTerm(currentTerm).setResult(result);
+    if (installedSnapshotIndex > 0) {
+      builder.setSnapshotIndex(installedSnapshotIndex);
+    }
+    return builder.build();
   }
 
+  static InstallSnapshotReplyProto toInstallSnapshotReplyProto(
+      RaftPeerId requestorId, RaftGroupMemberId replyId,
+      InstallSnapshotResult result) {
+    final RaftRpcReplyProto.Builder rb = toRaftRpcReplyProtoBuilder(requestorId,
+        replyId, result == InstallSnapshotResult.SUCCESS);
+    final InstallSnapshotReplyProto.Builder builder = InstallSnapshotReplyProto
+        .newBuilder().setServerReply(rb).setResult(result);
+    return builder.build();
+  }
+
+  static InstallSnapshotRequestProto toInstallSnapshotRequestProto(
+      RaftGroupMemberId requestorId, RaftPeerId replyId, String requestId, int requestIndex,
+      long term, TermIndex lastTermIndex, List<FileChunkProto> chunks,
+      long totalSize, boolean done) {
+    final InstallSnapshotRequestProto.SnapshotChunkProto.Builder snapshotChunkProto =
+        InstallSnapshotRequestProto.SnapshotChunkProto.newBuilder()
+            .setRequestId(requestId)
+            .setRequestIndex(requestIndex)
+            .setTermIndex(toTermIndexProto(lastTermIndex))
+            .addAllFileChunks(chunks)
+            .setTotalSize(totalSize)
+            .setDone(done);
+    return InstallSnapshotRequestProto.newBuilder()
+        .setServerRequest(toRaftRpcRequestProtoBuilder(requestorId, replyId))
+        // .setRaftConfiguration()  TODO: save and pass RaftConfiguration
+        .setLeaderTerm(term)
+        .setSnapshotChunk(snapshotChunkProto)
+        .build();
+  }
+
+  static InstallSnapshotRequestProto toInstallSnapshotRequestProto(
+      RaftGroupMemberId requestorId, RaftPeerId replyId, long leaderTerm, TermIndex firstAvailable) {
+    final InstallSnapshotRequestProto.NotificationProto.Builder notificationProto =
+        InstallSnapshotRequestProto.NotificationProto.newBuilder()
+            .setFirstAvailableTermIndex(toTermIndexProto(firstAvailable));
+    return InstallSnapshotRequestProto.newBuilder()
+        .setServerRequest(toRaftRpcRequestProtoBuilder(requestorId, replyId))
+        // .setRaftConfiguration()  TODO: save and pass RaftConfiguration
+        .setLeaderTerm(leaderTerm)
+        .setNotification(notificationProto)
+        .build();
+  }
+
+  @SuppressWarnings("parameternumber")
   static AppendEntriesReplyProto toAppendEntriesReplyProto(
-      RaftPeerId requestorId, RaftPeerId replyId, RaftGroupId groupId, long term,
-      long followerCommit, long nextIndex, AppendResult result, long callId) {
+      RaftPeerId requestorId, RaftGroupMemberId replyId, long term,
+      long followerCommit, long nextIndex, AppendResult result, long callId,
+      long matchIndex, boolean isHeartbeat) {
     RaftRpcReplyProto.Builder rpcReply = toRaftRpcReplyProtoBuilder(
-        requestorId, replyId, groupId, result == AppendResult.SUCCESS)
+        requestorId, replyId, result == AppendResult.SUCCESS)
         .setCallId(callId);
     return AppendEntriesReplyProto.newBuilder()
         .setServerReply(rpcReply)
         .setTerm(term)
         .setNextIndex(nextIndex)
+        .setMatchIndex(matchIndex)
         .setFollowerCommit(followerCommit)
-        .setResult(result).build();
+        .setResult(result)
+        .setIsHearbeat(isHeartbeat)
+        .build();
   }
 
   static AppendEntriesRequestProto toAppendEntriesRequestProto(
-      RaftPeerId requestorId, RaftPeerId replyId, RaftGroupId groupId, long leaderTerm,
+      RaftGroupMemberId requestorId, RaftPeerId replyId, long leaderTerm,
       List<LogEntryProto> entries, long leaderCommit, boolean initializing,
       TermIndex previous, Collection<CommitInfoProto> commitInfos, long callId) {
-    RaftRpcRequestProto.Builder rpcRequest = toRaftRpcRequestProtoBuilder(requestorId, replyId, groupId)
+    RaftRpcRequestProto.Builder rpcRequest = toRaftRpcRequestProtoBuilder(requestorId, replyId)
         .setCallId(callId);
     final AppendEntriesRequestProto.Builder b = AppendEntriesRequestProto
         .newBuilder()
@@ -334,7 +442,7 @@ public interface ServerProtoUtils {
     if (previous != null) {
       b.setPreviousLog(toTermIndexProto(previous));
     }
-    ProtoUtils.addCommitInfos(commitInfos, i -> b.addCommitInfos(i));
+    ProtoUtils.addCommitInfos(commitInfos, b::addCommitInfos);
     return b.build();
   }
 
