@@ -21,12 +21,16 @@ import org.apache.ratis.conf.RaftProperties;
 import org.apache.ratis.protocol.*;
 import org.apache.ratis.server.RaftServerConfigKeys;
 import org.apache.ratis.server.protocol.TermIndex;
+import org.apache.ratis.server.raftlog.RaftLog;
+import org.apache.ratis.server.raftlog.memory.MemoryRaftLog;
+import org.apache.ratis.server.raftlog.segmented.SegmentedRaftLog;
 import org.apache.ratis.server.storage.*;
 import org.apache.ratis.proto.RaftProtos.InstallSnapshotRequestProto;
 import org.apache.ratis.proto.RaftProtos.LogEntryProto;
 import org.apache.ratis.statemachine.SnapshotInfo;
 import org.apache.ratis.statemachine.StateMachine;
 import org.apache.ratis.statemachine.TransactionContext;
+import org.apache.ratis.util.TimeDuration;
 import org.apache.ratis.util.Timestamp;
 
 import java.io.Closeable;
@@ -40,8 +44,8 @@ import java.util.List;
 import java.util.Map;
 import java.util.Objects;
 import java.util.Optional;
-import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicLong;
+import java.util.concurrent.atomic.AtomicReference;
 import java.util.function.Consumer;
 
 import static org.apache.ratis.server.impl.RaftServerImpl.LOG;
@@ -50,8 +54,7 @@ import static org.apache.ratis.server.impl.RaftServerImpl.LOG;
  * Common states of a raft peer. Protected by RaftServer's lock.
  */
 public class ServerState implements Closeable {
-  private final RaftPeerId selfId;
-  private final RaftGroupId groupId;
+  private final RaftGroupMemberId memberId;
   private final RaftServerImpl server;
   /** Raft log */
   private final RaftLog log;
@@ -63,7 +66,7 @@ public class ServerState implements Closeable {
   private final RaftStorage storage;
   private final SnapshotManager snapshotManager;
   private volatile Timestamp lastNoLeaderTime;
-  private final long leaderElectionTimeoutMs;
+  private final TimeDuration noLeaderTimeout;
 
   /**
    * Latest term server has seen.
@@ -87,23 +90,23 @@ public class ServerState implements Closeable {
    * snapshot. Once we successfully install a snapshot, the SM may not pick it up immediately.
    * Further, this will not get updated when SM does snapshots itself.
    */
-  private volatile TermIndex latestInstalledSnapshot;
+  private final AtomicReference<TermIndex> latestInstalledSnapshot = new AtomicReference<>();
 
   ServerState(RaftPeerId id, RaftGroup group, RaftProperties prop,
               RaftServerImpl server, StateMachine stateMachine)
       throws IOException {
-    this.selfId = id;
-    this.groupId = group.getGroupId();
+    this.memberId = RaftGroupMemberId.valueOf(id, group.getGroupId());
     this.server = server;
     RaftConfiguration initialConf = RaftConfiguration.newBuilder()
         .setConf(group.getPeers()).build();
     configurationManager = new ConfigurationManager(initialConf);
-    LOG.info("{}:{} {}", id, groupId, configurationManager);
+    LOG.info("{}: {}", getMemberId(), configurationManager);
 
     // use full uuid string to create a subdirectory
     final File dir = chooseStorageDir(RaftServerConfigKeys.storageDirs(prop),
         group.getGroupId().getUuid().toString());
-    storage = new RaftStorage(dir, RaftServerConstants.StartupOption.REGULAR);
+    storage = new RaftStorage(dir, RaftServerConstants.StartupOption.REGULAR,
+        RaftServerConfigKeys.Log.corruptionPolicy(prop));
     snapshotManager = new SnapshotManager(storage, id);
 
     long lastApplied = initStatemachine(stateMachine, group.getGroupId());
@@ -111,21 +114,22 @@ public class ServerState implements Closeable {
     // On start the leader is null, start the clock now
     leaderId = null;
     this.lastNoLeaderTime = Timestamp.currentTime();
-    this.leaderElectionTimeoutMs =
-        RaftServerConfigKeys.leaderElectionTimeout(prop).toIntExact(TimeUnit.MILLISECONDS);
+    this.noLeaderTimeout = RaftServerConfigKeys.Notification.noLeaderTimeout(prop);
 
     // we cannot apply log entries to the state machine in this step, since we
     // do not know whether the local log entries have been committed.
-    log = initLog(id, prop, lastApplied, this::setRaftConf);
+    this.log = initRaftLog(getMemberId(), server, storage, this::setRaftConf, lastApplied, prop);
 
     RaftLog.Metadata metadata = log.loadMetadata();
     currentTerm.set(metadata.getTerm());
     votedFor = metadata.getVotedFor();
 
-    stateMachineUpdater = new StateMachineUpdater(stateMachine, server, log,
-         lastApplied, prop);
+    stateMachineUpdater = new StateMachineUpdater(stateMachine, server, this, lastApplied, prop);
   }
 
+  RaftGroupMemberId getMemberId() {
+    return memberId;
+  }
 
   static File chooseStorageDir(List<File> volumes, String targetSubDir) throws IOException {
     final Map<File, Integer> numberOfStorageDirPerVolume = new HashMap<>();
@@ -154,11 +158,10 @@ public class ServerState implements Closeable {
   private long initStatemachine(StateMachine sm, RaftGroupId gid)
       throws IOException {
     sm.initialize(server.getProxy(), gid, storage);
-    storage.setStateMachineStorage(sm.getStateMachineStorage());
     SnapshotInfo snapshot = sm.getLatestSnapshot();
 
     if (snapshot == null || snapshot.getTermIndex().getIndex() < 0) {
-      return RaftServerConstants.INVALID_LOG_INDEX;
+      return RaftLog.INVALID_LOG_INDEX;
     }
 
     // get the raft configuration from raft metafile
@@ -177,34 +180,22 @@ public class ServerState implements Closeable {
     stateMachineUpdater.start();
   }
 
-  /**
-   * note we do not apply log entries to the state machine here since we do not
-   * know whether they have been committed.
-   */
-  private RaftLog initLog(RaftPeerId id, RaftProperties prop,
-      long lastIndexInSnapshot, Consumer<LogEntryProto> logConsumer)
-      throws IOException {
+  private static RaftLog initRaftLog(RaftGroupMemberId memberId, RaftServerImpl server, RaftStorage storage,
+      Consumer<LogEntryProto> logConsumer, long lastIndexInSnapshot, RaftProperties prop) throws IOException {
     final RaftLog log;
     if (RaftServerConfigKeys.Log.useMemory(prop)) {
-      final int maxBufferSize =
-          RaftServerConfigKeys.Log.Appender.bufferByteLimit(prop).getSizeInt();
-      log = new MemoryRaftLog(id, lastIndexInSnapshot, maxBufferSize);
+      log = new MemoryRaftLog(memberId, lastIndexInSnapshot, prop);
     } else {
-      log = new SegmentedRaftLog(id, server, this.storage,
-          lastIndexInSnapshot, prop);
+      log = new SegmentedRaftLog(memberId, server, storage, lastIndexInSnapshot, prop);
     }
     log.open(lastIndexInSnapshot, logConsumer);
     return log;
   }
 
-  public RaftConfiguration getRaftConf() {
+  RaftConfiguration getRaftConf() {
     return configurationManager.getCurrent();
   }
 
-  public RaftPeerId getSelfId() {
-    return this.selfId;
-  }
-
   public long getCurrentTerm() {
     return currentTerm.get();
   }
@@ -231,7 +222,7 @@ public class ServerState implements Closeable {
    * Become a candidate and start leader election
    */
   long initElection() {
-    votedFor = selfId;
+    votedFor = getMemberId().getPeerId();
     setLeader(null, "initElection");
     return currentTerm.incrementAndGet();
   }
@@ -259,24 +250,28 @@ public class ServerState implements Closeable {
         Timestamp previous = lastNoLeaderTime;
         lastNoLeaderTime = null;
         suffix = ", leader elected after " + previous.elapsedTimeMs() + "ms";
+        server.getStateMachine().notifyLeaderChanged(getMemberId(), newLeaderId);
       }
-      LOG.info("{}:{} change Leader from {} to {} at term {} for {}{}",
-          selfId, groupId, leaderId, newLeaderId, getCurrentTerm(), op, suffix);
+      LOG.info("{}: change Leader from {} to {} at term {} for {}{}",
+          getMemberId(), leaderId, newLeaderId, getCurrentTerm(), op, suffix);
       leaderId = newLeaderId;
     }
   }
 
-  boolean checkForExtendedNoLeader() {
-    return getLastLeaderElapsedTimeMs() > leaderElectionTimeoutMs;
+  boolean shouldNotifyExtendedNoLeader() {
+    return Optional.ofNullable(lastNoLeaderTime)
+        .map(Timestamp::elapsedTime)
+        .filter(t -> t.compareTo(noLeaderTimeout) > 0)
+        .isPresent();
   }
 
-  long getLastLeaderElapsedTimeMs() {
+  public long getLastLeaderElapsedTimeMs() {
     final Timestamp t = lastNoLeaderTime;
     return t == null ? 0 : t.elapsedTimeMs();
   }
 
   void becomeLeader() {
-    setLeader(selfId, "becomeLeader");
+    setLeader(getMemberId().getPeerId(), "becomeLeader");
   }
 
   public RaftLog getLog() {
@@ -340,7 +335,7 @@ public class ServerState implements Closeable {
 
   @Override
   public String toString() {
-    return selfId + ":t" + currentTerm + ", leader=" + leaderId
+    return getMemberId() + ":t" + currentTerm + ", leader=" + leaderId
         + ", voted=" + votedFor + ", raftlog=" + log + ", conf=" + getRaftConf();
   }
 
@@ -358,9 +353,8 @@ public class ServerState implements Closeable {
   void setRaftConf(long logIndex, RaftConfiguration conf) {
     configurationManager.addConfiguration(logIndex, conf);
     server.getServerRpc().addPeers(conf.getPeers());
-    LOG.info("{}:{} set configuration {} at {}", getSelfId(), groupId, conf,
-        logIndex);
-    LOG.trace("{}: {}", getSelfId(), configurationManager);
+    LOG.info("{}: set configuration {} at {}", getMemberId(), conf, logIndex);
+    LOG.trace("{}: {}", getMemberId(), configurationManager);
   }
 
   void updateConfiguration(LogEntryProto[] entries) {
@@ -388,11 +382,9 @@ public class ServerState implements Closeable {
     try {
       stateMachineUpdater.stopAndJoin();
     } catch (InterruptedException e) {
-      LOG.warn(getSelfId() +
-          ": Interrupted when joining stateMachineUpdater", e);
+      LOG.warn("{}: Interrupted when joining stateMachineUpdater", getMemberId(), e);
     }
-    LOG.info("{}:{} closes. The last applied log index is {}",
-        getSelfId(), groupId, getLastAppliedIndex());
+    LOG.info("{}: closes. applyIndex: {}", getMemberId(), getLastAppliedIndex());
 
     log.close();
     storage.close();
@@ -407,20 +399,52 @@ public class ServerState implements Closeable {
     StateMachine sm = server.getStateMachine();
     sm.pause(); // pause the SM to prepare for install snapshot
     snapshotManager.installSnapshot(sm, request);
-    log.syncWithSnapshot(request.getTermIndex().getIndex());
-    this.latestInstalledSnapshot = ServerProtoUtils.toTermIndex(
-        request.getTermIndex());
+    updateInstalledSnapshotIndex(ServerProtoUtils.toTermIndex(request.getSnapshotChunk().getTermIndex()));
+  }
+
+  void updateInstalledSnapshotIndex(TermIndex lastTermIndexInSnapshot) {
+    log.syncWithSnapshot(lastTermIndexInSnapshot.getIndex());
+    latestInstalledSnapshot.set(lastTermIndexInSnapshot);
   }
 
   SnapshotInfo getLatestSnapshot() {
-    return server.getStateMachine().getStateMachineStorage().getLatestSnapshot();
+    return server.getStateMachine().getLatestSnapshot();
   }
 
-  public TermIndex getLatestInstalledSnapshot() {
-    return latestInstalledSnapshot;
+  public long getLatestInstalledSnapshotIndex() {
+    final TermIndex ti = latestInstalledSnapshot.get();
+    return ti != null? ti.getIndex(): 0L;
+  }
+
+  /**
+   * The last index included in either the latestSnapshot or the latestInstalledSnapshot
+   * @return the last snapshot index
+   */
+  long getSnapshotIndex() {
+    final SnapshotInfo s = getLatestSnapshot();
+    final long latestSnapshotIndex = s != null ? s.getIndex() : 0;
+    return Math.max(latestSnapshotIndex, getLatestInstalledSnapshotIndex());
+  }
+
+  public long getNextIndex() {
+    final long logNextIndex = log.getNextIndex();
+    final long snapshotNextIndex = getSnapshotIndex() + 1;
+    return Math.max(logNextIndex, snapshotNextIndex);
   }
 
   public long getLastAppliedIndex() {
-    return stateMachineUpdater.getLastAppliedIndex();
+    return stateMachineUpdater.getStateMachineLastAppliedIndex();
+  }
+
+  boolean containsTermIndex(TermIndex ti) {
+    Objects.requireNonNull(ti, "ti == null");
+
+    if (Optional.ofNullable(latestInstalledSnapshot.get()).filter(ti::equals).isPresent()) {
+      return true;
+    }
+    if (Optional.ofNullable(getLatestSnapshot()).map(SnapshotInfo::getTermIndex).filter(ti::equals).isPresent()) {
+      return true;
+    }
+    return log.contains(ti);
   }
 }
