@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one
  * or more contributor license agreements.  See the NOTICE file
  * distributed with this work for additional information
@@ -18,12 +18,11 @@
 package org.apache.ratis.server.storage;
 
 import org.apache.ratis.proto.RaftProtos.LogEntryProto;
+import org.apache.ratis.server.RaftServerConfigKeys.Log.CorruptionPolicy;
 import org.apache.ratis.server.impl.RaftConfiguration;
 import org.apache.ratis.server.impl.RaftServerConstants;
 import org.apache.ratis.server.impl.ServerProtoUtils;
 import org.apache.ratis.server.storage.RaftStorageDirectory.StorageState;
-import org.apache.ratis.statemachine.SnapshotInfo;
-import org.apache.ratis.statemachine.StateMachineStorage;
 import org.apache.ratis.util.Preconditions;
 import org.slf4j.Logger;
 import org.slf4j.LoggerFactory;
@@ -41,12 +40,17 @@ public class RaftStorage implements Closeable {
   // TODO support multiple storage directories
   private final RaftStorageDirectory storageDir;
   private final StorageState state;
+  private final CorruptionPolicy logCorruptionPolicy;
   private volatile MetaFile metaFile;
-  private StateMachineStorage stateMachineStorage;
 
   public RaftStorage(File dir, RaftServerConstants.StartupOption option)
       throws IOException {
-    storageDir = new RaftStorageDirectory(dir);
+    this(dir, option, CorruptionPolicy.getDefault());
+  }
+
+  public RaftStorage(File dir, RaftServerConstants.StartupOption option, CorruptionPolicy logCorruptionPolicy)
+      throws IOException {
+    this.storageDir = new RaftStorageDirectory(dir);
     if (option == RaftServerConstants.StartupOption.FORMAT) {
       if (storageDir.analyzeStorage(false) == StorageState.NON_EXISTENT) {
         throw new IOException("Cannot format " + storageDir);
@@ -63,12 +67,17 @@ public class RaftStorage implements Closeable {
             + ". Its state: " + state);
       }
     }
+    this.logCorruptionPolicy = logCorruptionPolicy;
   }
 
   StorageState getState() {
     return state;
   }
 
+  public CorruptionPolicy getLogCorruptionPolicy() {
+    return logCorruptionPolicy;
+  }
+
   private void format() throws IOException {
     storageDir.clearDirectory();
     metaFile = writeMetaFile(MetaFile.DEFAULT_TERM, MetaFile.EMPTY_VOTEFOR);
@@ -116,7 +125,7 @@ public class RaftStorage implements Closeable {
     storageDir.unlock();
   }
 
-  MetaFile getMetaFile() {
+  public MetaFile getMetaFile() {
     return metaFile;
   }
 
@@ -140,23 +149,8 @@ public class RaftStorage implements Closeable {
     }
   }
 
-  public SnapshotInfo getLatestSnapshot() throws IOException {
-    return getStateMachineStorage().getLatestSnapshot();
-  }
-
-  /**
-   * Called by the state machine after it has initialized the StateMachineStorage.
-   */
-  public void setStateMachineStorage(StateMachineStorage smStorage) {
-    this.stateMachineStorage = smStorage;
-  }
-
-  public StateMachineStorage getStateMachineStorage() {
-    return stateMachineStorage;
-  }
-
   @Override
   public String toString() {
-    return getStorageDir() + "";
+    return getClass().getSimpleName() + ":" + getStorageDir();
   }
 }
