@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one
  * or more contributor license agreements.  See the NOTICE file
  * distributed with this work for additional information
@@ -18,6 +18,7 @@
 package org.apache.ratis.server.storage;
 
 import org.apache.ratis.server.impl.RaftServerConstants;
+import org.apache.ratis.server.raftlog.RaftLog;
 import org.apache.ratis.util.AtomicFileOutputStream;
 import org.apache.ratis.util.FileUtils;
 import org.slf4j.Logger;
@@ -38,7 +39,6 @@ import java.util.regex.Matcher;
 import java.util.regex.Pattern;
 
 import static java.nio.file.Files.newDirectoryStream;
-import static org.apache.ratis.server.impl.RaftServerConstants.INVALID_LOG_INDEX;
 
 public class RaftStorageDirectory {
   static final Logger LOG = LoggerFactory.getLogger(RaftStorageDirectory.class);
@@ -153,7 +153,7 @@ public class RaftStorageDirectory {
     return new File(getCurrentDir(), META_FILE_NAME + CONF_EXTENSION);
   }
 
-  File getOpenLogFile(long startIndex) {
+  public File getOpenLogFile(long startIndex) {
     return new File(getCurrentDir(), getOpenLogFileName(startIndex));
   }
 
@@ -161,7 +161,7 @@ public class RaftStorageDirectory {
     return LOG_FILE_PREFIX + "_" + LOG_FILE_INPROGRESS + "_" + startIndex;
   }
 
-  File getClosedLogFile(long startIndex, long endIndex) {
+  public File getClosedLogFile(long startIndex, long endIndex) {
     return new File(getCurrentDir(), getClosedLogFileName(startIndex, endIndex));
   }
 
@@ -193,20 +193,9 @@ public class RaftStorageDirectory {
     try (DirectoryStream<Path> stream =
              Files.newDirectoryStream(getCurrentDir().toPath())) {
       for (Path path : stream) {
-        for (Pattern pattern : Arrays.asList(CLOSED_SEGMENT_REGEX, OPEN_SEGMENT_REGEX)) {
-          Matcher matcher = pattern.matcher(path.getFileName().toString());
-          if (matcher.matches()) {
-            if (pattern == OPEN_SEGMENT_REGEX && Files.size(path) == 0L) {
-              Files.delete(path);
-              LOG.info("Delete zero size file " + path);
-              break;
-            }
-            final long startIndex = Long.parseLong(matcher.group(1));
-            final long endIndex = matcher.groupCount() == 2 ?
-                Long.parseLong(matcher.group(2)) : INVALID_LOG_INDEX;
-            list.add(new LogPathAndIndex(path, startIndex, endIndex));
-            break;
-          }
+        LogPathAndIndex lpi = processOnePath(path);
+        if (lpi != null) {
+          list.add(lpi);
         }
       }
     }
@@ -214,6 +203,24 @@ public class RaftStorageDirectory {
     return list;
   }
 
+  public static LogPathAndIndex processOnePath(Path path) throws IOException {
+    for (Pattern pattern : Arrays.asList(CLOSED_SEGMENT_REGEX, OPEN_SEGMENT_REGEX)) {
+      Matcher matcher = pattern.matcher(path.getFileName().toString());
+      if (matcher.matches()) {
+        if (pattern == OPEN_SEGMENT_REGEX && Files.size(path) == 0L) {
+          Files.delete(path);
+          LOG.info("Delete zero size file " + path);
+          return null;
+        }
+        final long startIndex = Long.parseLong(matcher.group(1));
+        final long endIndex = matcher.groupCount() == 2 ?
+            Long.parseLong(matcher.group(2)) : RaftLog.INVALID_LOG_INDEX;
+        return new LogPathAndIndex(path, startIndex, endIndex);
+      }
+    }
+    return null;
+  }
+
   /**
    * Check to see if current/ directory is empty.
    */
@@ -288,7 +295,8 @@ public class RaftStorageDirectory {
    * @throws IOException if locking fails
    */
   public void lock() throws IOException {
-    FileLock newLock = tryLock();
+    final File lockF = new File(root, STORAGE_FILE_LOCK);
+    final FileLock newLock = FileUtils.attempt(() -> tryLock(lockF), () -> "tryLock " + lockF);
     if (newLock == null) {
       String msg = "Cannot lock storage " + this.root
           + ". The directory is already locked";
@@ -308,9 +316,8 @@ public class RaftStorageDirectory {
    * <code>null</code> if storage is already locked.
    * @throws IOException if locking fails.
    */
-  private FileLock tryLock() throws IOException {
+  private FileLock tryLock(File lockF) throws IOException {
     boolean deletionHookAdded = false;
-    File lockF = new File(root, STORAGE_FILE_LOCK);
     if (!lockF.exists()) {
       lockF.deleteOnExit();
       deletionHookAdded = true;
