@@ -17,13 +17,15 @@
  */
 package org.apache.ratis.statemachine;
 
+import org.apache.ratis.proto.RaftProtos;
 import org.apache.ratis.protocol.Message;
 import org.apache.ratis.protocol.RaftClientRequest;
-import org.apache.ratis.protocol.RaftGroup;
 import org.apache.ratis.protocol.RaftGroupId;
+import org.apache.ratis.protocol.RaftGroupMemberId;
+import org.apache.ratis.protocol.RaftPeerId;
 import org.apache.ratis.server.RaftServer;
 import org.apache.ratis.server.RaftServerConfigKeys;
-import org.apache.ratis.server.impl.RaftConfiguration;
+import org.apache.ratis.server.impl.ServerProtoUtils;
 import org.apache.ratis.server.protocol.TermIndex;
 import org.apache.ratis.server.storage.RaftStorage;
 import org.apache.ratis.proto.RaftProtos.RoleInfoProto;
@@ -136,8 +138,11 @@ public interface StateMachine extends Closeable {
       throws IOException;
 
   /**
-   * Write asynchronously the state machine data to this state machine.
-   *
+   * Write asynchronously the state machine data to this state machine. The call
+   * to stateMachine is made with the RaftLog lock held. All the stateMachineData
+   * calls are sequential and in order of the corresponding log operations.
+   * The operation must be lightweight and should be used to determine the task
+   * to be performed. The task if heavy should be performed inside the CompletableFuture.
    * @return a future for the write task if the state machine data should be sync'ed;
    *         otherwise, return null.
    */
@@ -216,10 +221,19 @@ public interface StateMachine extends Closeable {
    * Notify the Leader's state machine that one of the followers is slow
    * this notification is based on "raft.server.rpc.slowness.timeout"
    *
-   * @param group raft group information
    * @param roleInfoProto information about the current node role and rpc delay information
    */
-  default void notifySlowness(RaftGroup group, RoleInfoProto roleInfoProto) {
+  default void notifySlowness(RoleInfoProto roleInfoProto) {
+
+  }
+
+  /**
+   * Notify the state machine that the pipeline has failed.
+   * This notification is triggered when a log operation throws an Exception.
+   * @param t Exception which was caught, indicates possible cause.
+   * @param failedEntry if append failed for a specific entry, null otherwise.
+   */
+  default void notifyLogFailed(Throwable t, LogEntryProto failedEntry) {
 
   }
 
@@ -227,10 +241,9 @@ public interface StateMachine extends Closeable {
    * Notify the Leader's state machine that a leader has not been elected for a long time
    * this notification is based on "raft.server.leader.election.timeout"
    *
-   * @param group raft group information
    * @param roleInfoProto information about the current node role and rpc delay information
    */
-  default void notifyExtendedNoLeader(RaftGroup group, RoleInfoProto roleInfoProto) {
+  default void notifyExtendedNoLeader(RoleInfoProto roleInfoProto) {
 
   }
 
@@ -246,7 +259,11 @@ public interface StateMachine extends Closeable {
   /**
    * Truncates asynchronously the associated state machine data starting from the given log
    * index from the state machine. It will be a no op if the corresponding log entry does not
-   * have associated stateMachineData.
+   * have associated stateMachineData. The call to stateMachine is made with the RaftLog
+   * lock held. All the stateMachineData calls are sequential and in order of the
+   * corresponding log operations. The operation must be lightweight and should be used
+   * to determine the truncation task to be performed. The task if heavy should
+   * be performed inside the CompletableFuture.
    * @param index log Index starting from which the stateMachineData will be truncated.
    * @return a combined future for the remove task for all the log entries starting from
    *         given logIndex, null otherwise
@@ -254,4 +271,43 @@ public interface StateMachine extends Closeable {
   default CompletableFuture<Void> truncateStateMachineData(long index) {
     return CompletableFuture.completedFuture(null);
   }
+
+  /**
+   * Notify the Follower's state machine that the leader has purged entries
+   * from its log and hence to catch up, the Follower state machine would have
+   * to install the latest snapshot.
+   * @param firstTermIndexInLog TermIndex of the first append entry available
+   *                           in the Leader's log.
+   * @param roleInfoProto information about the current node role and
+   *                            rpc delay information
+   * @return After the snapshot installation is complete, return the last
+   * included term index in the snapshot.
+   */
+  default CompletableFuture<TermIndex> notifyInstallSnapshotFromLeader(
+      RoleInfoProto roleInfoProto, TermIndex firstTermIndexInLog) {
+    return CompletableFuture.completedFuture(null);
+  }
+
+  /**
+   * Notify the state machine that a RaftPeer has been elected as leader.
+   */
+  default void notifyLeaderChanged(RaftGroupMemberId groupMemberId, RaftPeerId raftPeerId) {
+  }
+
+  /**
+   * Notify about group removal in the state machine. This function is called
+   * during group removal after all the pending transactions have been applied
+   * by the state machine.
+   */
+  default void notifyGroupRemove() {
+  }
+
+  /**
+   * Converts the proto object into a useful log string to add information about state machine data.
+   * @param proto state machine proto
+   * @return
+   */
+  default String toStateMachineLogEntryString(RaftProtos.StateMachineLogEntryProto proto) {
+    return ServerProtoUtils.toStateMachineLogEntryString(proto, null);
+  }
 }
