@@ -18,6 +18,8 @@
 
 package org.apache.ratis.statemachine.impl;
 
+import com.codahale.metrics.Timer;
+import org.apache.ratis.proto.RaftProtos;
 import org.apache.ratis.protocol.Message;
 import org.apache.ratis.protocol.RaftClientRequest;
 import org.apache.ratis.protocol.RaftGroupId;
@@ -35,6 +37,7 @@ import org.apache.ratis.util.Preconditions;
 
 import java.io.IOException;
 import java.util.Collection;
+import java.util.Objects;
 import java.util.SortedMap;
 import java.util.TreeMap;
 import java.util.concurrent.CompletableFuture;
@@ -44,14 +47,18 @@ import java.util.concurrent.atomic.AtomicReference;
  * Base implementation for StateMachines.
  */
 public class BaseStateMachine implements StateMachine {
-  private final CompletableFuture<RaftServer> server = new CompletableFuture<>();
-  private volatile RaftGroupId groupId;
+  protected final CompletableFuture<RaftServer> server = new CompletableFuture<>();
+  protected volatile RaftGroupId groupId;
   protected final LifeCycle lifeCycle = new LifeCycle(getClass().getSimpleName());
 
   private final AtomicReference<TermIndex> lastAppliedTermIndex = new AtomicReference<>();
 
   private final SortedMap<Long, CompletableFuture<Void>> transactionFutures = new TreeMap<>();
 
+  public BaseStateMachine() {
+    setLastAppliedTermIndex(TermIndex.newTermIndex(0, -1));
+  }
+
   public RaftPeerId getId() {
     return server.isDone()? server.join().getId(): null;
   }
@@ -94,6 +101,8 @@ public class BaseStateMachine implements StateMachine {
   @Override
   public CompletableFuture<Message> applyTransaction(TransactionContext trx) {
     // return the same message contained in the entry
+    RaftProtos.LogEntryProto entry = Objects.requireNonNull(trx.getLogEntry());
+    updateLastAppliedTermIndex(entry.getTerm(), entry.getIndex());
     return CompletableFuture.completedFuture(
         Message.valueOf(trx.getLogEntry().getStateMachineLogEntry().getLogData()));
   }
@@ -107,6 +116,11 @@ public class BaseStateMachine implements StateMachine {
     lastAppliedTermIndex.set(newTI);
   }
 
+  @Override
+  public void notifyIndexUpdate(long term, long index) {
+    updateLastAppliedTermIndex(term, index);
+  }
+
   protected boolean updateLastAppliedTermIndex(long term, long index) {
     final TermIndex newTI = TermIndex.newTermIndex(term, index);
     final TermIndex oldTI = lastAppliedTermIndex.getAndSet(newTI);
@@ -148,6 +162,10 @@ public class BaseStateMachine implements StateMachine {
       @Override
       public void format() throws IOException {
       }
+
+      @Override
+      public void cleanupOldSnapshots(SnapshotRetentionPolicy snapshotRetentionPolicy) {
+      }
     };
   }
 
@@ -197,4 +215,19 @@ public class BaseStateMachine implements StateMachine {
     return getClass().getSimpleName() + ":"
         + (!server.isDone()? "uninitialized": getId() + ":" + groupId);
   }
+
+
+  protected CompletableFuture<Message> recordTime(Timer timer, Task task) {
+    final Timer.Context timerContext = timer.time();
+    try {
+      return task.run();
+    } finally {
+      timerContext.stop();
+    }
+  }
+
+  protected interface Task {
+    CompletableFuture<Message> run();
+  }
+
 }
