@@ -42,7 +42,7 @@ import org.apache.ratis.statemachine.StateMachine;
 import org.apache.ratis.thirdparty.com.google.protobuf.ByteString;
 import org.apache.ratis.thirdparty.com.google.protobuf.InvalidProtocolBufferException;
 import org.apache.ratis.util.JavaUtils;
-import org.apache.ratis.util.LogUtils;
+import org.apache.ratis.util.Log4jUtils;
 import org.apache.ratis.util.TimeDuration;
 import org.apache.ratis.util.function.CheckedRunnable;
 import org.junit.Assert;
@@ -57,6 +57,7 @@ import java.util.concurrent.CompletableFuture;
 import java.util.concurrent.CompletionException;
 import java.util.concurrent.ExecutionException;
 import java.util.concurrent.TimeUnit;
+import java.util.concurrent.TimeoutException;
 import java.util.concurrent.atomic.AtomicInteger;
 
 import static org.apache.ratis.RaftTestUtil.waitForLeader;
@@ -64,8 +65,8 @@ import static org.apache.ratis.RaftTestUtil.waitForLeader;
 public abstract class RaftAsyncTests<CLUSTER extends MiniRaftCluster> extends BaseTest
     implements MiniRaftCluster.Factory.Get<CLUSTER> {
   {
-    LogUtils.setLogLevel(RaftServerImpl.LOG, Level.DEBUG);
-    LogUtils.setLogLevel(RaftClient.LOG, Level.DEBUG);
+    Log4jUtils.setLogLevel(RaftServerImpl.LOG, Level.DEBUG);
+    Log4jUtils.setLogLevel(RaftClient.LOG, Level.DEBUG);
   }
 
   public static final int NUM_SERVERS = 3;
@@ -82,19 +83,14 @@ public abstract class RaftAsyncTests<CLUSTER extends MiniRaftCluster> extends Ba
     RaftClient.Builder clientBuilder = RaftClient.newBuilder()
         .setRaftGroup(RaftGroup.emptyGroup())
         .setProperties(properties);
-    int numThreads = RaftClientConfigKeys.Async.SCHEDULER_THREADS_DEFAULT;
     int maxOutstandingRequests = RaftClientConfigKeys.Async.MAX_OUTSTANDING_REQUESTS_DEFAULT;
     try(RaftClient client = clientBuilder.build()) {
-      RaftClientTestUtil.assertScheduler(client, numThreads);
       RaftClientTestUtil.assertAsyncRequestSemaphore(client, maxOutstandingRequests, 0);
     }
 
-    numThreads = 200;
     maxOutstandingRequests = 5;
     RaftClientConfigKeys.Async.setMaxOutstandingRequests(properties, maxOutstandingRequests);
-    RaftClientConfigKeys.Async.setSchedulerThreads(properties, numThreads);
     try(RaftClient client = clientBuilder.build()) {
-      RaftClientTestUtil.assertScheduler(client, numThreads);
       RaftClientTestUtil.assertAsyncRequestSemaphore(client, maxOutstandingRequests, 0);
     }
   }
@@ -107,16 +103,23 @@ public abstract class RaftAsyncTests<CLUSTER extends MiniRaftCluster> extends Ba
 
   @Test
   public void testRequestAsyncWithRetryFailure() throws Exception {
-    runWithNewCluster(1, false, cluster -> runTestRequestAsyncWithRetryFailure(false, cluster));
+    runTestRequestAsyncWithRetryFailure(false);
   }
 
   @Test
   public void testRequestAsyncWithRetryFailureAfterInitialMessages() throws Exception {
-    runWithNewCluster(1, true, cluster -> runTestRequestAsyncWithRetryFailure(true, cluster));
+    runTestRequestAsyncWithRetryFailure(true);
+  }
+
+  void runTestRequestAsyncWithRetryFailure(boolean initialMessages) throws Exception {
+    RaftClientConfigKeys.Async.Experimental.setSendDummyRequest(getProperties(), false);
+    runWithNewCluster(1, initialMessages, cluster -> runTestRequestAsyncWithRetryFailure(initialMessages, cluster));
+    RaftClientConfigKeys.Async.Experimental.setSendDummyRequest(getProperties(), true);
   }
 
   void runTestRequestAsyncWithRetryFailure(boolean initialMessages, CLUSTER cluster) throws Exception {
-    final RetryLimited retryPolicy = RetryPolicies.retryUpToMaximumCountWithFixedSleep(10, HUNDRED_MILLIS);
+    final TimeDuration sleepTime = HUNDRED_MILLIS;
+    final RetryLimited retryPolicy = RetryPolicies.retryUpToMaximumCountWithFixedSleep(10, sleepTime);
 
     try(final RaftClient client = cluster.createClient(null, retryPolicy)) {
       RaftPeerId leader = null;
@@ -147,7 +150,7 @@ public abstract class RaftAsyncTests<CLUSTER extends MiniRaftCluster> extends Ba
         }
 
         // sleep most of the retry time
-        retryPolicy.getSleepTime().apply(t -> t * (retryPolicy.getMaxAttempts() - 1)).sleep();
+        sleepTime.apply(t -> t * (retryPolicy.getMaxAttempts() - 1)).sleep();
 
         // send another half of the calls without starting the cluster
         for (; i < messages.length; i++) {
@@ -158,7 +161,7 @@ public abstract class RaftAsyncTests<CLUSTER extends MiniRaftCluster> extends Ba
 
       // sleep again so that the first half calls will fail retries.
       // the second half still have retry time remaining.
-      retryPolicy.getSleepTime().apply(t -> t*2).sleep();
+      sleepTime.apply(t -> t*2).sleep();
 
       if (leader != null) {
         cluster.restartServer(leader, false);
@@ -257,7 +260,7 @@ public abstract class RaftAsyncTests<CLUSTER extends MiniRaftCluster> extends Ba
   @Test
   public void testWithLoadAsync() throws Exception {
     runWithNewCluster(NUM_SERVERS,
-        cluster -> RaftBasicTests.testWithLoad(10, 500, true, cluster, LOG));
+        cluster -> RaftBasicTests.testWithLoad(5, 500, true, cluster, LOG));
   }
 
   @Test
@@ -346,11 +349,19 @@ public abstract class RaftAsyncTests<CLUSTER extends MiniRaftCluster> extends Ba
   @Test
   public void testRequestTimeout() throws Exception {
     final TimeDuration oldExpiryTime = RaftServerConfigKeys.RetryCache.expiryTime(getProperties());
-    RaftServerConfigKeys.RetryCache.setExpiryTime(getProperties(), TimeDuration.valueOf(5, TimeUnit.SECONDS));
+    RaftServerConfigKeys.RetryCache.setExpiryTime(getProperties(), FIVE_SECONDS);
+    RaftClientConfigKeys.Async.Experimental.setSendDummyRequest(getProperties(), false);
     runWithNewCluster(NUM_SERVERS, cluster -> RaftBasicTests.testRequestTimeout(true, cluster, LOG));
 
     //reset for the other tests
     RaftServerConfigKeys.RetryCache.setExpiryTime(getProperties(), oldExpiryTime);
+    RaftClientConfigKeys.Async.Experimental.setSendDummyRequest(getProperties(), true);
+  }
+
+  @Test
+  public void testStateMachineMetrics() throws Exception {
+    runWithNewCluster(NUM_SERVERS, cluster ->
+        RaftBasicTests.testStateMachineMetrics(true, cluster, LOG));
   }
 
   @Test
@@ -375,18 +386,53 @@ public abstract class RaftAsyncTests<CLUSTER extends MiniRaftCluster> extends Ba
       CompletableFuture<RaftClientReply> replyFuture = client.sendAsync(new SimpleMessage("abc"));
       Thread.sleep(waitTime);
       // replyFuture should not be completed until append request is unblocked.
-      Assert.assertTrue(!replyFuture.isDone());
+      Assert.assertFalse(replyFuture.isDone());
       // unblock append request.
       cluster.getServerAliveStream()
           .filter(impl -> !impl.isLeader())
           .map(SimpleStateMachine4Testing::get)
           .forEach(SimpleStateMachine4Testing::unblockWriteStateMachineData);
 
-      replyFuture.get();
+      Assert.assertTrue(replyFuture.get().isSuccess());
       Assert.assertTrue(System.currentTimeMillis() - time > waitTime);
     }
 
     //reset for the other tests
     RaftServerConfigKeys.RetryCache.setExpiryTime(getProperties(), oldExpiryTime);
   }
+
+  @Test
+  public void testNoRetryWaitOnNotLeaderException() throws Exception {
+    RaftClientConfigKeys.Async.Experimental.setSendDummyRequest(getProperties(), false);
+    runWithNewCluster(3, this::runTestNoRetryWaitOnNotLeaderException);
+    RaftClientConfigKeys.Async.Experimental.setSendDummyRequest(getProperties(), true);
+  }
+
+  private void runTestNoRetryWaitOnNotLeaderException(MiniRaftCluster cluster) throws Exception {
+    final RaftServerImpl leader = waitForLeader(cluster);
+    final List<RaftServerImpl> followers = cluster.getFollowers();
+    Assert.assertNotNull(followers);
+    Assert.assertEquals(2, followers.size());
+    Assert.assertNotSame(leader, followers.get(0));
+    Assert.assertNotSame(leader, followers.get(1));
+
+    // send a message to make sure that the leader is ready
+    try (final RaftClient client = cluster.createClient(leader.getId())) {
+      final CompletableFuture<RaftClientReply> f = client.sendAsync(new SimpleMessage("first"));
+      FIVE_SECONDS.apply(f::get);
+    }
+
+    final RetryPolicy r = event -> () -> {
+      final IllegalStateException e = new IllegalStateException("Unexpected getSleepTime: " + event);
+      setFirstException(e);
+      throw e;
+    };
+
+    try (final RaftClient client = cluster.createClient(followers.get(0).getId(), cluster.getGroup(), r)) {
+      final CompletableFuture<RaftClientReply> f = client.sendAsync(new SimpleMessage("abc"));
+      FIVE_SECONDS.apply(f::get);
+    } catch (TimeoutException e) {
+      throw new AssertionError("Failed to get async result", e);
+    }
+  }
 }
