@@ -24,10 +24,10 @@ import org.apache.ratis.client.RaftClientRpc;
 import org.apache.ratis.protocol.*;
 import org.apache.ratis.server.RaftServerConfigKeys;
 import org.apache.ratis.server.impl.RaftServerImpl;
-import org.apache.ratis.server.storage.RaftLog;
-import org.apache.ratis.server.storage.RaftLogIOException;
+import org.apache.ratis.server.raftlog.RaftLog;
+import org.apache.ratis.server.raftlog.RaftLogIOException;
 import org.apache.ratis.util.JavaUtils;
-import org.apache.ratis.util.LogUtils;
+import org.apache.ratis.util.Log4jUtils;
 import org.apache.ratis.util.SizeInBytes;
 import org.junit.Assert;
 import org.junit.Assume;
@@ -42,9 +42,9 @@ public abstract class RaftExceptionBaseTest<CLUSTER extends MiniRaftCluster>
     extends BaseTest
     implements MiniRaftCluster.Factory.Get<CLUSTER> {
   static {
-    LogUtils.setLogLevel(RaftServerImpl.LOG, Level.DEBUG);
-    LogUtils.setLogLevel(RaftLog.LOG, Level.DEBUG);
-    LogUtils.setLogLevel(RaftClient.LOG, Level.DEBUG);
+    Log4jUtils.setLogLevel(RaftServerImpl.LOG, Level.DEBUG);
+    Log4jUtils.setLogLevel(RaftLog.LOG, Level.DEBUG);
+    Log4jUtils.setLogLevel(RaftClient.LOG, Level.DEBUG);
   }
 
   static final int NUM_PEERS = 3;
@@ -80,7 +80,7 @@ public abstract class RaftExceptionBaseTest<CLUSTER extends MiniRaftCluster>
       }
 
       final RaftClientRpc rpc = client.getClientRpc();
-      JavaUtils.attempt(() -> assertNotLeaderException(newLeader, "m2", oldLeader, rpc, cluster),
+      JavaUtils.attemptRepeatedly(() -> assertNotLeaderException(newLeader, "m2", oldLeader, rpc, cluster),
           10, ONE_SECOND, "assertNotLeaderException", LOG);
 
       sendMessage("m3", client);
@@ -128,13 +128,13 @@ public abstract class RaftExceptionBaseTest<CLUSTER extends MiniRaftCluster>
 
       // it is possible that the remote peer's rpc server is not ready. need retry
       final RaftClientRpc rpc = client.getClientRpc();
-      final RaftClientReply reply = JavaUtils.attempt(
+      final RaftClientReply reply = JavaUtils.attemptRepeatedly(
           () -> assertNotLeaderException(newLeader, "m1", oldLeader, rpc, cluster),
           10, ONE_SECOND, "assertNotLeaderException", LOG);
 
       final Collection<RaftPeer> peers = cluster.getPeers();
-      final RaftPeer[] peersFromReply = reply.getNotLeaderException().getPeers();
-      Assert.assertEquals(peers.size(), peersFromReply.length);
+      final Collection<RaftPeer> peersFromReply = reply.getNotLeaderException().getPeers();
+      Assert.assertEquals(peers.size(), peersFromReply.size());
       for (RaftPeer p : peersFromReply) {
         Assert.assertTrue(peers.contains(p));
       }
