@@ -33,7 +33,8 @@ import org.apache.ratis.server.impl.DelayLocalExecutionInjection;
 import org.apache.ratis.server.impl.RaftServerImpl;
 import org.apache.ratis.server.impl.ServerProtoUtils;
 import org.apache.ratis.server.protocol.TermIndex;
-import org.apache.ratis.server.storage.RaftLog;
+import org.apache.ratis.server.raftlog.RaftLog;
+import org.apache.ratis.thirdparty.com.google.common.base.Preconditions;
 import org.apache.ratis.thirdparty.com.google.protobuf.ByteString;
 import org.apache.ratis.util.AutoCloseableLock;
 import org.apache.ratis.util.CollectionUtils;
@@ -55,6 +56,7 @@ import java.util.Objects;
 import java.util.Optional;
 import java.util.concurrent.CompletableFuture;
 import java.util.concurrent.TimeUnit;
+import java.util.concurrent.TimeoutException;
 import java.util.concurrent.atomic.AtomicLong;
 import java.util.concurrent.atomic.AtomicReference;
 import java.util.function.BooleanSupplier;
@@ -62,6 +64,7 @@ import java.util.function.Consumer;
 import java.util.function.Function;
 import java.util.function.IntSupplier;
 import java.util.function.Predicate;
+import java.util.function.Supplier;
 
 
 public interface RaftTestUtil {
@@ -94,7 +97,7 @@ public interface RaftTestUtil {
       exception.set(ise);
     };
 
-    final RaftServerImpl leader = JavaUtils.attempt(
+    final RaftServerImpl leader = JavaUtils.attemptRepeatedly(
         () -> cluster.getLeader(groupId, handleNoLeaders, handleMultipleLeaders),
         numAttempts, sleepTime, name, LOG);
 
@@ -119,6 +122,25 @@ public interface RaftTestUtil {
     return leader.getId();
   }
 
+  static void waitFor(Supplier<Boolean> check, int checkEveryMillis,
+      int waitForMillis) throws TimeoutException, InterruptedException {
+    Preconditions.checkNotNull(check);
+    Preconditions.checkArgument(waitForMillis >= checkEveryMillis);
+
+    long st = System.currentTimeMillis();
+    boolean result = check.get();
+
+    while (!result && (System.currentTimeMillis() - st < waitForMillis)) {
+      Thread.sleep(checkEveryMillis);
+      result = check.get();
+    }
+
+    if (!result) {
+      throw new TimeoutException("Timed out waiting for condition.");
+    }
+  }
+
+
   static boolean logEntriesContains(RaftLog log, SimpleMessage... expectedMessages) {
     return logEntriesContains(log, 0L, Long.MAX_VALUE, expectedMessages);
   }
@@ -180,12 +202,12 @@ public interface RaftTestUtil {
   }
 
   static void assertLogEntries(RaftServerImpl server, long expectedTerm, SimpleMessage... expectedMessages) {
-    LOG.info("checking raft log for " + server.getId());
+    LOG.info("checking raft log for {}", server.getMemberId());
     final RaftLog log = server.getState().getLog();
     try {
       RaftTestUtil.assertLogEntries(log, expectedTerm, expectedMessages);
     } catch (AssertionError e) {
-      LOG.error(server.getId() + ": Unexpected raft log", e);
+      LOG.error("Unexpected raft log in {}", server.getMemberId(), e);
       throw e;
     }
   }
@@ -359,7 +381,7 @@ public interface RaftTestUtil {
     final String name = JavaUtils.getCallerStackTraceElement().getMethodName() + "-changeLeader";
     cluster.setBlockRequestsFrom(oldLeader.toString(), true);
     try {
-      return JavaUtils.attempt(() -> {
+      return JavaUtils.attemptRepeatedly(() -> {
         final RaftPeerId newLeader = waitForLeader(cluster).getId();
         if (newLeader.equals(oldLeader)) {
           throw constructor.apply("Failed to change leader: newLeader == oldLeader == " + oldLeader);
