@@ -29,10 +29,10 @@ import org.apache.ratis.protocol.RaftPeer;
 import org.apache.ratis.protocol.RaftPeerId;
 import org.apache.ratis.server.impl.RaftServerImpl;
 import org.apache.ratis.server.impl.RaftServerTestUtil;
-import org.apache.ratis.server.storage.RaftLog;
-import org.apache.ratis.server.storage.RaftLogIOException;
+import org.apache.ratis.server.raftlog.RaftLog;
+import org.apache.ratis.server.raftlog.RaftLogIOException;
 import org.apache.ratis.util.JavaUtils;
-import org.apache.ratis.util.LogUtils;
+import org.apache.ratis.util.Log4jUtils;
 import org.apache.ratis.util.TimeDuration;
 import org.junit.Assert;
 import org.junit.Test;
@@ -46,7 +46,7 @@ public abstract class RetryCacheTests<CLUSTER extends MiniRaftCluster>
     extends BaseTest
     implements MiniRaftCluster.Factory.Get<CLUSTER> {
   static {
-    LogUtils.setLogLevel(RaftServerImpl.LOG, Level.DEBUG);
+    Log4jUtils.setLogLevel(RaftServerImpl.LOG, Level.DEBUG);
   }
 
   public static final int NUM_SERVERS = 3;
@@ -141,7 +141,7 @@ public abstract class RetryCacheTests<CLUSTER extends MiniRaftCluster>
     // trigger setConfiguration
     cluster.setConfiguration(allPeers);
 
-    final RaftPeerId newLeaderId = JavaUtils.attempt(() -> {
+    final RaftPeerId newLeaderId = JavaUtils.attemptRepeatedly(() -> {
       final RaftPeerId id = RaftTestUtil.waitForLeader(cluster).getId();
       Assert.assertNotEquals(leaderId, id);
       return id;
