@@ -19,20 +19,25 @@ package org.apache.ratis;
 
 import org.apache.log4j.Level;
 import org.apache.ratis.client.RaftClient;
+import org.apache.ratis.client.RaftClientConfigKeys;
 import org.apache.ratis.conf.RaftProperties;
 import org.apache.ratis.proto.RaftProtos.CommitInfoProto;
 import org.apache.ratis.proto.RaftProtos.ReplicationLevel;
 import org.apache.ratis.protocol.NotReplicatedException;
 import org.apache.ratis.protocol.RaftClientReply;
+import org.apache.ratis.protocol.RaftRetryFailureException;
+import org.apache.ratis.protocol.TimeoutIOException;
+import org.apache.ratis.retry.RetryPolicies;
 import org.apache.ratis.server.RaftServerConfigKeys;
 import org.apache.ratis.server.impl.RaftServerImpl;
 import org.apache.ratis.server.impl.RaftServerTestUtil;
 import org.apache.ratis.statemachine.SimpleStateMachine4Testing;
 import org.apache.ratis.statemachine.StateMachine;
-import org.apache.ratis.util.LogUtils;
+import org.apache.ratis.util.Log4jUtils;
 import org.apache.ratis.util.ProtoUtils;
 import org.apache.ratis.util.TimeDuration;
 import org.apache.ratis.util.function.CheckedConsumer;
+import org.apache.ratis.util.function.CheckedSupplier;
 import org.junit.Assert;
 import org.junit.Before;
 import org.junit.Test;
@@ -42,20 +47,23 @@ import java.util.ArrayList;
 import java.util.Collection;
 import java.util.List;
 import java.util.concurrent.CompletableFuture;
+import java.util.concurrent.ExecutionException;
 import java.util.concurrent.ThreadLocalRandom;
 import java.util.concurrent.TimeUnit;
 import java.util.stream.Stream;
 
+import static org.junit.Assert.fail;
+
 public abstract class WatchRequestTests<CLUSTER extends MiniRaftCluster>
     extends BaseTest
     implements MiniRaftCluster.Factory.Get<CLUSTER> {
   {
     RaftServerTestUtil.setWatchRequestsLogLevel(Level.DEBUG);
-    LogUtils.setLogLevel(RaftServerImpl.LOG, Level.DEBUG);
+    Log4jUtils.setLogLevel(RaftServerImpl.LOG, Level.DEBUG);
   }
 
   static final int NUM_SERVERS = 3;
-  static final int GET_TIMEOUT_SECOND = 5;
+  static final int GET_TIMEOUT_SECOND = 10;
 
   @Before
   public void setup() {
@@ -105,6 +113,24 @@ public abstract class WatchRequestTests<CLUSTER extends MiniRaftCluster>
       }
     }
 
+    CompletableFuture<RaftClientReply> sendWatchRequest(long logIndex)
+        throws Exception {
+
+      try (final RaftClient watchClient =
+          cluster.createClient(RaftTestUtil.waitForLeader(cluster).getId(),
+          RetryPolicies.retryUpToMaximumCountWithFixedSleep(5,
+          TimeDuration.valueOf(5, TimeUnit.SECONDS)))) {
+
+        CompletableFuture<RaftClientReply> reply =
+            watchClient.sendAsync(new RaftTestUtil.SimpleMessage("message"));
+        long writeIndex = reply.get().getLogIndex();
+        Assert.assertTrue(writeIndex > 0);
+        watchClient.sendWatchAsync(writeIndex, ReplicationLevel.MAJORITY_COMMITTED);
+        return watchClient.sendWatchAsync(logIndex, ReplicationLevel.MAJORITY);
+      }
+
+    }
+
     @Override
     public String toString() {
       return "numMessages=" + numMessages;
@@ -114,11 +140,23 @@ public abstract class WatchRequestTests<CLUSTER extends MiniRaftCluster>
   static void runTest(CheckedConsumer<TestParameters, Exception> testCase, MiniRaftCluster cluster, Logger LOG)
       throws Exception {
     try(final RaftClient client = cluster.createClient(RaftTestUtil.waitForLeader(cluster).getId())) {
-      final int[] numMessages = {1, 10, 100};
-      for(int i = 0; i < 5; i++) {
-        final int n = numMessages[ThreadLocalRandom.current().nextInt(numMessages.length)];
+      final int[] numMessages = {1, 10, 20};
+      for(int n : numMessages) {
         final TestParameters p = new TestParameters(n, client, cluster, LOG);
-        LOG.info("{}) {}, {}", i, p, cluster.printServers());
+        LOG.info("{}) {}, {}", n, p, cluster.printServers());
+        testCase.accept(p);
+      }
+    }
+  }
+
+  static void runSingleTest(CheckedConsumer<TestParameters, Exception> testCase,
+      MiniRaftCluster cluster, Logger LOG)
+      throws Exception {
+    try(final RaftClient client = cluster.createClient(RaftTestUtil.waitForLeader(cluster).getId())) {
+      final int[] numMessages = {1};
+      for(int n : numMessages) {
+        final TestParameters p = new TestParameters(n, client, cluster, LOG);
+        LOG.info("{}) {}, {}", n, p, cluster.printServers());
         testCase.accept(p);
       }
     }
@@ -280,9 +318,9 @@ public abstract class WatchRequestTests<CLUSTER extends MiniRaftCluster>
     futures.forEach(f -> {
       if (f.isDone()) {
         try {
-          Assert.fail("Done unexpectedly: " + f.get());
+          fail("Done unexpectedly: " + f.get());
         } catch(Exception e) {
-          Assert.fail("Done unexpectedly and failed to get: " + e);
+          fail("Done unexpectedly and failed to get: " + e);
         }
       }
     });
@@ -331,14 +369,14 @@ public abstract class WatchRequestTests<CLUSTER extends MiniRaftCluster>
   @Test
   public void testWatchRequestTimeout() throws Exception {
     final RaftProperties p = getProperties();
-    RaftServerConfigKeys.setWatchTimeout(p, TimeDuration.valueOf(500, TimeUnit.MILLISECONDS));
-    RaftServerConfigKeys.setWatchTimeoutDenomination(p, TimeDuration.valueOf(100, TimeUnit.MILLISECONDS));
+    RaftServerConfigKeys.Watch.setTimeout(p, TimeDuration.valueOf(500, TimeUnit.MILLISECONDS));
+    RaftServerConfigKeys.Watch.setTimeoutDenomination(p, TimeDuration.valueOf(100, TimeUnit.MILLISECONDS));
     try {
       runWithNewCluster(NUM_SERVERS,
           cluster -> runTest(WatchRequestTests::runTestWatchRequestTimeout, cluster, LOG));
     } finally {
-      RaftServerConfigKeys.setWatchTimeout(p, RaftServerConfigKeys.WATCH_TIMEOUT_DEFAULT);
-      RaftServerConfigKeys.setWatchTimeoutDenomination(p, RaftServerConfigKeys.WATCH_TIMEOUT_DENOMINATION_DEFAULT);
+      RaftServerConfigKeys.Watch.setTimeout(p, RaftServerConfigKeys.Watch.TIMEOUT_DEFAULT);
+      RaftServerConfigKeys.Watch.setTimeoutDenomination(p, RaftServerConfigKeys.Watch.TIMEOUT_DENOMINATION_DEFAULT);
     }
   }
 
@@ -347,8 +385,9 @@ public abstract class WatchRequestTests<CLUSTER extends MiniRaftCluster>
     final MiniRaftCluster cluster = p.cluster;
     final int numMessages = p.numMessages;
 
-    final TimeDuration watchTimeout = RaftServerConfigKeys.watchTimeout(cluster.getProperties());
-    final TimeDuration watchTimeoutDenomination = RaftServerConfigKeys.watchTimeoutDenomination(cluster.getProperties());
+    final RaftProperties properties = cluster.getProperties();
+    final TimeDuration watchTimeout = RaftServerConfigKeys.Watch.timeout(properties);
+    final TimeDuration watchTimeoutDenomination = RaftServerConfigKeys.Watch.timeoutDenomination(properties);
 
     // blockStartTransaction of the leader so that no transaction can be committed MAJORITY
     final RaftServerImpl leader = cluster.getLeader();
@@ -386,6 +425,45 @@ public abstract class WatchRequestTests<CLUSTER extends MiniRaftCluster>
     LOG.info("unblock follower {}", blockedFollower.getId());
   }
 
+  @Test
+  public void testWatchRequestClientTimeout() throws Exception {
+    final RaftProperties p = getProperties();
+    RaftServerConfigKeys.Watch.setTimeout(p, TimeDuration.valueOf(100,
+        TimeUnit.SECONDS));
+    RaftClientConfigKeys.Rpc.setWatchRequestTimeout(p,
+        TimeDuration.valueOf(15, TimeUnit.SECONDS));
+    try {
+      runWithNewCluster(NUM_SERVERS,
+          cluster -> runSingleTest(WatchRequestTests::runTestWatchRequestClientTimeout, cluster, LOG));
+    } finally {
+      RaftServerConfigKeys.Watch.setTimeout(p, RaftServerConfigKeys.Watch.TIMEOUT_DEFAULT);
+      RaftClientConfigKeys.Rpc.setWatchRequestTimeout(p,
+          RaftClientConfigKeys.Rpc.WATCH_REQUEST_TIMEOUT_DEFAULT);
+    }
+  }
+
+  static void runTestWatchRequestClientTimeout(TestParameters p) throws Exception {
+    final Logger LOG = p.log;
+
+    CompletableFuture<RaftClientReply> watchReply;
+    watchReply = p.sendWatchRequest(1000);
+
+    try {
+      watchReply.get();
+      fail("runTestWatchRequestClientTimeout failed");
+    } catch (Exception ex) {
+      LOG.error("error occurred", ex);
+      Assert.assertEquals(RaftRetryFailureException.class,
+          ex.getCause().getClass());
+      if (ex.getCause() != null) {
+        if (ex.getCause().getCause() != null) {
+          Assert.assertEquals(TimeoutIOException.class,
+              ex.getCause().getCause().getClass());
+        }
+      }
+    }
+  }
+
   static void checkTimeout(List<CompletableFuture<RaftClientReply>> replies,
       List<CompletableFuture<WatchReplies>> watches, Logger LOG) throws Exception {
     for(int i = 0; i < replies.size(); i++) {
@@ -397,17 +475,26 @@ public abstract class WatchRequestTests<CLUSTER extends MiniRaftCluster>
       final WatchReplies watchReplies = watches.get(i).get(GET_TIMEOUT_SECOND, TimeUnit.SECONDS);
       Assert.assertEquals(logIndex, watchReplies.logIndex);
 
-      final RaftClientReply watchAllReply = watchReplies.getAll();
-      assertNotReplicatedException(logIndex, ReplicationLevel.ALL, watchAllReply);
+      assertNotReplicatedException(logIndex, ReplicationLevel.ALL, watchReplies::getAll);
 
-      final RaftClientReply watchAllCommittedReply = watchReplies.getAllCommitted();
-      assertNotReplicatedException(logIndex, ReplicationLevel.ALL_COMMITTED, watchAllCommittedReply);
+      assertNotReplicatedException(logIndex, ReplicationLevel.ALL_COMMITTED, watchReplies::getAllCommitted);
     }
   }
 
-  static void assertNotReplicatedException(long logIndex, ReplicationLevel replication, RaftClientReply reply) {
-    Assert.assertFalse(reply.isSuccess());
-    final NotReplicatedException nre = reply.getNotReplicatedException();
+  static void assertNotReplicatedException(long logIndex, ReplicationLevel replication,
+      CheckedSupplier<RaftClientReply, Exception> replySupplier) throws Exception {
+    try {
+      replySupplier.get();
+      fail();
+    } catch (ExecutionException e) {
+      final Throwable cause = e.getCause();
+      assertNotReplicatedException(logIndex, replication, cause);
+    }
+  }
+
+  static void assertNotReplicatedException(long logIndex, ReplicationLevel replication, Throwable t) {
+    Assert.assertSame(NotReplicatedException.class, t.getClass());
+    final NotReplicatedException nre = (NotReplicatedException) t;
     Assert.assertNotNull(nre);
     Assert.assertEquals(logIndex, nre.getLogIndex());
     Assert.assertEquals(replication, nre.getRequiredReplication());
