@@ -22,29 +22,44 @@ import org.apache.ratis.BaseTest;
 import org.apache.ratis.MiniRaftCluster;
 import org.apache.ratis.RaftTestUtil;
 import org.apache.ratis.client.RaftClient;
+import org.apache.ratis.metrics.RatisMetricRegistry;
+import org.apache.ratis.protocol.RaftGroupId;
+import org.apache.ratis.protocol.RaftGroupMemberId;
 import org.apache.ratis.protocol.RaftPeerId;
-import org.apache.ratis.server.storage.RaftStorageTestUtils;
+import org.apache.ratis.server.metrics.LeaderElectionMetrics;
+import org.apache.ratis.server.raftlog.segmented.SegmentedRaftLogTestUtils;
 import org.apache.ratis.util.ExitUtils;
 import org.apache.ratis.util.JavaUtils;
-import org.apache.ratis.util.LogUtils;
+import org.apache.ratis.util.LifeCycle;
+import org.apache.ratis.util.Log4jUtils;
 import org.apache.ratis.util.TimeDuration;
+import org.apache.ratis.util.Timestamp;
 import org.junit.Assert;
 import org.junit.Test;
 import org.slf4j.Logger;
 
+import java.io.IOException;
 import java.util.Iterator;
 import java.util.Optional;
 import java.util.concurrent.ThreadLocalRandom;
 import java.util.concurrent.TimeUnit;
 
 import static org.apache.ratis.RaftTestUtil.waitForLeader;
+import static org.apache.ratis.server.metrics.LeaderElectionMetrics.LEADER_ELECTION_COUNT_METRIC;
+import static org.apache.ratis.server.metrics.LeaderElectionMetrics.LEADER_ELECTION_LATENCY;
+import static org.apache.ratis.server.metrics.LeaderElectionMetrics.LEADER_ELECTION_TIMEOUT_COUNT_METRIC;
+import static org.junit.Assert.assertEquals;
+import static org.junit.Assert.assertTrue;
+import static org.junit.Assert.fail;
+import static org.mockito.Mockito.mock;
+import static org.mockito.Mockito.when;
 
 public abstract class LeaderElectionTests<CLUSTER extends MiniRaftCluster>
     extends BaseTest
     implements MiniRaftCluster.Factory.Get<CLUSTER> {
   {
-    LogUtils.setLogLevel(RaftServerImpl.LOG, Level.DEBUG);
-    LogUtils.setLogLevel(RaftClient.LOG, Level.DEBUG);
+    Log4jUtils.setLogLevel(RaftServerImpl.LOG, Level.DEBUG);
+    Log4jUtils.setLogLevel(RaftClient.LOG, Level.DEBUG);
   }
 
   @Test
@@ -63,7 +78,7 @@ public abstract class LeaderElectionTests<CLUSTER extends MiniRaftCluster>
 
   @Test
   public void testChangeLeader() throws Exception {
-    RaftStorageTestUtils.setRaftLogWorkerLogLevel(Level.TRACE);
+    SegmentedRaftLogTestUtils.setRaftLogWorkerLogLevel(Level.TRACE);
     LOG.info("Running testChangeLeader");
     final MiniRaftCluster cluster = newCluster(3);
     cluster.start();
@@ -73,7 +88,7 @@ public abstract class LeaderElectionTests<CLUSTER extends MiniRaftCluster>
       leader = RaftTestUtil.changeLeader(cluster, leader, IllegalStateException::new);
       ExitUtils.assertNotTerminated();
     }
-    RaftStorageTestUtils.setRaftLogWorkerLogLevel(Level.INFO);
+    SegmentedRaftLogTestUtils.setRaftLogWorkerLogLevel(Level.INFO);
     cluster.shutdown();
   }
 
@@ -129,11 +144,72 @@ public abstract class LeaderElectionTests<CLUSTER extends MiniRaftCluster>
     // start the last server
     final RaftServerProxy lastServer = i.next();
     lastServer.start();
-    final RaftPeerId lastServerLeaderId = JavaUtils.attempt(
+    final RaftPeerId lastServerLeaderId = JavaUtils.attemptRepeatedly(
         () -> Optional.ofNullable(lastServer.getImpls().iterator().next().getState().getLeaderId())
             .orElseThrow(() -> new IllegalStateException("No leader yet")),
         10, ONE_SECOND, "getLeaderId", LOG);
     LOG.info(cluster.printServers());
     Assert.assertEquals(leader.getId(), lastServerLeaderId);
   }
+
+  @Test
+  public void testLeaderElectionMetrics() throws IOException, InterruptedException {
+    LOG.info("Running testLeaderElectionMetrics");
+    Timestamp timestamp = Timestamp.currentTime();
+    final MiniRaftCluster cluster = newCluster(3);
+    cluster.start();
+    RaftServerImpl leaderServer = waitForLeader(cluster);
+
+    RatisMetricRegistry ratisMetricRegistry = LeaderElectionMetrics.getLeaderElectionMetrics(leaderServer).getRegistry();
+
+    // Verify each metric individually.
+    long numLeaderElections = ratisMetricRegistry.counter(LEADER_ELECTION_COUNT_METRIC).getCount();
+    assertTrue(numLeaderElections > 0);
+
+    long numLeaderElectionTimeout = ratisMetricRegistry.counter(LEADER_ELECTION_TIMEOUT_COUNT_METRIC).getCount();
+    assertTrue(numLeaderElectionTimeout > 0);
+
+    Long leaderElectionLatency = (Long) ratisMetricRegistry.getGauges((s, metric) ->
+        s.contains(LEADER_ELECTION_LATENCY)).values().iterator().next().getValue();
+    assertTrue(leaderElectionLatency > 0 && leaderElectionLatency < timestamp.elapsedTimeMs());
+  }
+
+  @Test
+  public void testImmediatelyRevertedToFollower() {
+    RaftServerImpl server = createMockServer(true);
+    LeaderElection subject = new LeaderElection(server);
+
+    try {
+      subject.startInForeground();
+      assertEquals(LifeCycle.State.CLOSED, subject.getCurrentState());
+    } catch (Exception e) {
+      LOG.info("Error starting LeaderElection", e);
+      fail(e.getMessage());
+    }
+  }
+
+  @Test
+  public void testShutdownBeforeStart() {
+    RaftServerImpl server = createMockServer(false);
+    LeaderElection subject = new LeaderElection(server);
+
+    try {
+      subject.shutdown();
+      subject.startInForeground();
+      assertEquals(LifeCycle.State.CLOSED, subject.getCurrentState());
+    } catch (Exception e) {
+      LOG.info("Error starting LeaderElection", e);
+      fail(e.getMessage());
+    }
+  }
+
+  private static RaftServerImpl createMockServer(boolean alive) {
+    RaftServerImpl server = mock(RaftServerImpl.class);
+    when(server.isAlive()).thenReturn(alive);
+    when(server.isCandidate()).thenReturn(false);
+    when(server.getMemberId()).thenReturn(RaftGroupMemberId.valueOf(RaftPeerId.valueOf("any"), RaftGroupId.randomId()));
+    LeaderElectionMetrics leaderElectionMetrics = LeaderElectionMetrics.getLeaderElectionMetrics(server);
+    when(server.getLeaderElectionMetrics()).thenReturn(leaderElectionMetrics);
+    return server;
+  }
 }
