@@ -28,10 +28,10 @@ import org.apache.ratis.client.RaftClientRpc;
 import org.apache.ratis.proto.RaftProtos.LogEntryProto;
 import org.apache.ratis.protocol.*;
 import org.apache.ratis.server.RaftServerConfigKeys;
-import org.apache.ratis.server.storage.RaftLog;
+import org.apache.ratis.server.raftlog.RaftLog;
 import org.apache.ratis.server.storage.RaftStorageTestUtils;
 import org.apache.ratis.util.JavaUtils;
-import org.apache.ratis.util.LogUtils;
+import org.apache.ratis.util.Log4jUtils;
 import org.apache.ratis.util.TimeDuration;
 import org.junit.Assert;
 import org.junit.Test;
@@ -55,7 +55,7 @@ public abstract class RaftReconfigurationBaseTest<CLUSTER extends MiniRaftCluste
     extends BaseTest
     implements MiniRaftCluster.Factory.Get<CLUSTER> {
   static {
-    LogUtils.setLogLevel(RaftServerImpl.LOG, Level.DEBUG);
+    Log4jUtils.setLogLevel(RaftServerImpl.LOG, Level.DEBUG);
   }
 
   private static final DelayLocalExecutionInjection logSyncDelay =
@@ -521,7 +521,7 @@ public abstract class RaftReconfigurationBaseTest<CLUSTER extends MiniRaftCluste
 
       // find ConfigurationEntry
       final TimeDuration sleepTime = TimeDuration.valueOf(500, TimeUnit.MILLISECONDS);
-      final long confIndex = JavaUtils.attempt(() -> {
+      final long confIndex = JavaUtils.attemptRepeatedly(() -> {
         final long last = log.getLastEntryTermIndex().getIndex();
         for (long i = last; i >= 1; i--) {
           if (log.get(i).hasConfigurationEntry()) {
@@ -532,7 +532,7 @@ public abstract class RaftReconfigurationBaseTest<CLUSTER extends MiniRaftCluste
       }, 10, sleepTime, "confIndex", LOG);
 
       // wait till the old leader persist the new conf
-      JavaUtils.attempt(() -> log.getLatestFlushedIndex() >= confIndex,
+      JavaUtils.attemptRepeatedly(() -> log.getFlushIndex() >= confIndex,
           10, sleepTime, "FLUSH", LOG);
       final long committed = log.getLastCommittedIndex();
       Assert.assertTrue(committed < confIndex);
@@ -546,7 +546,7 @@ public abstract class RaftReconfigurationBaseTest<CLUSTER extends MiniRaftCluste
       Assert.assertTrue(gotNotLeader.get());
 
       // the old leader should have truncated the setConf from the log
-      JavaUtils.attempt(() -> log.getLastCommittedIndex() >= confIndex,
+      JavaUtils.attemptRepeatedly(() -> log.getLastCommittedIndex() >= confIndex,
           10, ONE_SECOND, "COMMIT", LOG);
       Assert.assertTrue(log.get(confIndex).hasConfigurationEntry());
       log2 = null;
@@ -575,16 +575,15 @@ public abstract class RaftReconfigurationBaseTest<CLUSTER extends MiniRaftCluste
       new Thread(() -> {
         final RaftClient client = cluster.createClient(leaderId);
         final RaftClientRpc sender = client.getClientRpc();
-
         final RaftClientRequest request = cluster.newRaftClientRequest(
             client.getId(), leaderId, new SimpleMessage("test"));
         while (!success.get()) {
           try {
-            RaftClientReply reply = sender.sendRequest(request);
+            final RaftClientReply reply = sender.sendRequest(request);
             success.set(reply.isSuccess());
-          } catch (LeaderNotReadyException e) {
-            LOG.info("Hit LeaderNotReadyException", e);
-            caughtNotReady.set(true);
+            if (reply.getException() != null && reply.getException() instanceof LeaderNotReadyException) {
+              caughtNotReady.set(true);
+            }
           } catch (IOException e) {
             LOG.info("Hit other IOException", e);
           }
