@@ -25,7 +25,7 @@ import org.apache.ratis.protocol.RaftGroupId;
 import org.apache.ratis.protocol.RaftPeer;
 import org.apache.ratis.protocol.RaftPeerId;
 import org.apache.ratis.util.JavaUtils;
-import org.apache.ratis.util.LogUtils;
+import org.apache.ratis.util.Log4jUtils;
 import org.apache.ratis.util.TimeDuration;
 import org.junit.Assert;
 import org.slf4j.Logger;
@@ -40,19 +40,22 @@ public class RaftServerTestUtil {
   static final Logger LOG = LoggerFactory.getLogger(RaftServerTestUtil.class);
 
   public static void setWatchRequestsLogLevel(Level level) {
-    LogUtils.setLogLevel(WatchRequests.LOG, level);
+    Log4jUtils.setLogLevel(WatchRequests.LOG, level);
+  }
+  public static void setPendingRequestsLogLevel(Level level) {
+    Log4jUtils.setLogLevel(PendingRequests.LOG, level);
   }
 
   public static void waitAndCheckNewConf(MiniRaftCluster cluster,
       RaftPeer[] peers, int numOfRemovedPeers, Collection<RaftPeerId> deadPeers)
       throws Exception {
     final TimeDuration sleepTime = cluster.getTimeoutMax().apply(n -> n * (numOfRemovedPeers + 2));
-    JavaUtils.attempt(() -> waitAndCheckNewConf(cluster, peers, deadPeers),
+    JavaUtils.attempt(() -> waitAndCheckNewConf(cluster, Arrays.asList(peers), deadPeers),
         10, sleepTime, "waitAndCheckNewConf", LOG);
   }
   private static void waitAndCheckNewConf(MiniRaftCluster cluster,
-      RaftPeer[] peers, Collection<RaftPeerId> deadPeers) {
-    LOG.info("waitAndCheckNewConf: peers={}, deadPeers={}, {}", Arrays.asList(peers), deadPeers, cluster.printServers());
+      Collection<RaftPeer> peers, Collection<RaftPeerId> deadPeers) {
+    LOG.info("waitAndCheckNewConf: peers={}, deadPeers={}, {}", peers, deadPeers, cluster.printServers());
     Assert.assertNotNull(cluster.getLeader());
 
     int numIncluded = 0;
@@ -78,7 +81,7 @@ public class RaftServerTestUtil {
         Assert.assertFalse(server.getRaftConf().containsInConf(server.getId()));
       }
     }
-    Assert.assertEquals(peers.length, numIncluded + deadIncluded);
+    Assert.assertEquals(peers.size(), numIncluded + deadIncluded);
   }
 
   public static long getRetryCacheSize(RaftServerImpl server) {
