@@ -1,4 +1,4 @@
-/**
+/*
  * Licensed to the Apache Software Foundation (ASF) under one
  * or more contributor license agreements.  See the NOTICE file
  * distributed with this work for additional information
@@ -17,23 +17,28 @@
  */
 package org.apache.ratis.server.storage;
 
-import org.apache.log4j.Level;
+import static org.apache.ratis.server.metrics.RaftLogMetrics.RAFT_LOG_FLUSH_TIME;
+import static org.apache.ratis.server.metrics.RaftLogMetrics.RATIS_LOG_WORKER_METRICS;
+
 import org.apache.ratis.protocol.RaftPeerId;
 import org.apache.ratis.server.impl.ServerProtoUtils;
+import org.apache.ratis.server.metrics.RatisMetrics;
 import org.apache.ratis.server.protocol.TermIndex;
-import org.apache.ratis.proto.RaftProtos;
+import org.apache.ratis.server.raftlog.RaftLog;
+import org.apache.ratis.server.raftlog.RaftLogIOException;
 import org.apache.ratis.util.AutoCloseableLock;
-import org.apache.ratis.util.LogUtils;
 
 import java.util.function.Consumer;
 
 public interface RaftStorageTestUtils {
-  static void setRaftLogWorkerLogLevel(Level level) {
-    LogUtils.setLogLevel(RaftLogWorker.LOG, level);
-  }
 
   static String getLogFlushTimeMetric(RaftPeerId serverId) {
-    return RaftLogWorker.class.getName() + "." + serverId + ".flush-time";
+    return getRaftLogFullMetric(serverId, RAFT_LOG_FLUSH_TIME);
+  }
+
+  static String getRaftLogFullMetric(RaftPeerId serverId, String metricName) {
+    return RatisMetrics.RATIS_APPLICATION_NAME_METRICS + "." + RATIS_LOG_WORKER_METRICS
+        + "." + serverId + "." + metricName;
   }
 
   static void printLog(RaftLog log, Consumer<String> println) {
@@ -46,7 +51,7 @@ public interface RaftStorageTestUtils {
     final long flushed, committed;
     try(AutoCloseableLock readlock = log.readLock()) {
       last = log.getLastEntryTermIndex();
-      flushed = log.getLatestFlushedIndex();
+      flushed = log.getFlushIndex();
       committed = log.getLastCommittedIndex();
     }
     final StringBuilder b = new StringBuilder();
