@@ -17,6 +17,12 @@
  */
 package org.apache.ratis.statemachine;
 
+import static org.apache.ratis.server.impl.StateMachineMetrics.RATIS_STATEMACHINE_METRICS;
+import static org.apache.ratis.server.impl.StateMachineMetrics.RATIS_STATEMACHINE_METRICS_DESC;
+import static org.apache.ratis.server.impl.StateMachineMetrics.STATEMACHINE_TAKE_SNAPSHOT_TIMER;
+import static org.apache.ratis.server.metrics.RaftLogMetrics.LOG_APPENDER_INSTALL_SNAPSHOT_METRIC;
+import static org.apache.ratis.server.metrics.RatisMetrics.RATIS_APPLICATION_NAME_METRICS;
+
 import org.apache.log4j.Level;
 import org.apache.ratis.BaseTest;
 import org.apache.ratis.MiniRaftCluster;
@@ -24,19 +30,22 @@ import org.apache.ratis.RaftTestUtil;
 import org.apache.ratis.RaftTestUtil.SimpleMessage;
 import org.apache.ratis.client.RaftClient;
 import org.apache.ratis.conf.RaftProperties;
+import org.apache.ratis.metrics.MetricRegistries;
+import org.apache.ratis.metrics.MetricRegistryInfo;
+import org.apache.ratis.metrics.RatisMetricRegistry;
 import org.apache.ratis.protocol.RaftClientReply;
 import org.apache.ratis.protocol.RaftPeerId;
 import org.apache.ratis.server.RaftServerConfigKeys;
 import org.apache.ratis.server.impl.RaftServerImpl;
 import org.apache.ratis.server.impl.RaftServerTestUtil;
-import org.apache.ratis.server.storage.RaftLog;
+import org.apache.ratis.server.raftlog.RaftLog;
 import org.apache.ratis.server.storage.RaftStorageDirectory;
 import org.apache.ratis.server.storage.RaftStorageDirectory.LogPathAndIndex;
 import org.apache.ratis.proto.RaftProtos.LogEntryProto;
 import org.apache.ratis.statemachine.impl.SimpleStateMachineStorage;
 import org.apache.ratis.util.FileUtils;
 import org.apache.ratis.util.JavaUtils;
-import org.apache.ratis.util.LogUtils;
+import org.apache.ratis.util.Log4jUtils;
 import org.junit.After;
 import org.junit.Assert;
 import org.junit.Before;
@@ -47,20 +56,24 @@ import org.slf4j.LoggerFactory;
 import java.io.File;
 import java.io.IOException;
 import java.util.List;
+import java.util.Optional;
 import java.util.stream.Collectors;
 import java.util.stream.LongStream;
 
+import com.codahale.metrics.Counter;
+import com.codahale.metrics.Timer;
+
 public abstract class RaftSnapshotBaseTest extends BaseTest {
-  static {
-    LogUtils.setLogLevel(RaftServerImpl.LOG, Level.DEBUG);
-    LogUtils.setLogLevel(RaftLog.LOG, Level.DEBUG);
-    LogUtils.setLogLevel(RaftClient.LOG, Level.DEBUG);
+  {
+    Log4jUtils.setLogLevel(RaftServerImpl.LOG, Level.DEBUG);
+    Log4jUtils.setLogLevel(RaftLog.LOG, Level.DEBUG);
+    Log4jUtils.setLogLevel(RaftClient.LOG, Level.DEBUG);
   }
 
   static final Logger LOG = LoggerFactory.getLogger(RaftSnapshotBaseTest.class);
   private static final int SNAPSHOT_TRIGGER_THRESHOLD = 10;
 
-  static List<File> getSnapshotFiles(MiniRaftCluster cluster, long startIndex, long endIndex) {
+  public static List<File> getSnapshotFiles(MiniRaftCluster cluster, long startIndex, long endIndex) {
     final RaftServerImpl leader = cluster.getLeader();
     final SimpleStateMachineStorage storage = SimpleStateMachine4Testing.get(leader).getStateMachineStorage();
     final long term = leader.getState().getCurrentTerm();
@@ -70,7 +83,7 @@ public abstract class RaftSnapshotBaseTest extends BaseTest {
   }
 
 
-  static void assertLeaderContent(MiniRaftCluster cluster) throws Exception {
+  public static void assertLeaderContent(MiniRaftCluster cluster) throws Exception {
     final RaftServerImpl leader = RaftTestUtil.waitForLeader(cluster);
     final RaftLog leaderLog = leader.getState().getLog();
     final long lastIndex = leaderLog.getLastEntryTermIndex().getIndex();
@@ -78,7 +91,9 @@ public abstract class RaftSnapshotBaseTest extends BaseTest {
     Assert.assertTrue(e.hasMetadataEntry());
     Assert.assertEquals(leaderLog.getLastCommittedIndex() - 1, e.getMetadataEntry().getCommitIndex());
 
-    final LogEntryProto[] entries = SimpleStateMachine4Testing.get(leader).getContent();
+    SimpleStateMachine4Testing simpleStateMachine = SimpleStateMachine4Testing.get(leader);
+    Assert.assertTrue("Is not notified as a leader", simpleStateMachine.isNotifiedAsLeader());
+    final LogEntryProto[] entries = simpleStateMachine.getContent();
     long message = 0;
     for (int i = 0; i < entries.length; i++) {
       LOG.info("{}) {} {}", i, message, entries[i]);
@@ -134,7 +149,7 @@ public abstract class RaftSnapshotBaseTest extends BaseTest {
     LOG.info("nextIndex = {}", nextIndex);
     // wait for the snapshot to be done
     final List<File> snapshotFiles = getSnapshotFiles(cluster, nextIndex - SNAPSHOT_TRIGGER_THRESHOLD, nextIndex);
-    JavaUtils.attempt(() -> snapshotFiles.stream().anyMatch(RaftSnapshotBaseTest::exists),
+    JavaUtils.attemptRepeatedly(() -> snapshotFiles.stream().anyMatch(RaftSnapshotBaseTest::exists),
         10, ONE_SECOND, "snapshotFile.exist", LOG);
 
     // restart the peer and check if it can correctly load snapshot
@@ -147,7 +162,7 @@ public abstract class RaftSnapshotBaseTest extends BaseTest {
     }
   }
 
-  static boolean exists(File f) {
+  public static boolean exists(File f) {
     if (f.exists()) {
       LOG.info("File exists: " + f);
       return true;
@@ -182,7 +197,7 @@ public abstract class RaftSnapshotBaseTest extends BaseTest {
       final long nextIndex = cluster.getLeader().getState().getLog().getNextIndex();
       LOG.info("nextIndex = {}", nextIndex);
       final List<File> snapshotFiles = getSnapshotFiles(cluster, nextIndex - SNAPSHOT_TRIGGER_THRESHOLD, nextIndex);
-      JavaUtils.attempt(() -> snapshotFiles.stream().anyMatch(RaftSnapshotBaseTest::exists),
+      JavaUtils.attemptRepeatedly(() -> snapshotFiles.stream().anyMatch(RaftSnapshotBaseTest::exists),
           10, ONE_SECOND, "snapshotFile.exist", LOG);
       logs = storageDirectory.getLogSegmentFiles();
     } finally {
@@ -211,13 +226,34 @@ public abstract class RaftSnapshotBaseTest extends BaseTest {
       // trigger setConfiguration
       cluster.setConfiguration(change.allPeersInNewConf);
 
+      // Verify installSnapshot counter on leader before restart.
+      verifyInstallSnapshotMetric(cluster.getLeader());
       RaftServerTestUtil.waitAndCheckNewConf(cluster, change.allPeersInNewConf, 0, null);
 
       // restart the peer and check if it can correctly handle conf change
       cluster.restartServer(cluster.getLeader().getId(), false);
       assertLeaderContent(cluster);
+      verifyTakeSnapshotMetric(cluster.getLeader());
     } finally {
       cluster.shutdown();
     }
   }
+
+  protected void verifyInstallSnapshotMetric(RaftServerImpl leader) {
+    Counter installSnapshotCounter = leader.getRaftServerMetrics().getCounter(LOG_APPENDER_INSTALL_SNAPSHOT_METRIC);
+    Assert.assertNotNull(installSnapshotCounter);
+    Assert.assertTrue(installSnapshotCounter.getCount() >= 1);
+  }
+
+  private static void verifyTakeSnapshotMetric(RaftServerImpl leader) {
+    MetricRegistryInfo info = new MetricRegistryInfo(leader.getMemberId().toString(),
+        RATIS_APPLICATION_NAME_METRICS,
+        RATIS_STATEMACHINE_METRICS, RATIS_STATEMACHINE_METRICS_DESC);
+    Optional<RatisMetricRegistry> opt = MetricRegistries.global().get(info);
+    Assert.assertTrue(opt.isPresent());
+    RatisMetricRegistry metricRegistry = opt.get();
+    Assert.assertNotNull(metricRegistry);
+    Timer timer = metricRegistry.timer(STATEMACHINE_TAKE_SNAPSHOT_TIMER);
+    Assert.assertTrue(timer.getCount() > 0);
+  }
 }
