@@ -24,8 +24,9 @@ import org.apache.ratis.proto.RaftProtos.LogEntryProto;
 import org.apache.ratis.proto.RaftProtos.RoleInfoProto;
 import org.apache.ratis.protocol.Message;
 import org.apache.ratis.protocol.RaftClientRequest;
-import org.apache.ratis.protocol.RaftGroup;
 import org.apache.ratis.protocol.RaftGroupId;
+import org.apache.ratis.protocol.RaftGroupMemberId;
+import org.apache.ratis.protocol.RaftPeerId;
 import org.apache.ratis.protocol.StateMachineException;
 import org.apache.ratis.server.RaftServer;
 import org.apache.ratis.server.RaftServerConfigKeys;
@@ -33,8 +34,8 @@ import org.apache.ratis.server.impl.RaftServerConstants;
 import org.apache.ratis.server.impl.RaftServerImpl;
 import org.apache.ratis.server.impl.ServerProtoUtils;
 import org.apache.ratis.server.protocol.TermIndex;
-import org.apache.ratis.server.storage.LogInputStream;
-import org.apache.ratis.server.storage.LogOutputStream;
+import org.apache.ratis.server.raftlog.segmented.SegmentedRaftLogInputStream;
+import org.apache.ratis.server.raftlog.segmented.SegmentedRaftLogOutputStream;
 import org.apache.ratis.server.storage.RaftStorage;
 import org.apache.ratis.statemachine.impl.BaseStateMachine;
 import org.apache.ratis.statemachine.impl.SimpleStateMachineStorage;
@@ -50,13 +51,17 @@ import org.slf4j.LoggerFactory;
 
 import java.io.File;
 import java.io.IOException;
+import java.nio.ByteBuffer;
+import java.util.Collection;
 import java.util.Collections;
 import java.util.EnumMap;
 import java.util.Objects;
 import java.util.SortedMap;
 import java.util.TreeMap;
+import java.util.concurrent.BlockingQueue;
 import java.util.concurrent.CompletableFuture;
 import java.util.concurrent.ExecutionException;
+import java.util.concurrent.LinkedBlockingQueue;
 import java.util.concurrent.TimeUnit;
 
 /**
@@ -71,6 +76,7 @@ public class SimpleStateMachine4Testing extends BaseStateMachine {
   private static final String RAFT_TEST_SIMPLE_STATE_MACHINE_TAKE_SNAPSHOT_KEY
       = "raft.test.simple.state.machine.take.snapshot";
   private static final boolean RAFT_TEST_SIMPLE_STATE_MACHINE_TAKE_SNAPSHOT_DEFAULT = false;
+  private boolean notifiedAsLeader;
 
   public static SimpleStateMachine4Testing get(RaftServerImpl s) {
     return (SimpleStateMachine4Testing)s.getStateMachine();
@@ -90,6 +96,36 @@ public class SimpleStateMachine4Testing extends BaseStateMachine {
 
   private volatile boolean running = true;
 
+  public static class Collecting {
+    public enum Type {
+      APPLY_TRANSACTION
+    }
+
+    private final EnumMap<Type, BlockingQueue<Runnable>> map = new EnumMap<>(Type.class);
+
+    BlockingQueue<Runnable> get(Type type) {
+      return map.get(type);
+    }
+
+    public BlockingQueue<Runnable> enable(Type type) {
+      final BlockingQueue<Runnable> q = new LinkedBlockingQueue<>();
+      final BlockingQueue<Runnable> previous = map.put(type, q);
+      Preconditions.assertNull(previous, "previous");
+      return q;
+    }
+
+    <T> CompletableFuture<T> collect(Type type, T value) {
+      final BlockingQueue<Runnable> q = get(type);
+      if (q == null) {
+        return CompletableFuture.completedFuture(value);
+      }
+
+      final CompletableFuture<T> future = new CompletableFuture<>();
+      final boolean offered = q.offer(() -> future.complete(value));
+      Preconditions.assertTrue(offered);
+      return future;
+    }
+  }
 
   static class Blocking {
     enum Type {
@@ -122,10 +158,13 @@ public class SimpleStateMachine4Testing extends BaseStateMachine {
   }
 
   private final Blocking blocking = new Blocking();
+  private final Collecting collecting = new Collecting();
   private long endIndexLastCkpt = RaftServerConstants.INVALID_LOG_INDEX;
   private volatile RoleInfoProto slownessInfo = null;
   private volatile RoleInfoProto leaderElectionTimeoutInfo = null;
 
+  private RaftGroupId groupId;
+
   public SimpleStateMachine4Testing() {
     checkpointer = new Daemon(() -> {
       while (running) {
@@ -141,6 +180,10 @@ public class SimpleStateMachine4Testing extends BaseStateMachine {
     });
   }
 
+  public Collecting collecting() {
+    return collecting;
+  }
+
   public RoleInfoProto getSlownessInfo() {
     return slownessInfo;
   }
@@ -163,6 +206,7 @@ public class SimpleStateMachine4Testing extends BaseStateMachine {
   public synchronized void initialize(RaftServer server, RaftGroupId groupId,
       RaftStorage raftStorage) throws IOException {
     LOG.info("Initializing " + this);
+    this.groupId = groupId;
     lifeCycle.startAndTransition(() -> {
       super.initialize(server, groupId, raftStorage);
       storage.init(raftStorage);
@@ -193,8 +237,9 @@ public class SimpleStateMachine4Testing extends BaseStateMachine {
     LogEntryProto entry = Objects.requireNonNull(trx.getLogEntry());
     put(entry);
     updateLastAppliedTermIndex(entry.getTerm(), entry.getIndex());
-    return CompletableFuture.completedFuture(
-        new SimpleMessage(entry.getIndex() + " OK"));
+
+    final SimpleMessage m = new SimpleMessage(entry.getIndex() + " OK");
+    return collecting.collect(Collecting.Type.APPLY_TRANSACTION, m);
   }
 
   @Override
@@ -210,8 +255,8 @@ public class SimpleStateMachine4Testing extends BaseStateMachine {
         termIndex.getIndex());
     LOG.debug("Taking a snapshot with t:{}, i:{}, file:{}", termIndex.getTerm(),
         termIndex.getIndex(), snapshotFile);
-    try (LogOutputStream out = new LogOutputStream(snapshotFile, false,
-        segmentMaxSize, preallocatedSize, bufferSize)) {
+    try (SegmentedRaftLogOutputStream out = new SegmentedRaftLogOutputStream(snapshotFile, false,
+        segmentMaxSize, preallocatedSize, ByteBuffer.allocateDirect(bufferSize))) {
       for (final LogEntryProto entry : indexMap.values()) {
         if (entry.getIndex() > endIndex) {
           break;
@@ -254,10 +299,9 @@ public class SimpleStateMachine4Testing extends BaseStateMachine {
           snapshot == null ? null : snapshot.getFile());
       return RaftServerConstants.INVALID_LOG_INDEX;
     } else {
-      LOG.info("Loading snapshot with t:{}, i:{}, file:{}", snapshot.getTerm(),
-          snapshot.getIndex(), snapshot.getFile().getPath());
+      LOG.info("Loading snapshot {}", snapshot);
       final long endIndex = snapshot.getIndex();
-      try (LogInputStream in = new LogInputStream(
+      try (SegmentedRaftLogInputStream in = new SegmentedRaftLogInputStream(
           snapshot.getFile().getPath().toFile(), 0, endIndex, false)) {
         LogEntryProto entry;
         while ((entry = in.nextEntry()) != null) {
@@ -265,8 +309,10 @@ public class SimpleStateMachine4Testing extends BaseStateMachine {
           updateLastAppliedTermIndex(entry.getTerm(), entry.getIndex());
         }
       }
+      // The end index is greater than last entry in indexMap as it also
+      // includes the configuration and metadata entries
       Preconditions.assertTrue(
-          !indexMap.isEmpty() && endIndex == indexMap.lastKey(),
+          !indexMap.isEmpty() && endIndex >= indexMap.lastKey(),
           "endIndex=%s, indexMap=%s", endIndex, indexMap);
       this.endIndexLastCkpt = endIndex;
       setLastAppliedTermIndex(snapshot.getTermIndex());
@@ -365,14 +411,35 @@ public class SimpleStateMachine4Testing extends BaseStateMachine {
   }
 
   @Override
-  public void notifySlowness(RaftGroup group, RoleInfoProto roleInfoProto) {
-    LOG.info("{}: notifySlowness {}, {}", this, group, roleInfoProto);
+  public void notifySlowness(RoleInfoProto roleInfoProto) {
+    LOG.info("{}: notifySlowness {}, {}", this, groupId, roleInfoProto);
     slownessInfo = roleInfoProto;
   }
 
   @Override
-  public void notifyExtendedNoLeader(RaftGroup group, RoleInfoProto roleInfoProto) {
-    LOG.info("{}: notifyExtendedNoLeader {}, {}", this, group, roleInfoProto);
+  public void notifyExtendedNoLeader(RoleInfoProto roleInfoProto) {
+    LOG.info("{}: notifyExtendedNoLeader {}, {}", this, groupId, roleInfoProto);
     leaderElectionTimeoutInfo = roleInfoProto;
   }
+
+  @Override
+  public void notifyNotLeader(Collection<TransactionContext> pendingEntries)
+      throws IOException {
+
+  }
+
+  @Override
+  public void notifyLeaderChanged(RaftGroupMemberId groupMemberId, RaftPeerId raftPeerId) {
+    if (groupMemberId.getPeerId().equals(raftPeerId)) {
+      notifiedAsLeader = true;
+    }
+  }
+
+  public boolean isNotifiedAsLeader() {
+    return notifiedAsLeader;
+  }
+
+  protected File getSMdir() {
+    return storage.getSmDir();
+  }
 }
