@@ -19,14 +19,16 @@ package org.apache.ratis;
 
 import org.apache.log4j.Level;
 import org.apache.ratis.conf.RaftProperties;
+import org.apache.ratis.metrics.RatisMetricRegistry;
 import org.apache.ratis.protocol.RaftPeerId;
 import org.apache.ratis.server.RaftServerConfigKeys;
 import org.apache.ratis.server.impl.RaftServerImpl;
+import org.apache.ratis.server.impl.RaftServerMetrics;
 import org.apache.ratis.server.simulation.MiniRaftClusterWithSimulatedRpc;
 import org.apache.ratis.proto.RaftProtos;
 import org.apache.ratis.statemachine.SimpleStateMachine4Testing;
 import org.apache.ratis.statemachine.StateMachine;
-import org.apache.ratis.util.LogUtils;
+import org.apache.ratis.util.Log4jUtils;
 import org.apache.ratis.util.TimeDuration;
 import org.junit.After;
 import org.junit.Assert;
@@ -37,8 +39,11 @@ import org.junit.Test;
 
 import java.io.IOException;
 import java.util.List;
+import java.util.SortedMap;
 import java.util.concurrent.TimeUnit;
 
+import com.codahale.metrics.Gauge;
+
 /**
  * Test Raft Server Slowness detection and notification to Leader's statemachine.
  */
@@ -46,7 +51,7 @@ import java.util.concurrent.TimeUnit;
 @Ignore
 public class TestRaftServerSlownessDetection extends BaseTest {
   static {
-    LogUtils.setLogLevel(RaftServerImpl.LOG, Level.DEBUG);
+    Log4jUtils.setLogLevel(RaftServerImpl.LOG, Level.DEBUG);
   }
 
   public static final int NUM_SERVERS = 3;
@@ -79,15 +84,30 @@ public class TestRaftServerSlownessDetection extends BaseTest {
 
   @Test
   public void testSlownessDetection() throws Exception {
-    RaftTestUtil.waitForLeader(cluster);
+    RaftServerImpl leaderServer = RaftTestUtil.waitForLeader(cluster);
     long slownessTimeout = RaftServerConfigKeys.Rpc
         .slownessTimeout(cluster.getProperties()).toIntExact(TimeUnit.MILLISECONDS);
     RaftServerImpl failedFollower = cluster.getFollowers().get(0);
 
+    RatisMetricRegistry ratisMetricRegistry =
+        RaftServerMetrics.getRaftServerMetrics(leaderServer).getRegistry();
+    SortedMap<String, Gauge> heartbeatElapsedTimeGauges =
+        ratisMetricRegistry.getGauges((s, metric) ->
+            s.contains("lastHeartbeatElapsedTime"));
+
+    String followerId = failedFollower.getId().toString();
+    Gauge metric = heartbeatElapsedTimeGauges.entrySet().parallelStream().filter(e -> e.getKey().contains(
+        followerId)).iterator().next().getValue();
+
+    long followerHeartBeatElapsedMetric = (long) metric.getValue();
+
     // fail the node and wait for the callback to be triggered
     cluster.killServer(failedFollower.getId());
     Thread.sleep( slownessTimeout * 2);
 
+    long followerHeartBeatElapsedMetricNew = (long) metric.getValue();
+    Assert.assertTrue(followerHeartBeatElapsedMetricNew > followerHeartBeatElapsedMetric);
+
     // Followers should not get any failed not notification
     for (RaftServerImpl followerServer : cluster.getFollowers()) {
       Assert.assertNull(SimpleStateMachine4Testing.get(followerServer).getSlownessInfo());
